;
;   This is the long, complex macro that handles Common Lisp structures
;   built with |defstruct|.
;
;   This file will work with either CScheme or Zen, assuming you
;   remove the lines that have ONLYZEN or ONLYCSCHEME in them,
;   respectively, first.  Something like the following works fine:
;
;   sed '/ONLYZEN/s/^/;/' <foo >bar
;
;   We build structures out of vectors.  The zeroth entry of
;   each vector is the `type' as a symbol.  The rest of each
;   vector is the actual records themselves.  We define the
;   appropriate ref and set macros for each record component.
;   We also define |copy-|, |make-|, and type predicate operations.
;
;   Note that |make-| becomes a macro itself, so it can check each
;   keyword at macro expansion time rather than execution time and
;   substitute the defaults then and there.
;
;   In addition, we define the symbol passed in as the name
;   to hold a list of the record components for reference by
;   any |defstruct|'s that |include| this one.
;
;   We must handle default values on each record entry, the use
;   of the keywords |:conc-name| and |:include|, and we add
;   to the |defsetf| association list for each record reference
;   type.
;
(define is-cscheme (eq? 'a 'A))
;
(declare (usual-integrations))
;
(set! garbage-collect-notify? #f) ; ONLYZEN
;
(define close-output-port close-port) ; ONLYZEN
;
(define-macro (dmb pattern . body) ; ONLYCSCHEME
  `(begin ; ONLYCSCHEME
     (define-macro ,pattern ,@body) ; ONLYCSCHEME
     (syntax-table-define system-global-syntax-table ; ONLYCSCHEME
        ',(car pattern) ; ONLYCSCHEME
       (macro ,(cdr pattern) ; ONLYCSCHEME
         ,@body)))) ; ONLYCSCHEME
;
(dmb (define-macro-both pattern . body) ; ONLYCSCHEME
  `(begin ; ONLYCSCHEME
     (define-macro ,pattern ,@body) ; ONLYCSCHEME
     (syntax-table-define system-global-syntax-table ; ONLYCSCHEME
         ',(car pattern) ; ONLYCSCHEME
       (macro ,(cdr pattern) ; ONLYCSCHEME
         ,@body)))) ; ONLYCSCHEME
;
(define-macro-both (defmac . body) ; ONLYCSCHEME
   (cons 'define-macro-both body)) ; ONLYCSCHEME
;
(define-macro (defmac . body) ; ONLYZEN
   (cons 'define-macro body)) ; ONLYZEN
;
(defmac (dolist iter . exp)
   `(let dolist-loop ((dolist-list ,(cadr iter)))
      (if dolist-list
         (let ((,(car iter) (car dolist-list)))
            ,@exp
            (dolist-loop (cdr dolist-list)))
         ,(if (cddr iter) (caddr iter) ''()))))
;
;
;   `When' is broken in zen.  We fix it here.
;
(defmac (when arg . stuff)
   (if (= (length stuff) 1)
      `(if ,arg ,(car stuff) '())
      `(if ,arg ,(cons 'begin stuff) '())))
;
(defmac (unless arg . stuff)
   (if (= (length stuff) 1)
      `(if ,arg '() ,(car stuff))
      `(if ,arg '() ,(cons 'begin stuff))))
;
(define (my-environment) (the-environment))
;
(defmac (defstruct name . components)
   (let ((conc-name '()))
      (define (struct-aux x n)
         (when x
            (let ((comp-name (car x)))
               (when (pair? comp-name)
                  (set! comp-name (car comp-name)))
               (let ((make-name (cons-symbol conc-name comp-name))
                     (set-name (cons-symbol conc-name comp-name "-set!")))
                `((defmac (,make-name x)
                    `(vector-ref ,x ,,n))
                 (defmac (,set-name x v)
                    `(vector-set! ,x ,,n ,v))
                 (define (,(cons-symbol make-name '-lambda) x)
                    (vector-ref x ,n))
                 (define (,(cons-symbol set-name '-lambda) x v)
                  (vector-set! x ,n v))
                 ,@(struct-aux (cdr x) (+ n 1)))))))
      (define (make-constructor name prefix prefixa)
        `(define (,name)
            (vector ',prefix
               ,@(map (lambda (a) ''()) prefixa))))
      (define (fix-comp a)
         (map (lambda (a) (if (pair? a) a (list a '()))) a)) ; ONLYCSCHEME
         (map (lambda (a) (if (pair? a) a (list a ''()))) a)) ; ONLYZEN
      (when (pair? name)
         (let ((name-stuff (cdr name)))
            (set! name (car name))
            (dolist (name-thing name-stuff)
               (case (car name-thing)
                  (':include
                     (set! components
                        (append (eval (cons-symbol "struct-" (cadr name-thing))
                                      (my-environment))
                           components)))
                  (':conc-name
                     (set! conc-name
                         (if (or (null? (cadr name-thing))
                                 (eq? (cadr name-thing) 'nil)) ""
                             (cadr name-thing))))
                  (else
                     (display-stuff "Warning: " name-thing
                                    " not understood in special" 'newline))))))
      (if (null? conc-name)
         (set! conc-name (cons-symbol name "-")))
      (let ((prefix (cons-symbol "struct-" name))
            (prefixa (fix-comp components)))
         `(begin
            (define ,prefix ',prefixa)
            (define (,(cons-symbol "copy-" name) exp)
               (vector-copy exp))
            ,(make-constructor (cons-symbol "make-" name) prefix prefixa)
            (define (,(cons-symbol name "?") x)
               (and (vector? x) (eq? (vector-ref x 0) ',prefix)))
            ,@(struct-aux components 1)))))
;
;   These functions are Elk-specific and are used for debugging.
;
(define (bt)
   (define (so s)
      (display (substring s 0 (min (string-length s) 75))) (newline))
   (let mloop ((bt (backtrace-list)))
      (if bt
         (if (eq? (vector-ref (car bt) 0) error-handler)
            (map (lambda (a)
                    (so (format #f "~a ~a" (vector-ref a 0) (vector-ref a 1))))
               (reverse (cdr bt)))
            (mloop (cdr bt)))
         (map (lambda (a)
                 (so (format #f "~a ~a" (vector-ref a 0) (vector-ref a 1))))
            (reverse (backtrace-list)))))
   '())
;
;   Some local macros we use ourselves for various purposes.  The first
;   just displays a bunch of stuff, terminating with a newline; the second
;   builds up a symbol from a bunch of other strings and/or symbols.
;
(define (display-stuff . stuff)
   (dolist (thing stuff)
      (if (eq? thing 'newline) (newline)
          (display thing))))
;
(define (display-stuff-port port . stuff)
   (dolist (thing stuff)
      (if (eq? thing 'newline) (newline port)
          (display thing port))))
;
(define (cons-symbol . stuff)
   (string->symbol
      (apply string-append
         (map (lambda (a)
                 (if (symbol? a) (symbol->string a)
                    (if (number? a) (number->string a)
                       (if (char? a) (list->string (list a)) a))))
              stuff))))
;
;   Returns a list (0 1 2 3 4 5 ...) of at least the length of the list
;   passed.  Caches it.
;
(define *iota-length* -1)
(define *iota-list* '())
;
(define (new-iota n m)
   (if (> n m)
      (begin
         (set! *iota-length* (- m 3))
         '())
      (cons n (new-iota (+ n 1) m))))
;
(define (iota-l l)
   (iota (length l)))
;
(define (iota n)
   (if (> n *iota-length*)
      (set! *iota-list* (new-iota 0 (* 2 (+ 3 n)))))
   *iota-list*)
;
(define (map-vector p v)
   (map p (vector->list v)))
;
;   We make nth a macro that special-cases the case of a constant n.
;
(define (nth-normal n l)
   (if (< n 1) (car l) (nth-normal (- n 1) (cdr l))))
;
(define (nth-expand n l)
   (if (> n 2)
      (nth-expand (- n 3) (list 'cdddr l))
      (list (if (= n 0) 'car (if (= n 1) 'cadr 'caddr)) l)))
;
(defmac (nth n l)
   (if (number? n)
      (nth-expand n l)
      `(nth-normal ,n ,l)))
;
(define (rand m) ; ONLYZEN
   (modulo (random) m)) ; ONLYZEN
;
(define (rand m) ; ONLYCSCHEME
   (random m)) ; ONLYCSCHEME
;
(define (runtime) 0) ; ONLYZEN
;
(define is-cscheme (eq? 'a 'A))
;
;   This function is a `safe-cdr' that doesn't crash if the
;   passed thing isn't a pair.
;
(define (safe-cdr a)
   (if (pair? a) (cdr a) '()))
;
(define (safe-car a)
   (if (pair? a) (car a) '()))
;
(defmac (safe-cadr a)
   `(let ((a ,a))
      (if (pair? a) (if (pair? (cdr a)) (cadr a) '()) '())))
;
(define (safe-cadr-lambda a)
   (safe-cadr a))
;
(define (safe-caddr a)
   (safe-cadr (safe-cdr a)))
;
(define (safe-cddr a)
   (safe-cdr (safe-cdr a)))
;
(define (my-sublist l a b)
   (if (<= a 0)
      (if (> b 0)
         (if l (cons (car l) (my-sublist (cdr l) a (- b 1))) '())
         '())
      (if l (my-sublist (cdr l) (- a 1) b) '())))
;
;   Map isn't guaranteed to do things in order.  This is.
;
(define (map-inorder f l)
   (if l (cons (f (car l)) (map-inorder f (cdr l))) '()))
;
(define (map-inorder2 f l l2)
   (if l (cons (f (car l) (car l2)) (map-inorder2 f (cdr l) (cdr l2))) '()))
;
(define (union a b)
   (let mloop ((sofar b) (a a))
      (if a
         (mloop (if (memq (car a) b) sofar (cons (car a) sofar)) (cdr a))
         sofar)))
;
(define (intersection a b)
   (let mloop ((sofar '()) (a a))
      (if a
         (mloop (if (memq (car a) b) (cons (car a) sofar) sofar) (cdr a))
         sofar)))
;
(define (set-difference a b)
   (let mloop ((sofar '()) (a a))
      (if a
         (mloop (if (memq (car a) b) sofar (cons (car a) sofar)) (cdr a))
         sofar)))
;
;   All debugging stuff goes here.
;
(define *debug-flags* '())
;
(define (set-debug . rest)
   (set! *debug-flags* rest))
;
(define (set-debug-all)
   (set! *debug-flags* '(a b c d e f g h i j k l m n o p q r s
                         t u v w x y z 0 1 2 3 4 5 6 7 8 9)))
;
(defmac (DBG flag expr)
   `(if (memq ,flag *debug-flags*) ,expr '()))
;
(define (identity a) a)
;
(define (bound? not-likley-en-more-so) ; ONLYCSCHEME
   (environment-bound? (the-environment) not-likley-en-more-so)) ; ONLYCSCHEME
;
;   In CScheme, we are careful to load the .scm file if the .com file is
;   less recent.
;
(define (sload filename) ; ONLYCSCHEME
   (let* ((scm-name (string-append filename ".scm")) ; ONLYCSCHEME
          (com-name (string-append filename ".com"))) ; ONLYCSCHEME
      (if (and (file-exists? com-name) ; ONLYCSCHEME
               (> (file-modification-time com-name) ; ONLYCSCHEME
                  (file-modification-time scm-name))) ; ONLYCSCHEME
         (load com-name) ; #f system-global-syntax-table #f) ; ONLYCSCHEME
         (load scm-name)))) ; ONLYCSCHEME
;
(define (oload filename) ; ONLYCSCHEME
   (if (file-exists? filename) ; ONLYCSCHEME
      (load filename) ; ONLYCSCHEME
      (load (string-append *src-dir* filename)))) ; ONLYCSCHEME
;
;   Zen doesn't have a system command (that I can find); CScheme has
;   unix/system.
;
(define (system s) (display-stuff s 'newline)) ; ONLYZEN
(define (system s) (unix/system s)) ; ONLYCSCHEME
