;
;   Now, the simulation code.
;
;   For simulation we need the following data structures:
;
;   - Global
;      - A list of all of the transitions
;      - A log of firings/values (for backtracking and search)
;      - An array of lists of all [] places
;      - An array of lists of all [] head places
;   - Per transition
;      - A dirty flag indicating whether to recalculate enabling
;      - An enabled/disabled flag
;   - Per place
;      - The current marking
;      - An indication of if it is a [] place
;      - Min and max times (if constrained)
;      - An indication of internal or external choice
;      - An indication of internal or external time
;      - An index count for the place
;      - A random number generator/lookup table
;   - Per data flow register
;      - The current value
;   - Per data flow node
;      - A list of input sources
;      - A function to compute the output values
;      - A list of list of outputs
;
(declare (usual-integrations))
;
(define *sgraph* '())
;
(define *current-state* 0)
;
(define *current-region* '())
;
(define *state-region* '())
;
;   We have some functions here for logging statistics so we can
;   optimize.
;
(define *stat-asc* '())
;
(define (log-info sym dat)
   (let ((asc (assoc sym *stat-asc*)))
      (if asc
         (set-cdr! asc (cons dat (cdr asc)))
         (begin
            (set! *stat-asc* (cons (cons sym '()) *stat-asc*))
            (log-info sym dat)))))
;
(define (log-inc sym)
   (let ((asc (assoc sym *stat-asc*)))
      (if asc
         (set-cdr! asc (+ 1 (cdr asc)))
         (begin
            (set! *stat-asc* (cons (cons sym 0) *stat-asc*))
            (log-inc sym)))))
;
(define (log-time sym otime)
   (let ((asc (assoc sym *stat-asc*)))
      (if asc
         (begin
            (set-car! (cdr asc) (+ 1 (cadr asc)))
            (set-car! (cddr asc) (+ (- (runtime) otime) (caddr asc))))
         (begin
            (set! *stat-asc* (cons (list sym 0 0) *stat-asc*))
            (log-time sym otime)))))
;
(define t3 '())
;
;
;   First, we build up lists of all of the different types of nodes.
;   This function takes a list of nodes, and it returns a list of
;   lists, each of the latter of the form (type node node node ...)
;
;   Must pass in an empty list to start it off.
;
(define (list-to-types n l)
   (if n
      (let* ((fn (car n))
             (ty (node-type fn))
             (as (assoc ty l)))
         (if as
            (begin
               (set-cdr! as (cons fn (cdr as)))
               (list-to-types (cdr n) l))
            (list-to-types (cdr n) (cons (list ty fn) l))))
      l))
;
(define (ascii->number n v a)
   (if n
      (if (eq? (car n) #\.)
         (+ (ascii->number (cdr n) 1 0) (/ a v))
         (ascii->number (cdr n) (* 10 v)
            (+ a (* v (- (char->integer (car n)) (char->integer #\0))))))
      a))
;
(define (ascii->symbol n)
   (let ((r (string->symbol (list->string n))))
      (if (eq? r 'inf) 'infinity r)))
;
;   This is a generic but stupid parsing routine.  It takes two strings,
;   one a template and the other a source, and tries to parse.  It
;   returns either #f (for failure) or else a list of the resulting
;   arguments (or #t if there are no arguments.)
;
;   The template is just a set of characters that must match
;   exactly, except that a # is matched with an integer or a symbol.
;   In addition, if the final symbol we end up with is 'inf, we
;   change it to 'infinity.
;
(define (smatch template s)
   (let loop ((tl (string->list template)) (sl (string->list s))
              (args '()) (sofar '()))
      (if (null? tl)
         (if (null? sl) (or args #t) #f)
         (if (null? sl) #f
            (if (eq? (car tl) #\#)
               (let ((r (char->integer (car sl))))
                  (if (or (<= (char->integer #\0) r (char->integer #\9))
                          (<= (char->integer #\a) r (char->integer #\z))
                          (<= (char->integer #\A) r (char->integer #\Z))
                          (eq? r (char->integer #\-))
                          (eq? r (char->integer #\*))
                          (eq? r (char->integer #\.)))
                     (loop tl (cdr sl) args (cons (car sl) sofar))
                     (loop (cdr tl) sl
                        (if (pair? sofar)
                           (if (<= (char->integer #\0)
                                   (char->integer (car (reverse sofar)))
                                   (char->integer #\9))
                              (cons (ascii->number sofar 1 0) args)
                              (cons (ascii->symbol (reverse sofar)) args))
                           (merror "Empty list in smatch" sofar)) '())))
               (if (eq? (car tl) (car sl))
                  (loop (cdr tl) (cdr sl) args sofar) #f))))))
;
;   This one looks for a match in the list.
;
(define (smatch-list templates string)
;  (display-stuff "Matching <" string ">" 'newline)
   (let loop ((t 1) (tl templates))
      (if tl
         (let ((r (smatch (car tl) string)))
            (if r (cons t r) (loop (+ t 1) (cdr tl))))
         (list 0))))
;
;   Our forever interval.
;
(define *iforever* (cons 0 'infinity))
;
(define *ilater* '(infinity . infinity))
;
;   We use the same *value* but a different cons cell for *ieps* so we
;   can calculate splace-triv? easily.
;
(define *ieps* (cons 0 'infinity))
;
(define (mstring->number s)
   (let fori ((i 0) (val 0) (rad 0))
      (if (>= i (string-length s))
         val
	 (if (eq? (string-ref s i) #\.)
	    (fori (+ i 1) val 10)
	    (let ((c (char->integer (string-ref s i))))
	       (if (<= (char->integer #\0) c (char->integer #\9))
		  (if (> rad 0)
                     (fori (+ i 1) (+ val (/ (- c (char->integer #\0)) rad))
		        (* rad 10))
		     (fori (+ i 1) (+ (* 10 val) (- c (char->integer #\0))) 0))
	          (merror "Bad char in mstring->num" s)))))))
;
;   Check that what we got is legal.  We allow:
;
;   integer, integer
;   integer, infinity
;   integer, symbol
;   symbol, integer
;   symbol, infinity
;   symbol, symbol
;
(define (timing-resolve n default)
   (if (or (eq? n 'infinity) (number? n))
      n
      (let ((star? -1) (ns (symbol->string n)))
         (let loop ((i (- (string-length ns) 2)))
            (when (> i 1)
               (if (eq? (string-ref ns i) #\*)
                  (set! star? i)
                  (loop (- i 1)))))
         (let* ((sym (if (< star? 0) n
                         (string->symbol (substring ns 0 star?))))
                (mult (if (< star? 0) 1
                          (mstring->number (substring
                                           ns (+ star? 1)
                                           (string-length ns)))))
                (v (or (assoc sym *global-assoc*)
                      (begin
                         (set! *global-assoc*
                            (cons (cons sym -1) *global-assoc*))
                         (car *global-assoc*)))))
           (if (eq? (cdr v) -2)
              default
              (if (eq? (cdr v) -3)
                 'infinity
                 (if (eq? (cdr v) -1)
                    n
                    (if (number? (cdr v))
                       (* mult (cdr v))
                       (merror "Funky value somehow in timing variable"
                               v)))))))))
;
(define (say-place-time p beg end)
   (if (and (eq? end 'infinity)
          (or (eq? beg 0) (eq? beg 'infinity)))
      (splace-time-marking-set! p
         (if (eq? beg 0) *iforever* *ilater*))
      (let* ((t (cons beg end))
             (sofar (if (splace-time-constraint? p)
                       (sgraph-con-times *sgraph*)
                       (sgraph-beh-times *sgraph*)))
             (v (assoc t sofar)))
         (splace-time-marking-set! p 'something)
         (if v
            (set-cdr! v (cons p (cdr v)))
            (if (splace-time-constraint? p)
               (sgraph-con-times-set! *sgraph* (cons (list t p) sofar))
               (sgraph-beh-times-set! *sgraph* (cons (list t p) sofar)))))))
;
(define *timing-important?* #f)
;
(define (timing-important?)
   (or (sgraph-con-times *sgraph*) (sgraph-beh-times *sgraph*)))
;
(define (fix-up-node-snode new-node old-node)
   (node-snode-set! old-node new-node)
   (let fori ((i 1))
      (when (< i (vector-length old-node))
         (vector-set! new-node i (vector-ref old-node i))
         (fori (+ i 1))))
   new-node)
;
;   Shows timing variables for constraint places.
;
(define (show-constraint-place-results)
   (define (time-scale n)
      (if n (* 1.0 (sgraph-global-tscale *sgraph*) n) n))
   (let ((tassoc '()))
      (for-each (lambda (p)
                   (when (and (splace-time-constraint? p) (node-label p)
                            (not (splace-triv? p))
                            (splace-trange-seen p))
                      (let* ((ts (car (node-label p)))
                             (v (or (assoc ts tassoc)
                                    (begin
                                       (set! tassoc
                                          (cons (cons ts '()) tassoc))
                                       (car tassoc))))
                             (r (splace-trange-seen p))
                             (res (cons (time-scale (car r))
                                        (time-scale (cdr r)))))
                        (when (not (member res (cdr v)))
                           (set-cdr! v (cons res (cdr v)))))))
         (sgraph-places *sgraph*))
      (for-each (lambda (t) (display-stuff t 'newline)) tassoc)))
;
;   Functions that build the auxilliary simulation node.
;
(define (place-to-splace p)
   (let ((nn (fix-up-node-snode (make-splace) p)))
      (splace-curval-set! nn 'unknown)
      (splace-time-marking-set! nn *iforever*)
      (splace-time-constraint?-set! nn #f)
      (splace-state-value-set! nn 0)
      (when (eq? (node-input-output? p) '*input*)
         (splace-time-constraint?-set! nn #t))
      (let loop ((ml (node-label p)))
         (when ml
            (let ((m (smatch-list '("[]" "[!]" "[#]" "[#,#]")
                                  (symbol->string (car ml)))))
               (case (car m)
                  (0 (when (eq? #\[ (car (string->list
                                            (symbol->string (car ml)))))
                        (merror "Syntax error on place" (car ml)))
                     (splace-wire-set! nn (car ml)))
                  (1 (splace-marked?-set! nn *iforever*)
                     (splace-time-marking-set! nn *ieps*))
                  (2 (splace-marked?-set! nn *iforever*)
                     (splace-time-marking-set! nn *ieps*)
                     (splace-immediate?-set! nn #t))
                  (3 (say-place-time nn (timing-resolve (cadr m) 0)
                                        (timing-resolve (cadr m) 'infinity)))
                  (4 (say-place-time nn (timing-resolve (caddr m) 0)
                                        (timing-resolve (cadr m) 'infinity)))
                  (else (merror "Internal error ??? in smatch-list"))))
               (loop (cdr ml))))
         (when (splace-eps? nn)
            (node-orbit-set! p #f)
            (node-orbit-set! nn #f))
      nn))
;
(define (splace-eps? sn)
   (eq? *ieps* (splace-time-marking sn)))
;
(define (splace-triv? sn)
   (or (eq? *iforever* (splace-time-marking sn))
      (eq? *ilater* (splace-time-marking sn))))
;
(define (splace-timing? sn)
   (not (splace-triv? sn)))
;
(define (trans-to-strans t)
   (let ((nn (fix-up-node-snode (make-strans) t))
         (ctpv (make-vector 3 '())))
      (strans-dirty?-set! nn #t)
      (strans-enabled?-set! nn #f)
      (strans-dead?-set! nn #f)
      (strans-input?-set! nn
         (intersection (sgraph-inputs *sgraph*) (label-list t)))
      (strans-output?-set! nn
         (intersection (sgraph-outputs *sgraph*) (label-list t)))
      (strans-state-value-set! nn 0)
      (strans-eps-connected-set! nn #f)
      (strans-eps-conflicts-set! nn '())
      (let loop ((el (node-outedges t)))
         (when el
            (let ((n (edge-outnode (car el))))
               (when (and (node-type? n 'P)
                          (null? (node-outedges n)))
                  (strans-dead?-set! nn #t))
               (loop (cdr el)))))
      (let loop ((el (node-inedges t)))
         (when el
            (let ((n (edge-innode (car el))))
               (if (node-type? n 'P)
                  (when (not (splace-eps? (node-snode n)))
                     (if (eq? (splace-time-marking (node-snode n)) *ilater*)
                        (strans-much-later?-set! nn #t)
                        (when (and (not (splace-triv? (node-snode n)))
                                   (not (splace-immediate? (node-snode n))))
                           (if (splace-time-constraint? (node-snode n))
                              (when (not (edge-ignore-timing? (car el)))
                                 (vector-set! ctpv 2
                                    (cons (node-snode n) (vector-ref ctpv 2))))
                              (let ((i (if (edge-ignore-timing? (car el))
                                           1 0)))
                                 (vector-set! ctpv i
                                    (cons (node-snode n)
                                       (vector-ref ctpv i))))))))
                  (set! *dataflow-important?* #t))
               (loop (cdr el)))))
      (when (> (length (vector-ref ctpv 0)) 1)
         (merror "Multiple beh places" 0))
      (when (vector-ref ctpv 0)
         (vector-set! ctpv 1 (cons (car (vector-ref ctpv 0))
                                   (vector-ref ctpv 1))))
      (strans-timed-places-set! nn ctpv)
      nn))
;
;   We find the input to the register here to make firing of the
;   register faster.
;
(define (reg-to-sreg r)
   (let ((nn (fix-up-node-snode (make-sdf) r)))
      (let* ((el (node-inedges r))
             (in (edge-innode
                    (if (= (edge-outpos (car el)) 1) (car el) (cadr el)))))
         (sdf-op-set! nn (lambda () (sdf-curval (node-snode in)))))
      (sdf-reg?-set! nn #t)
      (sdf-curval-set! nn 'unknown)
      (sdf-state-value-set! nn 0)
      nn))
;
(define (join-to-sjoin j)
   (let ((nn (fix-up-node-snode (make-sdf) j)))
      (sdf-op-set! nn (lambda (a) a))
      (sdf-reg?-set! nn #f)
      (sdf-curval-set! nn 'unknown)
      (sdf-state-value-set! nn 0)
      nn))
;
;   Functions are then straightforward.  We do collect up the
;   nodes that go to this function in the appropriate order so
;   that evaluation is faster.
;
(define (func-to-sfunc f)
   (let ((nn (fix-up-node-snode (make-sdf) f)))
      (sdf-op-set! nn (primitive-version (car (node-label f))))
      (sdf-reg?-set! nn #f)
      (sdf-curval-set! nn 'unknown)
      (sdf-state-value-set! nn 0)
      nn))
;
(define (another-df-pass nl)
   (when nl
      (let* ((f (car nl))
             (el (node-inedges f))
             (args (make-vector (length el) '())))
         (let loop ((el el))
            (when el
               (vector-set! args (edge-outpos (car el)) (edge-innode (car el)))
               (loop (cdr el))))
         (sdf-args-set! f (vector->list args)))
      (another-df-pass (cdr nl))))
;
;   Now we introduce our arithmetic on times.  Times are nonnegative
;   integers or 'infinity.
;
(defmac (tmin a b)
   `(let ((a ,a) (b ,b))
      (if (eq? a 'infinity) b
         (if (eq? b 'infinity) a
            (if (< a b) a b)))))
;
(define (tmax a b)
   (if (or (eq? a 'infinity) (eq? b 'infinity)) 'infinity (max a b)))
;
(define (t<= a b)
   (if (eq? b 'infinity) #t (if (eq? a 'infinity) #f (<= a b))))
;
(defmac (t< a b)
   `(let ((a ,a) (b ,b))
      (if (eq? b 'infinity)
         (not (eq? a 'infinity))
            (if (eq? a 'infinity) #f (< a b)))))
;
(defmac (t+ a b)
   `(let ((a ,a) (b ,b))
      (if (or (eq? a 'infinity) (eq? b 'infinity)) 'infinity (+ a b))))
;
(define (t- a)
   (if (eq? a 'infinity) 'neginfinity (- a)))
;
(defmac (t= a b)
   `(eqv? ,a ,b))
;
;   Intervals are similar.  Intervals are one of:
;      #f --> impossible interval
;      Or a pair of times, nonnegative, with the second greater
;         than or equal to the first.
;
;   We try not to mess with *iforever* to avoid garbage.
;
(define (iintersect a b)
   (if (and a b)
      (if (eq? a *iforever*) b
         (if (eq? b *iforever*) a
            (let ((mintime (tmax (car a) (car b)))
                  (maxtime (tmin (cdr a) (cdr b))))
               (if (t<= mintime maxtime)
                  (cons mintime maxtime)
                  #f))))
      #f))
;
;   This function adds a scalar time and an interval.
;
(define (i+t i t)
   (cons (t+ (car i) t) (t+ (cdr i) t)))
;
;   This function calculates a random time in the range.  If the last
;   time is infinity, then we use the first time plus 1, since it hardly
;   matters.
;
(define (ritime i)
   (if i
      (if (eq? (cdr i) 'infinity) (+ 1 (car i))
         (+ (car i) (rand (+ 1 (- (cdr i) (car i))))))
      (merror "No interval in supplied firing" i))) ; should really backtrack
;
(define (in-interval t i)
   (and i (t<= (car i) t) (t<= t (cdr i))))
;
;   This version is somewhat different.  Instead of treating each timing
;   value as a separate part of the state, thus allowing a seriously
;   exponential number of states in timing situations and requiring
;   drastic remedies to simulate, instead we treat a state as a marking
;   plus a convex (n+1)-dimensional polygon of permissable timing values,
;   where n is the number of marked timed places.  (We use the 0th row and
;   column to represent 0.)
;
;   The polygon is represented by the matrix with the following conventions:
;
;   x_0 .. x_n represent the timing variables.  x_0 is fixed as the constant
;   0, and x_1 .. x_n represent the actual values of the clocks in the
;   individual places.  Matrix element a_{ij} represents the equation
;
;      x_i + a_{ij} \ge x_j
;
;   With x_0 = 0, we can represent equations of the form:
;
;      x_i \ge k, k \ge x_i, x_i + k \ge x_j, and x_i = x_j.
;
;   There are multiple matrices that represent a specific polygon.
;   There is a single canonical representation of a specific polygon,
;   that can be determined by running the all pairs shortest path
;   algorithm over the matrix.
;
;   We have the following useful pieces of information about this matrix:
;
;      a_{ii} = 0
;
;   If after running the APSP algorithm over the matrix, any of the
;   diagonal elements are negative, this means that the inequalities
;   have no solution.
;
;   A legal assignment can be found by simply assigning to the
;   variables one at a time, such that all prior values are respected.
;   No backtracking is necessary.  (This assumes the canonical matrix,
;   and it is not true for a non-canonical matrix.)
;
;   The intersection of two such polygons is easily found; just
;   find the canonical matrices, then create a third matrix by
;   using the smaller value of each element from the two other
;   matrices, and then running the APSP algorithm.
;
;   For the canonical matrix:
;
;      a_{ij} + a_{jk} \ge a_{ik}
;      a_{ij} + a_{ji} \ge 0
;
;   The all-pairs shortest path algorithm.
;
(defmac (dref a i j)
   `(vector-ref (vector-ref ,a ,i) ,j))
;
(defmac (dset! a i j v)
   `(vector-set! (vector-ref ,a ,i) ,j ,v))
;
(defmac (dcopy v)
   `(vector-map ,v vector-copy))
;
(define *default-0-region* (vector (vector 0)))
;
(define (dmake n init)
;  (log-info 'dmake n)
   (if (< n 2)
      *default-0-region*
      (let ((v (make-vector n '())))
         (let loop ((i 0))
            (when (< i n)
               (vector-set! v i (make-vector n init))
               (loop (+ i 1))))
         v)))
;
(define (zero-diag v)
  (let loop ((i (- (vector-length v) 1)))
     (when (>= i 0)
        (dset! v i i 0)
        (loop (- i 1))))
  v)
;
;   This function returns 0 if successful, or a negative number if the
;   shape described includes no points.  If the shape described contains
;   no points, then this algorithm may abort prematurely, but that is okay,
;   because the result of apsp is undefined if there are any negative
;   cycles.
;
(define *slow-apsp?* #f)
;
(define (apsp v)
   (let ((n (vector-length v)))
;     (log-info 'apsp n)
      (let lp1 ((i 0))
         (when (< i n)
            (let lp2 ((j 0))
               (when (< j n)
                  (when (not (eq? (dref v j i) 'infinity))
                     (let lp3 ((k 0))
                        (when (< k n)
                           (when (not (eq? (dref v i k) 'infinity))
                              (let ((t (+ (dref v j i) (dref v i k)))
                                    (t2 (dref v j k)))
                                 (when (or (eq? t2 'infinity) (< t t2))
                                    (dset! v j k t))))
                           (lp3 (+ k 1)))))
                  (if (< (dref v j j) 0)
                     (dset! v 0 0 -1)
                     (lp2 (+ j 1)))))
            (lp1 (+ i 1)))))
   (dref v 0 0))
;
;   For debugging, we have this routine which checks for consistency.
;
(define (apsp-check v)
   (let ((n (vector-length v)))
;     (log-info 'apsp n)
      (let lp1 ((i 0))
         (when (< i n)
            (let lp2 ((j 0))
               (when (< j n)
                  (when (not (eq? (dref v j i) 'infinity))
                     (let lp3 ((k 0))
                        (when (< k n)
                           (when (not (eq? (dref v i k) 'infinity))
                              (let ((t (+ (dref v j i) (dref v i k)))
                                    (t2 (dref v j k)))
                                 (when (or (eq? t2 'infinity) (< t t2))
                                    (dset! v j j -1))))
                           (lp3 (+ k 1)))))
                  (if (< (dref v j j) 0)
                     (dset! v 0 0 -1)
                     (lp2 (+ j 1)))))
            (lp1 (+ i 1)))))
   (dref v 0 0))
;
;   This function is just like the previous, but it only recomputes the
;   0th row (if that's the only one that needs recomputation.)
;
(define (apsp-0row v)
;  (set! t3 (runtime))
   (if *slow-apsp?*
      (apsp v)
      (let ((n (vector-length v)))
;     (log-info 'apsp-0row n)
         (let lp1 ((i 0))
            (when (< i n)
               (when (not (eq? (dref v 0 i) 'infinity))
                  (let lp3 ((k 1))
                     (when (< k n)
                        (when (not (eq? (dref v i k) 'infinity))
                           (let ((t (+ (dref v 0 i) (dref v i k)))
                                 (t2 (dref v 0 k)))
                              (when (or (eq? t2 'infinity) (< t t2))
                                 (dset! v 0 k t))))
                        (lp3 (+ k 1)))))
            (lp1 (+ i 1))))
         0)))
;
;   This one is specialized for a new 0 row *and* column.
;
(define (apsp-new0 v)
;  (set! t3 (runtime))
   (if *slow-apsp?*
      (apsp v)
      (let ((n (vector-length v)))
         (let lp1 ((i 1))
            (when (< i n)
               (when (not (eq? (dref v 0 i) 'infinity))
                  (let lp3 ((k 1))
                     (when (< k n)
                        (when (not (eq? (dref v i k) 'infinity))
                           (let ((t (+ (dref v 0 i) (dref v i k)))
                                 (t2 (dref v 0 k)))
                              (when (or (eq? t2 'infinity) (< t t2))
                                 (dset! v 0 k t))))
                        (lp3 (+ k 1)))))
               (when (not (eq? (dref v i 0) 'infinity))
                  (let lp3 ((k 1))
                     (when (< k n)
                        (when (not (eq? (dref v k i) 'infinity))
                           (let ((t (+ (dref v k i) (dref v i 0)))
                                 (t2 (dref v k 0)))
                              (when (or (eq? t2 'infinity) (< t t2))
                                 (dset! v k 0 t))))
                        (lp3 (+ k 1)))))
               (lp1 (+ i 1))))
         (let lp1 ((i 1))
            (when (< i n)
               (when (not (eq? (dref v 0 i) 'infinity))
                  (let lp3 ((k 1))
                     (when (< k n)
                        (when (not (eq? (dref v k 0) 'infinity))
                           (let ((t (+ (dref v 0 i) (dref v k 0)))
                                 (t2 (dref v k i)))
                              (when (or (eq? t2 'infinity) (< t t2))
                                 (dset! v k i t))))
                        (lp3 (+ k 1)))))
               (lp1 (+ i 1))))
         0)))
;
;   If we know that the only change we have made is to reduce a single
;   constraint, we can use this n^2 version instead.
;
(define (apsp1 v ii jj)
   (if *slow-apsp?*
      (apsp v)
      (let ((n (vector-length v)))
;     (log-info 'apsp1 n)
         (let lp1 ((j 0))
            (when (< j n)
               (let ((t (+ (dref v ii jj) (dref v jj j))))
                  (when (< t (dref v ii j))
                     (dset! v ii j t)))
               (lp1 (+ j 1))))
         (let lp1 ((i 0))
            (when (< i n)
               (let ((t (t+ (dref v i ii) (dref v ii jj))))
                  (when (< t (dref v i jj))
                     (dset! v i jj t)
                     (let lp2 ((j 0))
                        (when (< j n)
                           (let ((t2 (+ t (dref v jj j))))
                              (when (< t2 (dref v i j))
                                 (dset! v i j t2)))
                           (lp2 (+ j 1))))))
               (lp1 (+ i 1))))
         0)))
;
;   This function hashes an array of integers or 'infinity.
;   For infinity, we use the value -31.
;
;   We try to keep things in fixnum range, but we don't insist.
;   If we get many more than 1M arrays, then this will seriously
;   hurt our hash performance.
;
(define (array-hash-aux a n s)
   (let* ((v (vector-ref a n))
          (v2 (+ (* 7 s) v)))
      (if (= n 0) (modulo v2 2790001)
         (array-hash-aux a (- n 1) v2))))
;
(define *was-a-new-array?* #f)
;
(define (array-hash a)
   (let ((n (- (vector-length a) 1)))
      (if *combine-rows?*
         (let loop ((i n) (s 131))
            (let* ((sv (vector-ref a i))
                        (h (array-hash-aux sv n 11))
                   (val (cons h sv))
                      (m? (iset-member**? *array-hash* val)))
                  (if m?
                     (vector-set! a i (cdar m?))
                     (iset-insert *array-hash* (list val)))
                  (if (= i 0)
                     (modulo (+ (* 3 s) h) 2790001)
                     (loop (- i 1) (+ (* 3 s) h)))))
         (let loop ((i n) (s 131))
            (if (= i 0)
                  (array-hash-aux (vector-ref a i) n (* 3 s))
                  (loop (- i 1) (array-hash-aux (vector-ref a i) n (* 3 s))))))))
;
(define *array-hash* '())
;
(define *combine-arrays?* #t)
;
(define *combine-rows?* #t)
;
(define (new-array-hash)
   (set! *array-hash* (new-hash caar)))
;
(define (copy-vector v1 v2 n)
   (vector-set! v1 n (vector-ref v2 n))
   (if (> n 0)
      (copy-vector v1 v2 (- n 1))
      v1))
;
(define (cache-array v)
   (set! *was-a-new-array?* #f)
   (let* ((k (array-hash v))
          (val '()) (m? '()))
      (when (or *require-exact-region-match?* (not *combine-rows?*)
                (not *destroy-shapes?*))
         (set! val (cons k v))
         (set! m? (iset-member**? *array-hash* val)))
      (if *require-exact-region-match?*
         (if m?
            (cdar m?)
            (begin
               (set! *was-a-new-array?* #t)
               (iset-insert *array-hash* (list val))
               v))
         (if m?
            (if *destroy-shapes?*
               (begin
                  (copy-vector v (cdar m?) (- (vector-length v) 1))
                  v)
               (cdar m?))
            (begin
               (if *destroy-shapes?*
                  (if val
                     (begin
                        (iset-insert *array-hash*
                           (list (cons k (vector-copy v))))
                        v)
                     (vector-copy v))
                  (begin
                     (iset-insert *array-hash* (list (cons k v)))
                     v)))))))
;
;   Assoc array.
;
(define (cache-array* v)
   (let* ((k (array-hash v))
          (val (cons k v)) (m? (iset-member**? *array-hash* val)))
      (or m? (iset-insert *array-hash* (list val)))))
;
;   This function takes the intersection of two state shapes.  It returns
;   #f if there is none, or else the shape if there is one.  It also sets
;   the global variable *shape-subset* to 'a, 'b, or 'both if one shape is
;   a subset of the other, or #f if neither is a subset of the other.  If
;   #f is returned, then *shape-subset* is undefined.  It assumes that the
;   two inputs are canonical.
;
;   This subroutine is n^2, unless it must call apsp in which case it is
;   n^3.  We only call apsp if neither shape is a subset of the other and
;   if they are not obviously disjoint (two constraints which are mutually
;   disjoint.)
;
(define *shape-subset* #f)
;
(define *use-shape-equiv-subset?* #t)
;
(define (shape-equiv-subset? a b)
   (set! *shape-subset* #f)
   (let ((a-sub #t) (b-sub #t) (n (vector-length a)))
      (let loop ((i 0))
         (when (< i n)
            (let ((bi0 (dref b i 0)) (ai0 (dref a i 0)))
               (when (not (eq? bi0 ai0))
                  (if (< ai0 bi0)
                     (set! b-sub #f)
                     (set! a-sub #f))))
            (loop (+ i 1))))
      (let loop ((i 0))
         (when (< i n)
            (let* ((bi (vector-ref b i)) (ai (vector-ref a i))
		   (bi0 (vector-ref bi 0)) (ai0 (vector-ref ai 0)))
               (when (not (eq? bi ai))
                  (let loop ((j 1))
                     (when (< j n)
                        (let ((bij (vector-ref bi j)) (aij (vector-ref ai j)))
                           (when (not (= bij aij))
                              (let ((p (vector-ref *timed-token-ar* j)))
                                 (if (< aij bij)
                                    (when (and b-sub
                                             (or (splace-time-constraint? p)
                                                   (< (- aij bi0)
                                                   (car (splace-time-marking
                                                           p)))))
                                       (set! b-sub #f))
                                    (when (and a-sub
                                             (or (splace-time-constraint? p)
                                                   (< (- bij ai0)
                                                   (car (splace-time-marking
                                                           p)))))
                                       (set! a-sub #f))))))
                        (loop (+ j 1))))))
            (if (or a-sub b-sub)
               (loop (+ i 1)))
               #f))
      (when (or a-sub b-sub)
         (set! *shape-subset*
            (if a-sub
               (if b-sub 'both 'a)
               (if b-sub 'b (merror "What happened?" 0)))))
      *shape-subset*))
;
(define (shape-subset? a b)
   (if *use-shape-equiv-subset?*
      (shape-equiv-subset? a b)
      (shape-normal-subset? a b)))
;
(define (shape-normal-subset? a b)
   (set! *shape-subset* #f)
   (let ((a-sub #t) (b-sub #t) (n (vector-length a)))
      (let loop ((i 0))
         (when (< i n)
            (let ((bi (vector-ref b i)) (ai (vector-ref a i)))
               (when (not (eq? bi ai))
                  (let loop ((j 0))
                     (when (< j n)
                        (let ((bij (vector-ref bi j)) (aij (vector-ref ai j)))
                           (when (not (= bij aij))
                              (if (< bij aij)
                                 (set! a-sub #f)
                                 (set! b-sub #f))))
                        (loop (+ j 1))))))
            (if (or a-sub b-sub)
               (loop (+ i 1)))
               #f))
      (when (or a-sub b-sub)
         (set! *shape-subset*
            (if a-sub
               (if b-sub 'both 'a)
               (if b-sub 'b (merror "What happened?" 0)))))
      *shape-subset*))
;
;   This checks two shapes for equality.
;
(define (shapes-eqv? a b)
   (set! *shape-subset* (if (equal? a b) 'both #f))
   *shape-subset*)
;
(define (shape-intersect a b)
   (set! *shape-subset* #f)
   (let ((c (dcopy a)) (a-sub #t) (b-sub #t) (bail #f) (n (vector-length a)))
      (let loop ((i 0))
         (when (< i n)
            (let ((bi (vector-ref b i)) (ci (vector-ref c i)))
               (let loop ((j 0))
                  (when (< j n)
                     (let ((bij (vector-ref bi j)) (cij (vector-ref ci j)))
                        (when (and (not (= i j)) (not (t= bij cij)))
                           (if (t< bij cij)
                              (begin
                                 (set! a-sub #f)
                                 (vector-set! ci j bij)
                                 (when (t< (t+ bij (dref c j i)) 0)
                                    (set! bail #t)))
                              (set! b-sub #f))))
                     (loop (+ j 1)))))
            (if bail
               #f
               (loop (+ i 1)))))
      (if bail
         #f
         (begin
            (set! *shape-subset*
               (if a-sub (if b-sub 'both 'a) (if b-sub 'b #f)))
            c)))) ; really want to run apsp if necessary here.
;
;   This function take the union of two shapes.  It returns #f if the
;   union is non-convex.  It sets the subset flag in the same way as
;   shape-intersect.  (As a matter of fact, the very first thing it does
;   is call shape-intersect.)
;
;   The way it works is the following.  If the shape has no intersection,
;   then the union must not be convex.  If one shape is a subset of the
;   other, or vice versa, then it is convex and it is simply the
;   larger shape.  Otherwise, we take the convex hull of the two shapes
;   (limited by our restrictions on planes) and try to show that the
;   this is indeed the union.
;
;   We form the convex hull by taking the max of each corresponding
;   constraint.  If a constraint is `extended' (overridden by a
;   larger one), we add it to the `sub' list for the corresponding
;   constraint.  All equal pairs are ignored here.
;
;   If we can find a point that is outside both shapes a and b, yet
;   inside the convex hull, then we know that the union shape is
;   concave.  So we try and find one constraint from the `a-sub'
;   list and one constraint from the `b-sub' list such that the
;   convex hull intersected with the converses of both of these
;   constraints is non-null.
;
;   The straightforward way to do this is to iterate through the
;   cross product a-sub x b-sub, introduce the two new constraints,
;   and run apsp on the result and see if it is void.  This,
;   unfortunately, has worst-case performance n^7 (two from a-sub,
;   two from b-sub, and three from apsp.)  In practice it is nowhere
;   near this bad, but still, we want to improve it if we can.  And
;   we can, all the way to n^4.
;
;   As it turns out, if we have a canonical shape, reducing a single
;   constraint a_{ij} will only make the shape void if
;
;      a_{ij} + a_{ji} < 0
;
;   So what we do is as follows.  We select the shorter of the two
;   lists and iterate over it for our outer loop.  First, we apply
;   the inverse of the corresponding constraint and call apsp.
;   We should never get a void result here.  Then we simply run
;   through the list of other constraints, checking if
;   a_{ij} + a_{ji} < 0, and failing if it is so.
;
(define *use-coalesce-approx?* #f)
;
(define (shape-union a b)
   (let ((int (shape-intersect a b)))
;     (log-info 'shape-intersect *shape-subset*)
      (if (or (not int) *shape-subset*)
         (if int
            (if (eq? *shape-subset* 'b) (dcopy a) (dcopy b))
            #f)
         (let ((c (dcopy a)) (a-sub 0) (b-sub 0) (n (vector-length a)))
            (let loop ((i 0))
               (when (< i n)
                  (let ((bi (vector-ref b i)) (ci (vector-ref c i)))
                     (let loop ((j 0))
                        (when (< j n)
                           (when (and (not (= i j))
                                      (not (t= (vector-ref bi j)
                                               (vector-ref ci j))))
                              (if (t< (vector-ref ci j) (vector-ref bi j))
                                 (begin
                                    (set! a-sub (+ 1 a-sub))
                                    (vector-set! ci j (vector-ref bi j)))
                                 (set! b-sub (+ 1 b-sub))))
                           (loop (+ j 1)))))
                  (loop (+ i 1))))
            (if *use-coalesce-approx?* c
               (call-with-current-continuation
                  (lambda (cont)
                     (if (< a-sub b-sub)
                        (shape-union-aux a b c n cont)
                        (shape-union-aux b a c n cont)))))))))
;
;   The idea here is that we run through the `a' matrix, looking for
;   constraints that are not necessarily met by `c'.  We make a new matrix
;   `d', initialize it to `c', and add the converse of the found constraint.
;   We then run APSP.  We can make APSP faster since we know that we have
;   only reduced a single value.  (We know we have reduced it because the
;   union of the two shapes was non-null, and thus c[j,i] + a[i,j] >= 0,
;   and thus c[j,i] >= -a[i,j].)
;
(define (shape-union-aux a b c n cont)
;  (log-info 'shape-union-aux n)
   (let ((d (vector-copy c)))
      (let forj ((j 0))
         (when (< j n)
            (let fori ((i 0))
               (vector-set! d j (vector-copy (vector-ref c j)))
               (when (< i n)
                  (when (and (not (= i j)) (t< (dref a i j) (dref c i j)))
                     (dset! d j i (- -1 (dref a i j)))
                     (let fork ((k 0))
                        (when (< k n)
                           (let ((t (t+ (dref d j i) (dref d i k))))
                              (when (t< t (dref d j k))
                                 (dset! d j k t)))
                           (fork (+ k 1))))
                     (let forii ((ii 0))
                        (when (< ii n)
                           (let forjj ((jj 0))
                              (when (< jj n)
                                 (let ((t (t+ (dref d ii j) (dref d j jj))))
                                    (when (and (not (= ii jj))
                                             (t< (dref b ii jj)
                                                (tmin t (dref d ii jj))))
                                       (cont #f)))
                                 (forjj (+ jj 1))))
                           (forii (+ ii 1)))))
                  (fori (+ i 1))))
            (forj (+ j 1))))
      c))
;
;   This routine does a change of variables; v_j' = v_j + d.  It is used
;   to place a bound on the shape space defined by constraints of the
;   form [constant,infinity].  If the minimum value of v_j is greater
;   than constant, this is precisely the same as though the minimum
;   value were that constant, so we can prune it down a bit.
;
(define (shape-translate v j d)
   (when *use-po-timing-now?*
      (po-trans-push (node-orbit (vector-ref *timed-token-ar* j)) d))
; (display-stuff "Translating v " v " at j " j " by d " d 'newline)
   (let ((n (vector-length v)))
      (let fori ((i 0))
         (when (< i n)
            (when (not (= i j))
               (dset! v i j (t+ (dref v i j) d))
               (dset! v j i (t+ (dref v j i) (- d))))
            (fori (+ i 1))))))
;
(define *destroy-shapes?* #t)
;
(define (shape-destroy v)
   (when (and *destroy-shapes?* (not *breadth-first?*))
      (let ((n (vector-length v)))
         (let fori ((i 0))
            (when (< i n)
               (vector-set! v i 'destroyed)
               (fori (+ i 1)))))))
;
(define (shape-destroyed? v)
   (eq? (vector-ref v 0) 'destroyed))
;
;   Here is our routine for taking the difference of two dbm's.  When
;   this becomes important, we will improve it (to use a n^2 instead of
;   a n^3 apsp) but for now it suffices.  Note that it returns a list of
;   dbms, rather than a single dbm.
;
;   Takes a-b, returns difference as a list of dbms.  The count will be
;   equal to or less than the number of faces of b that intersect a.
;   And, the resulting list will be disjoint---no discrete solution will
;   be a member of more than one list.  It potentially returns an
;   exponential number of dbms.
;
(define (dbm-difference a b)
   (let ((t (shape-intersect a b)))
      (if (or (eq? *shape-subset* 'a) (eq? *shape-subset* 'both)) '()
         (if (not t) (list a)
            (dbm-difference-hard a b)))))
;
(define (dbm-difference-hard a b)
   (let ((a (dcopy a)) (n (vector-length a)) (ess (dcopy b)) (r '()))
      (let fork ((k 0))
         (when (< k n)
            (let fori ((i 0))
               (when (< i n)
                  (when (not (= i k))
                     (let forj ((j 0))
                        (when (< j n)
                           (when (and (not (= i j)) (not (= j k))
                                    (= (+ (dref b i k) (dref b k j))
                                       (dref b i j)))
                              (dset! ess i j #f))
                           (forj (+ j 1)))))
                  (fori (+ i 1))))
            (fork (+ k 1))))
      (let fork ((k 0))
         (when (< k 2)
            (let fori ((i 0))
               (when (and a (< i n))
                  (let forj ((j 0))
                     (when (and a (< j n))
                        (when (and (not (= i j)) (or (= k 1) (dref ess i j))
                                 (> (dref a i j) (dref b i j)))
                           (let ((ap (dcopy a)))
                              (dset! ap j i (- (+ 1 (dref b i j))))
                              (apsp1 ap j i)
                              (set! r (cons ap r))
                              (dset! a i j (dref b i j))
                              (when (< (apsp1 a i j) 0)
                                 (set! a #f))))
                        (forj (+ j 1))))
                  (fori (+ i 1))))
            (fork (+ k 1))))
   r))
;
;   How does the backtracking search work in the context of these regions?
;   What we do is maintain, in addition to each state, a list of time
;   regions for which those states are valid.  Every time we find a new
;   set of regions is valid, we iterate through this list, trying to
;   combine it with the existing regions using state-union.  If we
;   cannot, we simply add the new region on to the list and continue.
;   If we can, then we iterate again using the new joined region over
;   the list trying to continue to merge.  If during this combination
;   procedure, we end up with a region that is a superset of the `new'
;   region, then we don't explore it further.  Otherwise we continue
;   from the reached point.
;
;   On backtracking, each time we backtrack we check to see if the
;   existing matrix is still a member of the state search list.  If it
;   is not, then some later state has subsumed it and backtracking is
;   not necessary.  (We only remove a matrix from the list if we can
;   merge it with another matrix.)
;
;   The problem we will try to avoid is that the number of regions on
;   the state list might end up large, and then our whole process will
;   slow down.  But since we are now working with large chunks of states
;   rather than individual states themselves, we should have a major
;   win even if this does happen.
;
;   One remaining issue is how we associate places with the indices of
;   the matrices in any particular state.  We might be able to use the
;   orbit information to assist this a bit, but even so there are many
;   more orbits than there will be marked timed places in a given state.
;   Actually, this isn't too bad.  We start with the orbits of the
;   marked places.  We sort this list, and then initialize two arrays
;   for the two mappings.  This is n^2 if we use a slow sort and n log n
;   if we use a fast, but either way the time is dominated by the region
;   operations.  We sort the tokens directly in the timed-token-ar, and
;   then `unsort' them as we fire a transition.
;
;   So to look things up, we need the following data structures:
;
;      token-locator:  orbit# -> node
;      timed-tokens:   count of timed tokens in current state
;      timed-token-ar: index -> node
;      timed-locator:  orbit# -> index (in timed-tokens)
;
(define *timed-tokens* 0)
(define *otimed-tokens* 0)
;
(define *timed-token-ar* '())
(define *otimed-token-ar* '())
;
(define *timed-locator* '())
(define *otimed-locator* '())
;
(define (notice-timed-token p)
   (set! *timed-tokens* (+ 1 *timed-tokens*))
   (vector-set! *timed-token-ar* *timed-tokens* p)
   (vector-set! *timed-locator* (node-orbit p) *timed-tokens*))
;
(define (forget-timed-token p)
   (let* ((n (vector-ref *timed-locator* (node-orbit p)))
          (dp (vector-ref *timed-token-ar* *timed-tokens*)))
      (vector-set! *otimed-token-ar*
         (vector-ref *otimed-locator* (node-orbit p)) '())
      (vector-set! *timed-token-ar* n dp)
      (vector-set! *timed-locator* (node-orbit dp) n)
      (set! *timed-tokens* (- *timed-tokens* 1))))
;
(define (swap-elements a i j)
   (let ((t (vector-ref a i)))
      (vector-set! a i (vector-ref a j))
      (vector-set! a j t)))
;
(define (sort-timed-tokens)
   (let ((a *timed-token-ar*) (n *timed-tokens*))
      (if (< n 4)
         (case n
            (0 '())
            (1 '())
            (2 (when (> (splace-sort-key (vector-ref a 1))
                      (splace-sort-key (vector-ref a 2)))
                  (swap-elements a 1 2)))
            (3 (when (> (splace-sort-key (vector-ref a 1))
                      (splace-sort-key (vector-ref a 2)))
                  (swap-elements a 1 2))
               (when (> (splace-sort-key (vector-ref a 2))
                      (splace-sort-key (vector-ref a 3)))
                  (swap-elements a 2 3)
                  (when (> (splace-sort-key (vector-ref a 1))
                         (splace-sort-key (vector-ref a 2)))
                     (swap-elements a 1 2)))))
         (let loop ((i 1) (sv 0))
            (if (<= i n)
               (loop (+ i 1) (cons (vector-ref *timed-token-ar* i) sv))
               (let lp2 ((i 1) (sv (sort sv (lambda (a b)
                                               (< (splace-sort-key a)
                                                  (splace-sort-key b))))))
                  (when (<= i n)
                     (vector-set! *timed-token-ar* i (car sv))
                     (lp2 (+ i 1) (cdr sv)))))))
      (let fori ((i 1))
         (when (<= i n)
            (vector-set! *timed-locator* (node-orbit (vector-ref a i)) i)
            (fori (+ i 1))))))
;
(define (mark-enabled-dirty t)
   (let* ((esg (strans-eps-connected t)))
      (when (not (strans-dirty? t))
         (strans-dirty?-set! t #t)
         (sgraph-consider-set! *sgraph* (cons t (sgraph-consider *sgraph*)))
         (when (and esg (not (eps-subgraph-dirty? esg)))
            (eps-subgraph-dirty?-set! (strans-eps-connected t) #t)
            (sgraph-consider-esg-set! *sgraph*
               (cons esg (sgraph-consider-esg *sgraph*)))))))
;
;   This function propogates a new data flow value down the graph.
;   The main subtlety is handling values from wires.  We don't
;   propogate past registers or into places or into transitions.
;   If the value does go to a transition, though, we mark it as
;   dirty.
;
(define (new-dfs-value n)
   (case (node-type n)
      (F
         (let* ((newval
            (let ((args (map (lambda (a) (sdf-curval a))
                           (sdf-args n))))
               (if (memq 'unknown args) 'unknown
                  (apply (sdf-op n) args))))
                (oldval (sdf-curval n)))
            (if (eqv? newval oldval)
               '() ; do nothing the value is the same
               (begin
                  (sdf-curval-set! n newval)
                  (for-each (lambda (a) (new-dfs-value (edge-outnode a)))
                              (node-outedges n))))))
      (J (for-each (lambda (a) (new-dfs-value (edge-outnode a)))
                     (node-outedges n)))
      (T (mark-enabled-dirty n))))
;
;   This function calculates all of the joins that are upwardly
;   reachable from a particular data flow node.
;
;   At the moment we do not handle places correctly.
;
(define (dataflow-check-joins-of n)
   (if (eq? (sdf-joins n) 'uncalc)
      (let ((r (if (node-type? n 'J) (list n)
                  (let loop ((el (node-inedges n)) (res '()))
		     (if el
			(loop (cdr el)
                           (if (edge-type? (car el) 'D)
                              (union res (dataflow-check-joins-of
					    (edge-innode (car el))))
			      res))
			res)))))
         (sdf-joins-set! n r)
	 r)
      (sdf-joins n)))
;
;   This function finds all of the constants and propogates their
;   values down the functional tree.  Constants are defined as all
;   functional units with no inputs.
;
;   Then, it does all the rest of the work necessary for setting up
;   the data flow.  Specifically, it calculates the joins for each
;   data flow node.  Then, it calculates the strans-df-rw for all
;   transitions  This is '() or a cons cell; if it is a cons cell, the
;   car contains a list of joins that are read and the cdr contains
;   a list of joins that are written.  These lists are disjoint.
;
(define (prop-constants)
   (let loop ((nl (sgraph-dfs *sgraph*)))
      (when nl
         (let ((n (car nl)))
	    (sdf-joins-set! n 'uncalc)
            (when (null? (node-inedges n))
               (new-dfs-value n)))
         (loop (cdr nl))))
   (let loop ((nl (sgraph-dfs *sgraph*)))
      (when nl
         (let ((n (car nl)))
            (dataflow-check-joins-of (car nl)))
	 (loop (cdr nl))))
   (let loop ((nl (sgraph-transitions *sgraph*)))
      (when nl
         (let ((t (car nl)) (r '()) (w '()))
	    (for-each (lambda (e)
		         (when (edge-type? e 'D)
			    (set! r (union r (sdf-joins (edge-innode e))))))
	       (node-inedges t))
	    (for-each (lambda (e)
		         (when (node-type? (edge-outnode e) 'R)
			    (set! w (union w (list (edge-outnode
						      (car (node-outedges
							      (edge-outnode
							         e)))))))))
	       (node-outedges t))
	    (set! r (set-difference r w))
	    (strans-df-rw-set! t
	       (if (or r w) (cons r w) '())))
	 (loop (cdr nl)))))
;
;   This function calculates, for each transition connected to any
;   epsilon place, all transitions connected through epsilon places
;   (directly or indirectly) to that transition.  It also calculates
;   all places connected in such a way.
;
;   We do this through recursive search.
;
;   Note that the only thing connected to an epsilon place is
;   (can be) transitions.
;
;   During this search, we may encounter an epsilon place again,
;   but when we encounter the appropriate transition, that will
;   cause the search to end.
;
(define (eps-search-place esg p places trans)
   (if (eq? (splace-eps-connected p) esg)
      (cons places trans)
      (begin
         (splace-eps-connected-set! p esg)
         (let loop ((places (cons p places)) (trans trans)
                    (tl (append
                       (map edge-innode-lambda (node-inedges p))
                       (map edge-outnode-lambda (node-outedges p)))))
             (if tl
                (let ((more (eps-search-trans esg (car tl) places trans)))
                   (loop (car more) (cdr more) (cdr tl)))
                (cons places trans))))))
;
;   This next does the same for transitions.
;
(define (eps-search-trans esg t places trans)
   (if (eq? (strans-eps-connected t) esg)
      (cons places trans)
      (begin
         (strans-eps-connected-set! t esg)
         (let loop ((places places) (trans (cons t trans))
                    (pl (append
                       (map edge-innode-lambda (node-inedges t))
                       (map edge-outnode-lambda (node-outedges t)))))
            (if pl
               (if (and (node-type? (car pl) 'P)
                        (splace-eps? (car pl)))
                  (let ((more (eps-search-place esg (car pl) places trans)))
                     (loop (car more) (cdr more) (cdr pl)))
                  (loop places trans (cdr pl)))
               (cons places trans))))))
;
;   This function creates a list.  Each element of the list is a
;   5-tuple (as a list) consisting of:
;
;      input transitions for the corresponding place that are seen
;      output transitions for the corresponding place that are seen
;      input transitions that haven't been seen
;      output transitions that haven't been seen
;      the place in question
;
(define (make-accel ps seen)
   (if ps
      (let* ((ins (map edge-innode-lambda (node-inedges (car ps))))
             (outs (map edge-outnode-lambda (node-outedges (car ps))))
             (oldin (remove-if (lambda (a)
                                 (not (eq? (strans-magik a) seen))) ins))
             (oldout (remove-if (lambda (a)
                                  (not (eq? (strans-magik a) seen))) outs))
             (newin (remove-if (lambda (a)
                                 (eq? (strans-magik a) seen)) ins))
             (newout (remove-if (lambda (a)
                                  (eq? (strans-magik a) seen)) outs)))
         (for-each (lambda (a) (strans-magik-set! a seen)) ins)
         (for-each (lambda (a) (strans-magik-set! a seen)) outs)
         (cons (list oldin oldout newin newout (car ps))
            (make-accel (cdr ps) seen)))
      '()))
;
;   This function builds a list of normal places that are inputs to the
;   transitions in this epsilon group.  For each place that is an
;   input to more than one transition, all of those transitions are
;   marked as `in conflict'.
;
;   We don't have to do this for outputs because the nets are safe by
;   construction.
;
(define (calc-eps-conflicts esg)
   (let loop ((tl (eps-subgraph-transitions esg)))
      (when tl
         (let loop2 ((el (node-inedges (car tl))))
            (when el
               (let* ((t (car tl))
                      (n (edge-innode (car el))))
                  (when (and (node-type? n 'P) (not (splace-eps? n)))
                     (let loop3 ((el (node-outedges n)) (ts '()))
                        (if el
                           (let ((n2 (edge-outnode (car el))))
                              (loop3 (cdr el)
                                 (if (and (not (eq? n2 t))
                                        (node-type? n2 'T)
                                        (eq? esg (strans-eps-connected n2)))
                                    (cons n2 ts) ts)))
                           (when (> (length ts) 0)
                              (strans-eps-conflicts-set! t
                                 (union ts (strans-eps-conflicts t))))))))
               (loop2 (cdr el))))
         (loop (cdr tl)))))
;
(define (put-immediates-first pl tail)
   (if pl
      (if (splace-immediate? (car pl))
         (cons (car pl) (put-immediates-first (cdr pl) tail))
         (put-immediates-first (cdr pl) (cons (car pl) tail)))
      tail))
;
;   Finally, our eps-connected calculator.
;
(define (calc-eps-connected)
   (let loop ((tl (sgraph-transitions *sgraph*)))
      (when tl
         (when (not (strans-eps-connected (car tl)))
            (let* ((esg (make-eps-subgraph))
                   (r (eps-search-trans esg (car tl) '() '()))
                   (ps (car r)) (ts (cdr r)))
               (if ps
                   (let ((nps (put-immediates-first ps '())))
                      (sgraph-esgs-set! *sgraph*
                         (cons esg (sgraph-esgs *sgraph*)))
                      (eps-subgraph-transitions-set! esg ts)
                      (eps-subgraph-places-set! esg ps)
                      (eps-subgraph-dirty?-set! esg #t)
                      (eps-subgraph-accel-set! esg
                         (make-accel nps (cons '() '())))
                      (calc-eps-conflicts esg))
                   (strans-eps-connected-set! (car tl) #f))))
         (loop (cdr tl)))))
;
;   This function calculates some information about the orbits.
;   Specifically, it calculates how many bits are required for each,
;   just taking into account the number of places and whether the
;   orbit is marked or not.
;
(define (calc-orbit-info)
   (let* ((n (graph-maxorbits *graph*))
          (tokens (make-vector n 0))
          (preserved? (make-vector n #t))
          (count (make-vector n 0))
          (tok-loc (make-vector n '())))
      (set! *timed-tokens* 0)
      (set! *timed-token-ar* (make-vector (+ n 1) '()))
      (set! *timed-locator* (make-vector n '()))
      (set! *otimed-locator* (make-vector n '()))
      (set! *otimed-token-ar* (make-vector (+ n 1) '()))
      (set! *otimed-tokens* 0)
      (sgraph-token-locator-set! *sgraph* tok-loc)
      (for-each
         (lambda (t)
            (let ((in (remove-if not
                         (map (lambda (a) (node-orbit (edge-innode a)))
                            (node-inedges t))))
                  (out (remove-if not
                          (map (lambda (a) (node-orbit (edge-outnode a)))
                             (node-outedges t)))))
               (strans-orbits-set! t (sort (union in out) <))
               (for-each (lambda (a)
                            (when (not (and (member a in) (member a out)))
                               (vector-set! preserved? a #f)))
                  (strans-orbits t))))
         (sgraph-transitions *sgraph*))
      (for-each
         (lambda (p)
            (let ((i (node-orbit p)))
               (when i
                  (splace-orbit-seq-set! p (vector-ref count i))
                  (vector-set! count i (+ 1 (vector-ref count i)))
                  (when (node-marked p)
                     (vector-set! tokens i (+ 1 (vector-ref tokens i)))))))
         (sgraph-places *sgraph*))
      (DBG 'o
         (let loop ((i 0))
            (when (< i (graph-maxorbits *graph*))
               (display-stuff "Orbit " i " has " (vector-ref tokens i)
                  " tokens and " (vector-ref count i) " places and is "
                  (if (vector-ref preserved? i) "" "not ") "preserved."
                  'newline)
               (loop (+ i 1)))))
      (for-each (lambda (toks pres k)
                   (when (or (> toks 1) (and (= toks 0) pres (> k 0)))
                      (merror "Some token loop is messed up" toks)))
         (vector->list tokens) (vector->list preserved?) (vector->list count))
      (sgraph-orbit-info-set! *sgraph* (cons preserved? count))))
;
;   This function calculates a list of lists of places for each place;
;   list is dot of p dot.  This is used to speed up stubborn set
;   calculations.  (Actually, we stick the relevant transition at the
;   front of each of the sublists so we can check if it is enabled or
;   not first.)  Any transitions that only have this place as their
;   input are not added to the list.
;
(define (calc-dot-of-p-dot p)
   (let ((lst '()))
      (let loop ((el (node-outedges p)))
         (when el
            (let* ((t (edge-outnode (car el))))
               (when (node-type? t 'T)
                  (let loop2 ((el (node-inedges t)) (sofar (list t)))
                     (if el
                        (let ((pp (edge-innode (car el))))
                           (if (and (node-type? pp 'P) (not (eq? pp p)))
                              (loop2 (cdr el) (cons pp sofar))
                              (loop2 (cdr el) sofar)))
                        (when (> (length sofar) 1)
                           (set! lst (cons (reverse sofar) lst)))))))
            (loop (cdr el))))
      (splace-dot-of-p-dot-set! p lst)
      (when (splace-immediate? p)
         (let loop ((toadd '()) (lst lst))
            (if lst
               (loop (union toadd (map node-orbit-lambda (cdar lst)))
                        (cdr lst))
               (let loop ((el (node-inedges p)))
                  (when el
                     (let* ((t (edge-innode (car el))))
                        (when (node-type? t 'T)
                           (strans-orbits-set! t
                              (union toadd (strans-orbits t)))))
                     (loop (cdr el)))))))))
;
;   This function calculates whether a transition is weakly enabled.
;   This means, enabled except for epsilon places.
;
;   The enabling is as follows:
;
;      Input places must be marked.
;      Input data flow nodes must have a non-zero or non-false value
;         if the input is from arity 0, else it must have a 0 or false
;         value if the input is from arity 1.
;
;   And that's all that can be connected to transitions!
;
;   *BUG*:  we don't deal with places with a data flow node that goes
;   directly to a transition.
;
(define (weak-enabling t)
   (let ((r (let loop ((tl (node-inedges t)))
               (if tl
                  (let* ((n (edge-innode (car tl))))
                     (case (node-type n)
                        (P (if (splace-marked? n)
                              (loop (cdr tl))
                              #f))
                        ((J F R)
                           (let ((v (sdf-curval n)))
                              (if (eq? v 'unknown) #f
                                 (if (= (edge-inpos (car tl))
                                       (if (and (number? v)
                                              (not (= v 0))) 0 1))
                                    (loop (cdr tl))
                                    #f))))))
                  #t))))
      (when (not (eq? (strans-enabled? t) r))
         (let ((esg (strans-eps-connected t)))
            (strans-enabled?-set! t r)
            (when (and esg *use-esg-hash?*)
               (eps-subgraph-state-set! esg
                  ((if r + -) (eps-subgraph-state esg)
                     (strans-esg-bit t))))))
      (strans-dirty?-set! t #f)))
;
;   This function gets the `old' value from a register, which is
;   really the value at the join it points to.  Assumption:  each
;   register has only one output edge; true by construction.
;
(define (reg-old-value r)
   (sdf-curval (edge-outnode (car (node-outedges r)))))
;
;   This is where registers get fired.  The input value is also passed on
;   to any following nodes that are joins (as indicated by the sdf
;   structure.)
;
(define (fire-register r)
   (let* ((oldval (reg-old-value r))
          (newval ((sdf-op r))))
      (DBG 'v
         (display-stuff "/" oldval "/" newval "/"))
      (when (not (eqv? oldval newval))
         (new-reg-val r oldval newval))))
;
(define *new-reg-vals* '())
;
(define (new-reg-val r oldval newval)
;  (display-stuff "Setting new value of " newval " (from " oldval ")" 'newline)
   (sdf-curval-set! r newval)
   (let loop ((tl (node-outedges r)))
      (when tl
         (let ((n (edge-outnode (car tl))))
            (set! *new-reg-vals* (cons (cons n newval) *new-reg-vals*))
            (loop (cdr tl))))))
;
(define (flush-regs)
   (for-each (lambda (c)
                (let ((n (car c)) (v (cdr c)))
                   (update-df-state n v)
                   (new-dfs-value n)))
      *new-reg-vals*)
   (set! *new-reg-vals* '()))
;
;   This is where we try to mark a transition as disabled.  If it is
;   connected to epsilon places, we instead mark it as dirty because,
;   though it is definitely disabled, we want to make really sure that
;   the epsilon group is correctly calculated.
;
(define (mark-disable n)
   (if (strans-eps-connected n)
      (mark-enabled-dirty n)
      (strans-enabled?-set! n #f)))
;
;   This is where we add a token to a place.  If the timing is
;   *iforever*, we use *iforever* as the marking.  We also update
;   the locator arrays.  Currently no difference between `mark' and
;   `remark', but kept separate for historical reasons.
;
(define (mark-place p)
   (let ((tl (node-outedges p)))
      (when (not (splace-eps? p))
;  (display-stuff "(+" (log2 (splace-state-value p)) ")")
         (when (splace-marked? p)
            (merror "Place is already marked!" 0))
         (vector-set! (sgraph-token-locator *sgraph*) (node-orbit p) p)
         (splace-marked?-set! p #t)
         (when (not (splace-triv? p))
            (notice-timed-token p))
         (DBG 'f
            (if (null? tl) (display-stuff "*DEAD*")))
         (let loop ((tl tl))
            (if tl
               (let ((n (edge-outnode (car tl))))
                  (case (node-type n)
                     (T (mark-enabled-dirty n)))
                  (loop (cdr tl))))))))
;
(define (remark-place p)
   (let ((tl (node-outedges p)))
      (when (not (splace-eps? p))
;  (display-stuff "(++" (log2 (splace-state-value p)) ")")
         (when (splace-marked? p)
            (merror "Place is already marked!" 0))
         (vector-set! (sgraph-token-locator *sgraph*) (node-orbit p) p)
         (splace-marked?-set! p #t)
         (when (not (splace-triv? p))
            (notice-timed-token p))
         (let loop ((tl tl))
            (if tl
               (let ((n (edge-outnode (car tl))))
                  (case (node-type n)
                     (T (mark-enabled-dirty n)))
                  (loop (cdr tl))))))))
;
(define (sim-marked? n)
   (and (splace-marked? n) (not (splace-eps? n))))
;
;   This is where we remove a token from a place.
;
(define (unmark-place p)
   (when (not (splace-eps? p))
;  (display-stuff "(-" (log2 (splace-state-value p)) ")")
      (when (not (splace-marked? p))
         (merror "Place is already not marked!" 0))
      (when (not (splace-triv? p))
         (forget-timed-token p))
      (splace-marked?-set! p #f)
      (let loop ((tl (node-outedges p)))
         (if tl
            (let ((n (edge-outnode (car tl))))
               (case (node-type n)
                  (T (mark-disable n)))
               (loop (cdr tl)))))))
;
;
;   This routine fires a transition.  It does the following:
;
;      Removes all tokens from input places.  For each of these
;      places, it marks all output transitions as disabled (strongly).
;
;      It places new tokens in all output places.  For each of these
;      places, it marks all output transitions as dirty.
;
;      If a register is connected to the transition, it propogates new
;      data flow values down.
;
;      It then marks the current transition as dirty.
;
;   It also puts together a collection of the previous contents of
;   values and returns that.  This then is used to unfire the node.
;
;   For simulation, the constructed log is dropped on the floor.  For
;   verification, it is recorded and retained.
;
;   As the log, we save the transition, the time, and any changed values
;   or changed non-trivial markings.
;
;   We also keep track of the times that the place has been fired, so that
;   we can check for covering.  (This only for behavioral constraint places.)
;
(define *firing-failed?* #f)
;
;   We also need to make sure that the time lies within the range
;   specified by the constraint (if any).
;
(define (prefire t)
   (strans-fired?-set! t #t)
   (let loop ((tl (node-inedges t)))
      (when tl
         (let* ((n (edge-innode (car tl))))
            (case (node-type n)
               (P (unmark-place n)))
            (loop (cdr tl))))))
;
(define (postfire t)
   (let ((log '()))
      (let loop ((tl (node-outedges t)))
         (when tl
            (let ((n (edge-outnode (car tl))))
               (case (node-type n)
                  (P (mark-place n))
                  (R (set! log (cons (reg-old-value n) log))
                     (fire-register n)))
               (loop (cdr tl)))))
      (if *big-states?*
         (cons t (cons *current-region* log))
         (cons t log))))
;
;   This function `unfires' a node given the log of the node firing.
;   If a place that is being unmarked contains a `seen' firing, but
;   the set of seen firings does not cover the enabling of the
;   transition, then we set the failed firing flag.
;
(define (unfire log)
   (let ((t '()))
      (set! t (if (vector? log) log (car log)))
      (when *big-states?*
         (set! log (cdr log)))
      (DBG 'f
         (display-stuff (fire-label t)))
      (let loop ((tl (reverse (node-outedges t))))
         (when tl
            (let ((n (edge-outnode (car tl))))
               (case (node-type n)
                  (P 
;   (DBG 'w
;      (when (splace-firings-seen n)
;        (display-stuff "Saw " (splace-firings-seen (node-snode n)) 'newline)))
                      (unmark-place n))
                  (R (new-reg-val n #f (cadr log))
;  (display-stuff "Unfiring to " (cadr log) 'newline)
                     (set! log (cdr log))))
               (loop (cdr tl)))))
      (let loop ((tl (reverse (node-inedges t))))
         (when tl
            (let ((n (edge-innode (car tl))))
               (case (node-type n)
                  (P (remark-place n)))
               (loop (cdr tl)))))))
;
(define (unfire-complex log)
   (set! *firing-failed?* #f)
   (set! *num-unfirings* (+ *num-unfirings* 1))
   (DBG 'f
      (display-stuff "<"))
   (if (or *timing-important?* *dataflow-important?*)
      (begin
         (if (pair? (car log))
            (for-each unfire (reverse log))
            (unfire log))
         (when *new-reg-vals*
            (flush-regs))
         (when *timing-important?*
            (set! *current-region*
               (if (pair? (car log)) (cadar log) (cadr log)))
            (sort-timed-tokens)
            (when *use-po-timing-now?*
               (if (pair? (car log))
                  (undo-po-work (cons '() (map car log)))
                  (undo-po-work (car log))))))
      (if (pair? log)
         (for-each unfire (cdr log))
         (unfire log))))
;
(define (plus-minus lab)
   (append (map (lambda (a) (cons-symbol a '+)) lab)
           (map (lambda (a) (cons-symbol a '-)) lab)))
;
(define *use-esg-hash?* #t)
;
(define (find-orbital-seq-from t o)
   (let ((seq 0))
      (for-each (lambda (e)
                   (let ((n (edge-innode e)))
                      (when (and (node-type? n 'P)
                               (eqv? (node-orbit n) o))
                         (set! seq (splace-orbit-seq n)))))
         (node-inedges t))
      seq))
;
(define (assign-esg-trans esg s)
   (let* ((n (graph-maxorbits *graph*))
          (v (make-vector (+ n 1) '()))
          (osize (cdr (sgraph-orbit-info *sgraph*))))
      (for-each (lambda (t)
                   (let loop ((os (strans-orbits t)) (m 1) (mo n))
                      (if os
                         (if (>= (vector-ref osize (car os)) m)
                            (loop (cdr os) (vector-ref osize (car os))
                               (car os))
                            (loop (cdr os) m mo))
                         (begin
                            (when (not (vector-ref v mo))
                               (vector-set! v mo (make-vector m '())))
                            (let* ((oseq (find-orbital-seq-from t mo)))
                               (dset! v mo oseq
                                  (cons t (dref v mo oseq))))))))
         (eps-subgraph-transitions esg))
      (let loop ((o 0))
         (when (<= o n)
            (when (vector-ref v o)
               (let ((tl '()) (tc 0))
                  (for-each (lambda (v)
                               (when v
                                  (set! tc (+ tc (length v)))
                                  (set! tl (cons (cons (length v) v) tl))))
                     (vector->list (vector-ref v o)))
                  (set! tl (sort tl (lambda (a b) (> (car a) (car b)))))
                  (if (< tc (* 2 (caar tl)))
                     (let loop2 ((tlist (apply append (map cdr tl))) (b s))
                        (if tlist
                           (begin
                              (strans-esg-bit-set! (car tlist) b)
                              (loop2 (cdr tlist) (* b 2)))
                           (set! s b)))
                     (let ((sb (* s (pow2 (caar tl)))) (ta 1))
                        (let loop2 ((tl tl) (i 0))
                           (when tl
                              (let ((ta2 (log2-ceil (* i (caar tl)))))
                                 (when (> ta2 ta)
                                    (set! ta ta2)))
                              (let loop ((tl (cdar tl)) (b s))
                                 (when tl
                                    (strans-esg-bit-set! (car tl)
                                       (+ b (* i sb)))
                                    (loop (cdr tl) (* b 2))))
                              (loop2 (cdr tl) (+ i 1))))
                        (set! s (* sb (pow2 ta)))))))
            (loop (+ o 1))))))
;
(define (init-esg-hash)
   (when (or *use-esg-hash?* *use-po-pruning-now?*)
      (let loop ((i 1) (esgs (sgraph-esgs *sgraph*)))
         (if esgs
            (let ((esg (car esgs)))
               (eps-subgraph-serial-set! esg i)
               (loop (+ i 1) (cdr esgs)))
            (let loop1 ((s (pow2 (log2 (+ 1 (* i 2)))))
                        (esgs (sgraph-esgs *sgraph*)))
               (let loop2 ((tl (sgraph-transitions *sgraph*)) (v s))
                  (when tl
                     (if (strans-eps-connected (car tl))
                        (loop2 (cdr tl) v)
                        (begin
                           (strans-esg-bit-set! (car tl) v)
                           (loop2 (cdr tl) (+ v s))))))
               (when esgs
                  (assign-esg-trans (car esgs) s)
                  (loop1 s (cdr esgs))))))
      (sgraph-esg-hash-set! *sgraph* (new-hash car))))
;
;   A non-recursive map.
;
(define (nrmap f l)
   (define (nrmapint l1 l2)
      (if l1
         (nrmapint (cdr l1) (cons (f (car l1)) l2))
	 l2))
   (reverse! (nrmapint l '())))
;
;   This function builds up the simulation graph.
;
(define (build-sgraph g inputs outputs)
   (logit "Building sgraph")
   (set! *dataflow-important?* #f)
   (when (intersection inputs outputs)
      (merror "Signals can only be inputs or outputs, not both"
         (intersection inputs outputs)))
   (let* ((type-assoc (list-to-types (graph-nodes g) '()))
          (places (safe-cdr (assoc 'P type-assoc)))
          (transitions (safe-cdr (assoc 'T type-assoc)))
          (registers (safe-cdr (assoc 'R type-assoc)))
          (joins (safe-cdr (assoc 'J type-assoc)))
          (funcs (safe-cdr (assoc 'F type-assoc)))
          (ng (make-sgraph))
          (dead-trans (new-node-unadded 'T)))
;
;   Check to see that the numbers add up.
;
      (set! *sgraph* ng)
      (set! *closure-sizes* '())
      (sgraph-inputs-set! ng (plus-minus inputs))
      (sgraph-outputs-set! ng (plus-minus outputs))
      (when (not (= (length (graph-nodes g))
                    (+ (length places) (length transitions)
                       (length registers) (length joins) (length funcs))))
         (merror "Some funky types in the graph" (map car type-assoc)))
      (set! places (nrmap place-to-splace places))
      (set! transitions (nrmap trans-to-strans transitions))
      (let ((sn (trans-to-strans dead-trans)))
         (strans-dead?-set! sn #t)
         (strans-esg-bit-set! sn 0)
         (strans-state-value-set! sn 1))
      (set! registers (map reg-to-sreg registers))
      (set! joins (map join-to-sjoin joins))
      (set! funcs (map func-to-sfunc funcs))
      (for-each (lambda (nn) (let ((n (node-snode nn)))
                                (node-outedges-set! n '())
                                (node-inedges-set! n '())))
         (graph-nodes g))
      (ge-for-each (lambda (e) (let* ((ne (vector-copy e))
                                   (in (edge-innode e)) (out (edge-outnode e))
                                   (sin (node-snode in))
                                   (sout (node-snode out)))
                               (edge-innode-set! ne sin)
                               (edge-outnode-set! ne sout)
                               (node-inedges-set! sout
                                  (cons ne (node-inedges sout)))
                               (node-outedges-set! sin
                                  (cons ne (node-outedges sin))))))
      (sgraph-log-set! ng '())
      (sgraph-graph-set! ng g)
      (sgraph-places-set! ng places)
      (sgraph-eps-places-set! ng
         (remove-if (lambda (a) (not (splace-eps? a))) places))
      (sgraph-transitions-set! ng transitions)
      (sgraph-dead-transition-set! ng (node-snode dead-trans))
      (sgraph-dfs-set! ng (append funcs registers joins))
      (another-df-pass (sgraph-dfs ng))
;     (logit "Propogating constants")
      (prop-constants)
;     (logit "Calculating eps-connected")
      (calc-eps-connected)
      (calc-orbit-info)
      (when *expand-eps-places*
         (for-each calc-dot-of-p-dot places))
      (sgraph-consider-set! *sgraph* transitions)
      (sgraph-consider-esg-set! *sgraph* (sgraph-esgs *sgraph*))
      (display-stuff " " (length transitions) " tr " (length places) "("
         (length (sgraph-eps-places ng)) ") pl " (length (sgraph-dfs ng))
         " df " (length (sgraph-esgs ng))
         " esg " (graph-maxorbits *graph*) " orbits")
      (init-esg-hash)
;     (logit "Setting verification variables")
;
;   This last two steps should be idempotent so we can repeat them for
;   timing iteration.
;
      (handle-beh-con-places)
      (when (null? (sgraph-tb-assoc *sgraph*))
         (set-verification-variables 1 #f))))
;
;   This is our second attempt at handling the epsilon places.  The
;   first worked, but wasn't as efficient as we would have liked.
;   (It wasn't that bad, but it was large and complex and hairy and
;   we felt that we could do better.)
;
;   We also use a list created during create-eps-connected that helps
;   us avoid a lot of garbage collection during simulation.  There is
;   still some garbage, however.
;
(define (esg-enabling esg)
; (display-stuff "Recalculating esg: "
;    (map (lambda (a) (list (length (car a)) (length (cadr a))
;                           (length (caddr a)) (length (car (cdddr a)))))
;       (eps-subgraph-accel esg)) 'newline)
   (let* ((s (eps-subgraph-state esg))
          (v (and *use-esg-hash?*
                (iset-member? (sgraph-esg-hash *sgraph*) s))))
      (if (not v)
         (begin
            (eps-subgraph-enabled?-set! esg '())
            (do-one-eps-place esg '() (eps-subgraph-accel esg))
            (eps-subgraph-enabled?-set! esg
               (remove 'dummy (eps-subgraph-enabled? esg)))
            (when (and *use-esg-hash?*
                     (not (iset-member? (sgraph-esg-hash *sgraph*) s)))
               (iset-insert (sgraph-esg-hash *sgraph*)
                  (cons s (eps-subgraph-enabled? esg)))))
         (eps-subgraph-enabled?-set! esg (cdr v)))
      (eps-subgraph-dirty?-set! esg #f)))
;
(define (count-in-set tl set n)
   (if tl
      (count-in-set (cdr tl) set
         (if (memq (car tl) set) (+ n 1) n))
      n))
;
;   This function is much easier than it looks.  When we consider the
;   current place, we have a list of transitions that have already been
;   considered that are inputs to this place, a list of transitions that
;   have already been considered that are outputs from this place, a list
;   of transitions that have not been considered that are inputs, and a
;   list of transitions that have been considered that are outputs.
;
;   So we look at the ones that we've already decided to enable, and we
;   count how many for inputs and outputs.  If either are greater than
;   1, we punt.  Otherwise, we split into three cases.
;
;   If no input or output transitions have been enabled for this transition,
;   first, we try to continue with no additional transitions, as though
;   this place were not to `fire'.  Then, we `doubly-loop' over all
;   remaining input and output transitions, looking for all pairs that are
;   enabled, add then to the `on' set, and recurse.
;
;   If exactly one input (or output) transition is enabled, and none of
;   the output (input) transitions are, then we work our way through
;   the list of remaining output (input) transitions, looking for one to
;   enable.
;
;   Finally, if exactly one input and exactly one output transition is
;   already enabled, we just recurse.
;
;   When we come to the end of the place list, we take what we have, and
;   if there are any transitions in it, we add it to the list of
;   enabled virtual transitions in the eps-subgraph structure.
;
;   This code is made just slightly more illegible by the fact that the
;   enabled transitions must not be in conflict.
;
;   We need a tiny function to check for conflicts.
;
;   Finally, we handle immediate epsilon places here as well.  If an
;   immediate place is encountered, we the following.
;
;   First, we handle the case where the places is not included at all.
;
;   Then, we save the current list of transitions on the eps group.
;
;   Then, we try it as a normal eps place.
;
;   If no new transitions are added, then we try it as a receptiveness
;   failure place, only looking for one input.
;
(define (no-conflict n on)
   (null? (intersection (strans-eps-conflicts n) on)))
;
;   This superset function is made much easier by the fact that
;   transitions are always considered in precisely the same order.
;
;   If we ignore it because it is a superset, we still stick a dummy
;   `transition' onto the list so that receptiveness works correctly.
;
(define (is-superset? al bl)
   (if (null? bl) #t
      (if (null? al) #f
         (if (eq? (car al) (car bl))
            (is-superset? (cdr al) (cdr bl))
            (is-superset? (cdr al) bl)))))
;
(define (add-dummy-to esg)
   (eps-subgraph-enabled?-set! esg (cons 'dummy (eps-subgraph-enabled? esg))))
;
(define (add-complex-delta t)
   (set-car! (cdddr (car t))
      (apply + (map strans-state-value-lambda (cdr t))))
   t)
;
(define (add-if-not-superset t esg)
   (let ((ts t) (others (eps-subgraph-enabled? esg)))
      (let loop ((tl others))
         (if tl
            (if (and (not (eq? 'dummy (car tl))) (is-superset? ts (cdar tl)))
               (add-dummy-to esg)
               (loop (cdr tl)))
            (if *use-esg-hash?*
               (let* ((s (complex-esg-bit (cons '() t)))
                      (m? (iset-member? (sgraph-esg-hash *sgraph*) s)))
                  (if m?
                     (eps-subgraph-enabled?-set! esg
                        (cons (cadr m?) (eps-subgraph-enabled? esg)))
                     (let ((t (add-complex-delta
                                 (cons (gather-timed-places #f '() '() '() t)
                                    t))))
                        (iset-insert (sgraph-esg-hash *sgraph*)
                           (cons s (list t)))
                        (eps-subgraph-enabled?-set! esg
                           (cons t (eps-subgraph-enabled? esg))))))
               (eps-subgraph-enabled?-set! esg
                  (cons (add-complex-delta
                           (cons (gather-timed-places #f '() '() '() t) t))
                     (eps-subgraph-enabled? esg))))))))
;
(define dummy-list (list 'dummy))
;
(define (add-dead-to-esg on)
   (if (memq (sgraph-dead-transition *sgraph*) on)
      on
      (append on (list (sgraph-dead-transition *sgraph*)))))
;
(define (gather-timed-places infinity? b bi c on)
   (if on
      (let* ((n (car on))
             (v (strans-timed-places n))
             (nb (vector-ref v 0))
             (nbi (vector-ref v 1))
             (nc (vector-ref v 2)))
         (gather-timed-places (or infinity? (strans-much-later? n))
            (if b (if nb (merror "Oops; too many beh places" 0) b) nb)
            (if nb
                (if bi (append (cdr nbi) bi) (cdr nbi))
                (if bi (append nbi bi) nbi))
            (if c (if nc (append nc c) c) nc)
            (cdr on)))
      (list infinity? (vector b (if b (cons (car b) bi) bi) c) '() '())))
;
(define (do-one-eps-place esg on accel)
   (if accel
      (let* ((inold (caar accel))
             (outold (cadar accel))
             (innew (car (cddar accel)))
             (outnew (cadr (cddar accel)))
             (insofar (count-in-set inold on 0))
             (outsofar (count-in-set outold on 0))
             (p (caddr (cddar accel))))
         (cond
            ((and (= insofar 0) (= outsofar 0))
               (do-one-eps-place esg on (cdr accel))
               (let ((state (eps-subgraph-enabled? esg)))
                  (let loop1 ((in innew))
                     (when in
                        (let* ((n (car in))
                               (int2 (strans-enabled? n)))
                           (when int2
                              (if (no-conflict n on)
                                 (let loop2 ((out outnew))
                                    (when out
                                       (let ((n2 (car out)))
                                          (when (strans-enabled? n2)
                                             (if (no-conflict n2 on)
                                                (do-one-eps-place esg
                                                   (cons n (cons n2 on))
                                                   (cdr accel))
                                                (add-dummy-to esg))))
                                      (loop2 (cdr out))))
                                 (add-dummy-to esg))))
                        (loop1 (cdr in))))
                  (when (and (splace-immediate? p)
                             (eq? state (eps-subgraph-enabled? esg)))
                     (let loop1 ((in innew))
                        (when in
                           (let* ((n (car in))
                                  (non (add-dead-to-esg on))
                                  (int2 (strans-enabled? n)))
                              (when (and int2 (no-conflict n on))
                                 (do-one-eps-place esg (cons n non)
                                    (cdr accel))))
                           (loop1 (cdr in)))))))
            ((and (= insofar 1) (= outsofar 0))
               (let ((state (eps-subgraph-enabled? esg)))
                  (let loop2 ((out outnew))
                     (when out
                        (let* ((n2 (car out))
                               (int3 (strans-enabled? n2)))
                           (when int3
                              (if (no-conflict n2 on)
                                 (do-one-eps-place esg (append on (list n2))
                                    (cdr accel))
                                 (add-dummy-to esg))))
                        (loop2 (cdr out))))
                  (when (and (splace-immediate? p)
                             (eq? state (eps-subgraph-enabled? esg)))
                     (do-one-eps-place esg (add-dead-to-esg on)
                        (cdr accel)))))
            ((and (= insofar 0) (= outsofar 1))
               (let loop ((in innew))
                  (when in
                     (let* ((n (car in))
                            (int2 (strans-enabled? n)))
                        (when int2
                           (if (no-conflict n on)
                              (do-one-eps-place esg (cons n on) (cdr accel))
                              (add-dummy-to esg))))
                    (loop (cdr in)))))
            ((and (= insofar 1) (= outsofar 1))
               (do-one-eps-place esg on (cdr accel)))
            (else #t)))
      (when on
         (add-if-not-superset on esg))))
;
;   This function updates the enabling of all dirty transitions.
;
;   First, we calculate the weak enabling.  As we do this, we keep track
;   of any epsilon place sets that need recalculating.
;
;   Then, for each epsilon place set that needs recalculating, we
;   calculate the list of virtual transitions that can enable it.
;
;   One problem we need to avoid is that of having something on the
;   consider list multiple times.  We avoid this by creating a new cell
;   and sticking a pointer to it in magik.  We ignore any transitions or
;   groups that already have the magik number.
;
(define *magik-counter* 0)
;
(define (calculate-enabling)
;  (set! t3 (runtime))
   (set! *magik-counter* (+ 1 *magik-counter*))
   (let loop ((tl (sgraph-consider *sgraph*)) (enabled '()))
      (if tl
         (let* ((n (car tl)) (te (strans-eps-connected n)))
            (when (strans-dirty? n)
               (weak-enabling n))
            (if (eq? (strans-magik n) *magik-counter*)
               (loop (cdr tl) enabled)
               (begin
                  (strans-magik-set! n *magik-counter*)
                  (loop (cdr tl)
                     (if (and (not te) (strans-enabled? n))
                        (cons n enabled) enabled)))))
         (let loop ((esgl (sgraph-consider-esg *sgraph*))
                    (en2 enabled) (next '()))
            (if esgl
               (let ((esg (car esgl)))
                  (when (eps-subgraph-dirty? esg)
                     (esg-enabling esg))
                  (if (eq? (eps-subgraph-magik esg) *magik-counter*)
                     (loop (cdr esgl) en2 next)
                     (begin
                        (eps-subgraph-magik-set! esg *magik-counter*)
                        (loop (cdr esgl)
                           (append (eps-subgraph-enabled? esg) en2)
                           (if (eps-subgraph-enabled? esg)
                              (cons esg next) next)))))
               (begin
                  (sgraph-consider-set! *sgraph* enabled)
                  (sgraph-consider-esg-set! *sgraph* next)
                  (DBG 'e
                     (display-stuff "(" (length enabled) "+"
                        (- (length en2) (length enabled)) ")"))
;                 (log-time 'calc-enabling t3)
                  en2))))))
;
;   This function figures out the enabling time.  It's either a
;   virtual transition (as a list) or a real transition.
;
(define (complex-dead? t)
   (if (pair? t)
      (let loop ((tl (cdr t)))
         (if tl
            (or (strans-dead? (car tl)) (loop (cdr tl)))
            #f))
      (strans-dead? t)))
;
(define (complex-orbits t)
   (if (pair? t)
      (begin
         (when (null? (car (cddar t)))
            (set-car! (cddar t)
               (sort (apply append (map strans-orbits-lambda (cdr t))) <)))
         (car (cddar t)))
      (strans-orbits t)))
;
(define (complex-input? t)
   (if (pair? t)
      (let loop ((tl (cdr t)))
         (if tl
            (or (strans-input? (car tl)) (loop (cdr tl)))
            #f))
      (strans-input? t)))
;
(define (complex-visible t)
   (let ((ci
            (if (pair? t)
               (let loop ((tl (cdr t)) (sofar '()))
                  (if tl
                     (loop (cdr tl)
                        (union (append (strans-input? (car tl))
                                       (strans-output? (car tl)))
                           sofar))
                     sofar))
               (append (strans-input? t)
                       (strans-output? t)))))
      (if (> (length ci) 1)
         (begin
            (display-stuff " Length: " (length (cdr t)))
            (merror "Oops; multiple visible transitions" ci))
         (and ci (car ci)))))
;
;   This one says whether a node is a non-input dead node.  If we see
;   an input node, we immediately return #f.
;
(define (complex-dead-non-input? t)
   (if (pair? t)
      (let loop ((tl (cdr t)) (dead? #f))
         (if tl
            (let ((s (car tl)))
               (if (strans-input? s) #f
                  (loop (cdr tl) (or (strans-dead? s) dead?))))
            dead?))
      (and (not (strans-input? t)) (strans-dead? t))))
;
(define (complex-much-later? t)
   (if (pair? t)
      (caar t)
      (strans-much-later? t)))
;
(define (complex-timed-places t)
   (if (pair? t)
      (cadar t)
      (strans-timed-places t)))
;
(define (complex-timed-b t)
   (vector-ref (complex-timed-places t) 0))
;
(define (complex-timed-bi t)
   (vector-ref (complex-timed-places t) 1))
;
(define (complex-timed-c t)
   (vector-ref (complex-timed-places t) 2))
;
;   When we check a timing constraint, we need to make sure that the
;   lower bound of the constraint is met.  The upper bound is
;   automatically met by our region arithmetic (upper bounds of behavioral
;   places limit the region itself.)  So we make sure the lower bound
;   of firing times is less than or equal to the upper bound of the
;   corresponding timing variable.
;
;   We include a flag that allows us to say that all maximums must be
;   violated in order to raise an error.
;
(define (all-timed-satisfied ctp)
   (if ctp
      (let* ((p (car ctp))
             (i (vector-ref *timed-locator* (node-orbit p))))
         (and
            (t<= (car (splace-time-marking p))
               (dref *current-region* 0 i))
            (all-timed-satisfied (cdr ctp))))
      #t))
;
;   We check to see if all of its behavioral
;   timing constraints are satisfied in the current region.  If
;   they are, we leave it on the list.
;
;   If any of the remaining transitions are *DEAD* and non-input
;   nodes, we return just that transition because the others
;   don't matter anyway.
;
;   We also need to make sure that each behavioral place that is
;   `viable' for the current region is also represented by an
;   enabled transition that would remove that token.
;
(define (prune-enabling enabling)
   (if *timing-important?*
      (time-prune-enabling enabling)
      (let loop ((inf '()) (ok '()) (todo enabling))
         (if todo
            (let* ((t (car todo)))
               (if (complex-much-later? t)
                  (loop (cons t inf) ok (cdr todo))
                  (if (and (complex-dead-non-input? t) *die-on-dead*)
                     (begin
                        (display-stuff " dead trans " (fire-label t))
                        (list t))
                     (loop inf (cons t ok) (cdr todo)))))
            (if ok ok
               (if inf inf
                  (begin
                     (display-stuff " deadlock ")
                     '())))))))
;
;   This routine prints out basic information about a place.
;
(define (identify-place p)
   (display-stuff "Place " (node-seq p))
   (when (splace-time-marking p)
      (if (splace-time-constraint? p)
         (display-stuff " time constraint of " (splace-time-marking p))
         (display-stuff " time behavior of " (splace-time-marking p))))
   (display-stuff " input trans")
   (for-each (lambda (e)
                (let ((n (edge-innode e)))
                   (when (node-type? n 'T)
                      (display-stuff " " (fire-label n)))))
      (node-inedges p))
   (display-stuff " output trans")
   (for-each (lambda (e)
                (let ((n (edge-outnode e)))
                   (when (node-type? n 'T)
                      (display-stuff " " (fire-label n)))))
      (node-outedges p))
   (display-stuff 'newline))
;
(define (time-prune-enabling enabling)
   (let ((vt (make-vector (+ *timed-tokens* 1) #f)))
      (let loop ((inf '()) (ok '()) (todo enabling))
         (if todo
            (let* ((t (car todo)) (ctp (complex-timed-bi t)))
               (for-each (lambda (p)
                            (vector-set! vt
                               (vector-ref *timed-locator* (node-orbit p)) #t))
                  ctp)
               (if (complex-much-later? t)
                  (loop (cons t inf) ok (cdr todo))
                  (if (all-timed-satisfied (complex-timed-b t))
                     (if (and (complex-dead-non-input? t) *die-on-dead*)
                        (begin
                           (display-stuff " dead trans " (fire-label t))
                           (list t))
                        (loop inf (cons t ok) (cdr todo)))
                     (loop inf ok (cdr todo)))))
            (let fori ((i 1))
               (if (> i *timed-tokens*)
                  (begin
; (display-stuff "Initially " (length enabling) " inf " (length inf)
;  " ok " (length ok) 'newline)
                     (if ok ok
                         (if inf inf
                                  (begin
                                     (display-stuff " deadlock ")
                                     '()))))
                  (if (or (vector-ref vt i)
                         (splace-time-constraint?
                            (vector-ref *timed-token-ar* i))
                         (t< (dref *current-region* 0 i)
                            (car (splace-time-marking
                                    (vector-ref *timed-token-ar* i)))))
                     (fori (+ i 1))
                     (begin
                        (display-stuff " beh place violation " i " "
                            (dref *current-region* 0 i) " "
                            *current-region* 'newline)
                        (identify-place (vector-ref *timed-token-ar* i))
                        '()))))))))
;
(define (prune-enabling-no-beh enabling)
   (let loop ((inf '()) (ok '()) (todo enabling))
      (if todo
         (let* ((t (car todo)) (ctp (complex-timed-places t)))
            (if (complex-much-later? t)
               (loop (cons t inf) ok (cdr todo))
               (if (or (null? ctp) (all-timed-satisfied ctp))
                  (if (complex-dead-non-input? t)
                     (list t)
                     (loop inf (cons t ok) (cdr todo)))
                  (loop inf ok (cdr todo)))))
; (display-stuff "Initially " (length enabling) " inf " (length inf)
;  " ok " (length ok) 'newline)
         (if ok ok inf))))
;
;   This routine implements the stubborn set method, only on orbits
;   rather than transitions.  For now it enforces the restriction
;   that all enabled transitions must be simple and must be enabled
;   for perpetuity.  (We hope later to extend this method.)  We don't
;   actually check here, though.
;
;   The first thing we do is build closure sets for enabled transitions
;   based on the orbits that are enabling each transition.
;
;   Since typically there will be (many) fewer enabled transitions
;   than orbits, we want to do this by transitions.  But we don't want
;   a potential quadratic algorithm here, since in many cases there
;   are a *whole bunch* of enabled transitions.  (For instance, the
;   dining philosophers.)  So we bite the bullet and do it by orbits.
;   (This isn't necessarily bad; it's linear and there's only a couple
;   of scans which should proceed rapidly.)
;
;   Then we sort each closure set.
;
(define *closure-array* '())
;
(define *closure-list* '())
;
(define *closure-sizes* '())
;
(define *closure-min* 0)
;
(define *closure-add* 0)
;
; (rest moved to ss.scm since it is no longer current)
;
;   Finally, we select a region to fire, and fire.
;
(define (choose-and-fire-transition enabling)
   (let ((t (nth (rand (length enabling)) enabling)))
      (fire-complex t)))
;
(define *num-firings* 0)
;
(define *num-unfirings* 0)
;
(define (current-depth) (- *num-firings* *num-unfirings*))
;
;   The most complex operation we need to perform is to `fire' a
;   transition.  This involves removing tokens from certain timed
;   places and adding tokens to other timed places in such a way
;   that we capture all the reachable states.  (Firing maintains
;   the convexity of the polygon.)  Firing is done with the following
;   procedure:
;
;   - Apply the required firing inequalities to the matrix based on
;      the behavioral timed places.  If there were any of these,
;      run apsp.  (If there were only one, as will be common, then
;      run the abbreviated apsp based on the reduction of a single
;      constraint.)
;
;   - Check that all constraint places that are still marked are
;      fully satisfied.  (We cannot check before we select a firing.)
;      If they are not, then set *firing-failed*.  Extend the set of
;      timing constraints seen if we have now seen more than before.
;
;   - Save the current values of the region, the timed locator, the
;      timed token array, and the timed token count.
;
;   - Actually do the firing.
;
;   - Sort the timed token array.
;
;   - Create a new region matrix.  Copy all of the old values that
;      still exist into the new matrix.  This covers (0 x old),
;      (0 x 0), (old x old), and (old x 0).
;
;   - For all new timed places, set a_{ij} = 0 since they must increment
;      in sync.  This also initializes the new diagonal entries.
;      This covers (new x new).
;
;   - For all pairs of (new x old), set a_{ik} = a_{0k}.  For all
;      pairs of (old x new), set a_{ik} = a_{i0}.
;
;   - For all (old and new) constraints, update a_{0i} from the
;      maximums from the places themselves.  Make sure the constraint
;      places get infinity.  This corresponds to allowing time to
;      advance in the new region.  This covers (0 x old) and (0 x new).
;
;   - For all new constraints, set a_{i0} to 0.  This covers (new x 0)
;      and thus, with the others, the entire matrix.
;
;   - Run APSP over this matrix to make it canonical.  If it disappears,
;      set *firing-failed*.
;
(define (adjust-region t)
   (let loop ((pl (complex-timed-b t)))
; (display-stuff "Region before adjustment: " *current-region* 'newline)
      (if pl
         (let* ((n (car pl))
                (i (vector-ref *timed-locator* (node-orbit n))))
            (if (t<= (car (splace-time-marking n))
                     (- (dref *current-region* i 0)))
               *current-region*
               (let ((v (dcopy *current-region*)))
                  (DBG 'f
                     (display-stuff "*" (t+ (car (splace-time-marking n))
                                              (dref *current-region* i 0))
                                    " "))
                  (dset! v i 0 (- (car (splace-time-marking n))))
                  (when (< (apsp1 v i 0) 0)
                     (merror "Oops; empty region in adjust region!" v))
                  v)))
         *current-region*)))
;
(define *require-all-maxes-for-violation?* #f)
;
(define (check-constraint-places-in-region v pl)
   (let loop ((pl pl) (maxok #f) (maxnotok '()))
      (if pl
         (let* ((p (car pl))
                (t (splace-time-marking p)) (tr (splace-trange-seen p))
                (i (vector-ref *timed-locator* (node-orbit p))))
            (when (t< (- (dref v i 0)) (car t))
               (set! maxnotok (cons (cons i t) maxnotok))
               (set! *firing-failed?* #t))
            (if (t< (cdr t) (dref v 0 i))
               (set! maxnotok (cons (cons i t) maxnotok))
               (set! maxok #t))
            (when (or (null? (car tr)) (< (- (dref v i 0)) (car tr)))
               (set-car! tr (- (dref v i 0))))
            (when (or (null? (cdr tr)) (> (dref v 0 i) (cdr tr)))
               (set-cdr! tr (dref v 0 i)))
            (loop (cdr pl) maxok maxnotok))
         (when (or *firing-failed?*
                  (and maxnotok
                     (or (not *require-all-maxes-for-violation?*)
                        (not maxok))))
            (display-stuff " constraint violation: " maxnotok " in " v)
            (for-each (lambda (a)
                         (identify-place (vector-ref *timed-token-ar*
                                                     (car a))))
               maxnotok)
            (set! *firing-failed?* #t)))))
;
(define (save-location-information)
   (set! *otimed-tokens* *timed-tokens*)
   (let loop ((i 1))
      (when (<= i *timed-tokens*)
         (vector-set! *otimed-token-ar* i (vector-ref *timed-token-ar* i))
         (vector-set! *otimed-locator*
            (node-orbit (vector-ref *timed-token-ar* i)) i)
         (loop (+ i 1)))))
;
(define (update-current-region ov)
   (let* ((n *timed-tokens*) (on *otimed-tokens*)
          (v (dmake (+ n 1) 0)) (mapping '()))
;(display-stuff "From " on " to " n " timed places." 'newline)
      (set! *current-region* v)
      (when #t ; (not *use-po-timing-now?*)
         (set! mapping (make-vector (+ n 1) '()))
         (vector-set! mapping 0 0)
         (let fori ((i 1))
            (when (<= i on)
               (when (vector-ref *otimed-token-ar* i)
                  (vector-set! mapping
                     (vector-ref *timed-locator*
                        (node-orbit (vector-ref *otimed-token-ar* i))) i))
               (fori (+ i 1)))))
      (let fori ((i 1))
         (when (<= i n)
            (let ((p (vector-ref *timed-token-ar* i)))
               (dset! v 0 i
                  (if (splace-time-constraint? p) 'infinity
                      (cdr (splace-time-marking p)))))
            (let* ((mi (and mapping (vector-ref mapping i)))
                   (vi (vector-ref v i))
                   (noi (node-orbit (vector-ref *timed-token-ar* i)))
                   (ovmi (if mi (vector-ref ov mi) 0)))
               (if *use-po-timing-now?*
                  (begin
                     (vector-set! vi 0 (if mi (vector-ref ovmi 0) 0))
                     (let forj ((j 1))
                        (when (<= j n)
                           (vector-set! vi j
                              (calc-po-timing
                                          (node-orbit
                                             (vector-ref *timed-token-ar*
                                                      j)) noi))
                           (forj (+ j 1)))))
                  (let forj ((j 0))
                     (when (<= j n)
                        (let ((mj (vector-ref mapping j)))
                           (if (and mi mj)
                              (vector-set! vi j (dref ov mi mj))
                              (when (or mi mj)
                                 (vector-set! vi j
                                    (if mi
                                        (vector-ref ovmi 0)
                                        (dref ov 0 mj))))))
                        (forj (+ j 1))))))
            (fori (+ i 1))))
;
;   Now we need to increase any of the time constraints with infinity
;   as the upper bound.
;

;  (display-stuff "Region before apsp: " v 'newline)
      (if *use-po-timing-now?*
         (begin
            (when (< (apsp-new0 v) 0)
               (merror "Somehow we went void in update current region" v)))
         (when (< (apsp-0row v) 0)
            (merror "Somehow we went void in update current region" v)))
;  (display-stuff "Region after apsp: " v 'newline)
;
;   Normalization here.  Also do it for constraint places > upper limit.
;
;   Also set a flag from the maximum to indicate whether we need the
;   post normalization phase.
;
   (my-normalize v n)))
;
(define *future-equiv-hack?* #f)
;
(define (my-normalize v n)
;   (display-stuff "Pre-normalize: " v 'newline)
   (when *future-equiv-hack?*
      (let fori ((i 1) (f? #f))
         (if (<= i n)
            (let* ((p (vector-ref *timed-token-ar* i))
                   (tm (splace-time-marking p)))
               (if (not (splace-time-constraint? p))
                  (begin
                     (dset! v 0 i (cdr tm))
                     (let forj ((j 1))
                        (if (<= j n)
                           (begin
                              (when (>= (dref v j i)
                                       (+ (car tm) (dref v j 0)))
                                 (dset! v j i 'infinity))
                              (forj (+ j 1)))
                           (fori (+ i 1) #t))))
                  (fori (+ i 1) #t)))
             (when f?
               (apsp v))
            )))
   (let ((fixup '()) (premax (make-vector (+ n 1) '())))
      (let fori ((i 1))
         (when (<= i n)
            (let* ((p (vector-ref *timed-token-ar* i))
                   (tm (splace-time-marking p)))
               (vector-set! premax i
                  (if (splace-time-constraint? p)
                     (if (eq? (cdr tm) 'infinity)
                         (car tm)
                         (+ (cdr tm) 1))
                     (if (eq? (cdr tm) 'infinity)
                         (car tm)
                         (cdr tm)))))
            (fori (+ i 1))))
      (let fori ((i 1))
         (when (<= i n)
            (when (> (- (dref v i 0)) (vector-ref premax i))
               (shape-translate v i (+ (dref v i 0) (vector-ref premax i))))
            (fori (+ i 1))))
;
;   After that is done, we get into the interesting normalization phase.
;   We chop the maximums as much as we can (n^2 here) such that the
;   rows and columns are not adversely affected.  When we do this, we
;   also eliminate any infinities that might be left, leading a strictly
;   numerical matrix.
;
      (let fori ((i 1))
         (when (<= i n)
            (let ((o (dref v 0 i)) (t (vector-ref premax i)))
               (when (t< t o)
                  (let forj ((j 1))
                     (when (<= j n)
                        (let ((vv (tmin (dref v 0 j)
                                     (vector-ref premax j))))
                        (when (t< (t+ t (dref v i j)) vv)
                           (set! t (- vv (dref v i j)))))
                        (forj (+ j 1))))
                  (when (t< t o)
                     (dset! v 0 i t)
                     (set! fixup (cons i fixup)))))
            (fori (+ i 1))))
      (when fixup
;  (display-stuff "Fixup: " fixup 'newline)
         (let fori ((i 1))
            (when (<= i n)
               (for-each (lambda (j)
                            (let ((t (+ (dref v i 0) (dref v 0 j))))
                               (when (t< t (dref v i j))
                                  (dset! v i j t))))
                  fixup)
               (fori (+ i 1))))
;          (display-stuff "Post-normalize: " v 'newline)
   )))
;
(define (show-status-intermed)
   (DBG 'x
      (let ((sum 0))
         (hash-iter (lambda (s)
                       (for-each (lambda (s)
                                    (set! sum (+ sum (vector-length s))))
                          (state-shapes s)))
            *state-hash*)
         (display-stuff "[NF=" *num-firings* " NR=" sum " TM="
            (- (round (* (runtime) 1000)) *last-runtime*) "]"))))
;
(define (fire-complex t)
   (DBG 'f
      (display-stuff ">"))
   (set! *num-firings* (+ 1 *num-firings*))
   (set! *firing-failed?* #f)
   (when (= 0 (modulo *num-firings* 50000))
      (show-status-intermed))
   (set! *current-state* (+ *current-state* (complex-delta t)))
   (let ((v *current-region*) (ctp '()))
      (when *timing-important?*
         (set! ctp (complex-timed-c t))
         (when #t ; (or ctp (not *use-po-timing-now?*))
            (set! v (adjust-region t))
            (check-constraint-places-in-region v ctp))
         (save-location-information))
      (let ((r (if (pair? t)
                  (begin
                     (for-each prefire (cdr t))
                     (DBG 'f
                        (display-stuff (printable-list-label (cdr t))))
                     (map-inorder postfire (cdr t)))
                  (begin
                     (prefire t)
                     (DBG 'f
                        (display-stuff (printable-label t)))
                     (postfire t)))))
         (DBG 'f
            (display " "))
         (DBG 't
            (display-stuff 'newline))
         (when *new-reg-vals*
            (flush-regs))
         (if *timing-important?*
            (begin
               (sort-timed-tokens)
               (when *use-po-timing-now?*
                  (do-po-work t))
               (update-current-region v)
               r)
            (if *dataflow-important?*
               r
               t)))))
;
(define (fire-label t)
   (let ((l (if (pair? t)
               (printable-list-label (cdr t))
               (printable-label t))))
      (if (or (null? l) (eq? l ""))
         "*"
         l)))
;
(define *save-arcs* #f)
;
;   A successor element is a label consed with a state.  The label
;   is normally a symbol, but it can also be a symbol followed by
;   additional information.
;
(define (state-not-already-in-succ s2 succ)
   (or (null? succ) (and (not (eq? s2 (cdar succ)))
		       (state-not-already-in-succ s2 (cdr succ)))))
;
(define (update-arcs s1 s2 t ci?)
   (when *save-arcs*
; (display-stuff "Updating arcs for state " (mstate-val s1) " to state "
;   (mstate-val s2) " ci? " ci? " t " (fire-label t) 'newline) 
     (let ((v (cons (fire-label t) (complex-visible t))))
         (when (state-not-already-in-succ s2 (state-succ s1))
            (state-succ-set! s1 (cons (cons v s2) (state-succ s1)))
            (when (not ci?)
               (state-utility-set! s2 (cons s1 (state-utility s2))))))))
;
(define (update-min-arcs s1 s2 t)
   (state-succ-set! s1 (cons (cons t s2) (state-succ s1)))
   (state-utility-set! s2 (cons s1 (state-utility s2))))
;
;   These functions write the state graph.
;
(define *write-state-graph-file* #f)
;
(define (write-state-graph n)
   (set! *write-state-graph-file* (open-output-file n)))
;
(define (close-state-graph)
   (close-output-port *write-state-graph-file*)
   (set! *write-state-graph-file* #f))
;
(define (state-graph-node-write n)
   (display-stuff-port *write-state-graph-file* "N F F" n " " n 'newline))
;
(define *arc-counter* 0)
;
(define (state-graph-arc-write s1 s2 t)
   (display-stuff-port *write-state-graph-file*
      "N E A" *arc-counter* " " (if (symbol? t) t (car t)) 'newline)
   (display-stuff-port *write-state-graph-file*
      "E P F" (state-seq s1) " 0 A" *arc-counter* " 0" 'newline)
   (display-stuff-port *write-state-graph-file*
      "E P A" *arc-counter* " 0 F" (state-seq s2) " 0" 'newline)
   (set! *arc-counter* (+ *arc-counter* 1)))
;
(define (write-state-graph-of h)
   (when *save-arcs*
      (let ((live 0))
         (hash-iter (lambda (a) (when (not (state-dead? a))
                                   (set! live (+ 1 live))))
            h)
         (if (> live 300000)
            (display-stuff
               " (state graph has more than 300000 nodes; not writing)")
            (begin
               (hash-iter (lambda (a)
                             (when (not (state-dead? a))
                                (state-graph-node-write (state-seq a))))
                  h)
               (hash-iter (lambda (a)
                             (when (not (state-dead? a))
                                (map (lambda (ar)
                                        (when (not (state-dead? (cdr ar)))
                                           (state-graph-arc-write
                                              a (cdr ar) (car ar))))
                                   (state-succ a)))) h))))
      (close-state-graph)))
;
;   This function writes out a state graph.
;
   (when *write-state-graph-file*
      (state-graph-arc-write s1 s2 t))
;
(defmac (mstate-val a)
   `(if *big-states?* (state-val ,a) ,a))
;
(define (new-state-hash)
   (set! *state-counter* 1)
   (if *big-states?*
      (new-hash state-val-lambda)
      (new-hash (lambda (a) a))))
;
(define *state-counter* 1)
;
;   This function combines the search and insert functions.  It is
;   slightly faster than the two combined because it only does a single
;   hash rather than two.  It also adds a counter to the state to make
;   it easier to write a state graph file.
;
;   I hate to do this this way, but unfortunately there was an earlier bug
;   that is most easily fixed with a global variable . . .
;
(define *latest-was-new* #f)
;
(define (new-state?)
   *latest-was-new*)
;
(define *big-states?* #f)
(define *print-interval* 50)
;
(define (find-new-state? h k)
   (set! *latest-was-new* #f)
   (when (>= (hash-size h) (hash-limit h))
      (rehash h))
   (let ((hv (modulo k (hash-prime h))))
      (let loop ((hc (vector-ref (hash-table h) hv)))
         (if hc
            (if (= k (if *big-states?* (state-val (car hc)) (car hc)))
                (car hc) (loop (cdr hc)))
            (let ((s (if *big-states?* (make-state) k)))
               (set! *latest-was-new* #t)
               (when (= 0 (remainder *state-counter* *print-interval*))
                  (if (= 0 (remainder *state-counter* (* 10 *print-interval*)))
                     (display-stuff *state-counter*)
                     (display-stuff ".")))
               (when *big-states?*
                  (state-val-set! s k)
                  (state-seq-set! s *state-counter*)
                  (state-dead?-set! s #f))
               (hash-size-set! h (+ 1 (hash-size h)))
               (vector-set! (hash-table h) hv
                  (cons s (vector-ref (hash-table h) hv)))
               (set! *state-counter* (+ 1 *state-counter*))
               s)))))
;
;   This routine is where we check out the regions.  If there are
;   currently no timed tokens, we just see if the region list is empty
;   or not.  If the state space is not empty, we do the following.
;
;   First, we run through the existing list of shapes, trying to find
;   one to `union' with.  If we find one, but the current shape is a
;   subset of the larger one, we just return with no changes.  Otherwise,
;   if we find one, we delete the shape we unioned with from the list
;   of state shapes, set the current shape to the result of the union,
;   and recurse.
;
(define *coalesce-regions?* 0)
(define *require-exact-region-match?* #f)
;
(define (add-new-state? h k)
   (let* ((r (find-new-state? h k)))
; (display-stuff "** " *current-region* 'newline)
; (display-stuff "{" (length reg) "->")
      (set! *shape-subset* 'both) ; in case we are using eq?
      (if (= *timed-tokens* 0)
         (when *big-states?*
            (when (null? (state-shapes r))
               (state-shapes-set! r (list *current-region*))
               (set! *latest-was-new* #t)))
         (if (and *combine-arrays?* *require-exact-region-match?*)
            (let ((t (cache-array *current-region*)))
               (if (and (not *was-a-new-array?*)
                      (or (= *state-counter* 2) (memq t (state-shapes r))))
                  #f
                  (begin
                     (set! *latest-was-new* #t)
                     (state-shapes-set! r (cons t (state-shapes r))))))
            (let loop ((reglist (state-shapes r)))
               (if reglist
                  (let ((t ((if *require-exact-region-match?*
                               shapes-eqv? shape-subset?)
                              *current-region* (car reglist))))
                     (if t
                        (when (not (or (eq? *shape-subset* 'a)
                                      (eq? *shape-subset* 'both)))
                           (shape-destroy (car reglist))
                           (state-shapes-set! r
                              (remove (car reglist) (state-shapes r)))
                           (loop (cdr reglist)))
                        (loop (cdr reglist))))
                  (begin
                     (set! *latest-was-new* #t)
                     (when *combine-arrays?*
                        (set! *current-region*
                              (cache-array *current-region*)))
                     (if (< (length (state-shapes r)) *coalesce-regions?*)
                        (let loop2 ((reglist (state-shapes r)))
                           (if reglist
                              (let ((t (shape-union
                                          *current-region* (car reglist))))
                                 (if t
                                     (begin
                                        (shape-destroy (car reglist))
                                        (shape-destroy *current-region*)
                                        (state-shapes-set! r
                                           (remove (car reglist)
                                              (state-shapes r)))
                                        (set! *current-region* t)
                                        (loop (state-shapes r)))
                                     (loop2 (cdr reglist))))
                              (state-shapes-set! r
                                 (cons *current-region* (state-shapes r)))))
                        (state-shapes-set! r
                           (cons *current-region* (state-shapes r)))))))))
; (display-stuff (length (state-shapes r)) "}")
      r))
;
;   So how do we represent state?  We identify the following cases,
;   from simplest to most complex.
;
;   NOTE:  MUCH OF THIS COMMENT IS OBSOLETE BUT I'M TOO LAZY TO FIX
;   IT AT THE MOMENT.
;
;   1.  No data flow, all timing has a max of infinity:
;     In this case, state is simply which places have tokens in them.
;     Each non-eps place gets one bit.
;
;   2.  No data flow, timing is small integers:
;     In this case, state is simply the time left of the places that
;     have tokens in them.   Each place gets the number of bits
;     required to represent the time value.
;
;   3.  No data flow, timing is large integers:
;     This one is harder, equivalent to linear programming.  We handle
;     this by bracketing, both conservative and liberal, and tightening
;     in on things.
;
;   4.  Data flow, but independent of timing or `correctness'.
;     We handle this case by `ignoring' the data flow, always allowing
;     all arcs to be enabled.
;
;   5.  Data flow, not independent of timing.  (State space search.)
;     This is trickier, since the data flow values can take any number
;     of bits.  The lower portion of the state space is the same as
;     before.  The upper consists of the number of bits required for
;     all of the values, which are maintained as part of the state.
;     Thus, we need to declare the number of bits for each state
;     variable.  But this stinks, so instead we assign a certain
;     ordering for the variables, and then we interleave bytes
;     representing them, so each variable gets a whole byte initially
;     but can get larger.
;
;   5 is useful for checking protocols, for instance, mutual exclusion.
;
;   The first trick is to identify which of the cases we have.  If
;   there are no data flow arcs into transitions, we have 1 through 3
;   (even if there is data flow.)  If there are such arcs, we have the
;   user set a flag indicating whether they `matter' or not.  Selecting
;
;   Actually, these are orthogonal, so we really have nine cases, three
;   that talk about data flow by three that talk about timing.
;
;   Determining the timing characteristics is easy.  So is determining
;   the data flow characteristics.
;
;   We are talking some *mighty* large integers, here.  But we will
;   put off the optimizations for just a little bit now.
;
;   So now let's consider the place markings.  Each place has an
;   integer (possibly large) associated with it.
;
(define *simulation-times?* #f)
;
(define *dataflow-important?* #t)
;
(define *dataflow-transitions?* #f)
;
;   This function gives a join a new value (from a register.)  It updates
;   the state based on the difference from the new and the old values.
;
(define *df-state-spread* 1)
;
(define (spread-integer n)
   (if (eq? n 'unknown) 1
      (if (< n 0)
         (+ (spread-integer (- n)) 1)
         (let loop ((n (* n 2)))
            (if (< n 16) n
               (+ (modulo n 16)
                 (* *df-state-spread*
                    (loop (quotient n 16)))))))))
;
(define (update-df-state dfn newval)
   (let* ((o (spread-integer (sdf-curval dfn)))
          (n (spread-integer newval)))
;  (display-stuff "Updating dfs state " o " " n " "
;                               (sdf-state-value dfn) 'newline)
      (sdf-curval-set! dfn newval)
      (set! *current-state*
         (+ *current-state* (* (sdf-state-value dfn) (- n o))))))
;
;   This set of routines handles nonnegative integer ranges.  An
;   integer range is:
;
;      - A rational `r', that each number in the range must satisfy.
;        If '(), then any value is fine (passes this test, anyway.)
;      - A list of the `on' set, where each is a pair.  The endpoints
;        are considered `in'.  Point ranges are allowed (where both
;        numbers are the same.)  Infinity is not allowed.
;
;   Thus, the total range (all nonnegatives) is (() (0 . max))
;   if `max' is larger than any other value.
;
(define (range-all max)
   (list '() (cons 0 max)))
;
;   This function removes a single subrange from a given range suffix
;   (not including the divider part.)
;
(define (sub-from-range r min max)
   (if r
      (if (<= max (caar r)) r
         (if (>= min (cdar r))
            (cons (car r) (sub-from-range (cdr r) min max))
            (if (> max (cdar r))
               (if (< min (caar r))
                  (sub-from-range (cdr r) min max)
                  (cons (cons (caar r) min) (sub-from-range (cdr r) min max)))
               (if (< min (caar r))
                  (cons (cons max (cdar r)) (cdr r))
                  (cons (cons (caar r) min)
                     (cons (cons max (cdar r)) (cdr r)))))))
      '()))
;
;   This is the converse of the above; it finds the intersection
;   between a range and a min/max value.  It requires that the
;   min/max have numerical values.
;
(define (inter-with-range r min max)
   (if r
      (if (< max (caar r)) '()
         (if (> min (cdar r))
            (inter-with-range (cdr r) min max)
            (if (>= max (cdar r))
               (if (<= min (caar r))
                  (cons (car r) (inter-with-range (cdr r) min max))
                  (cons (cons min (cdar r))
                     (inter-with-range (cdr r) min max)))
               (if (<= min (caar r))
                  (cons (cons (caar r) max) '())
                  (cons (cons min max) '())))))
      '()))
;
(define (range-intersect r1 r2)
   (clean-range
      (cons
         (if (and (car r1) (car r2))
            (gcd (car r1) (car r2))
            (or (car r1) (car r2)))
         (apply append
            (map (lambda (a) (inter-with-range
                                (cdr r1) (car a) (cdr a))) (cdr r2))))))
;
;   Here we select a number in the range.  If the range is fragmented, we
;   always select from the lower portion.  Then, we always select the
;   number with the smaller denominator.  But we do check to see if there
;   is some integer that is in range first.
;
(define (select-midrange r)
   (let ((lo (caadr r)) (hi (cdadr r)))
      (if (car r)
         (/ (car r) (floor (/ (car r) lo)))
         (if (<= (ceiling lo) hi)
            (ceiling lo)
            (if (> (denominator hi) (denominator lo)) lo hi)))))
;
(define (new-range min max)
   (list '() (cons min max)))
;
(define (range-null? r)
   (null? (cdr r)))
;
;   This function makes sure that each subrange of the range contains
;   at least one integral divisor of the divisor.
;
(define (clean-range r)
   (define (clean-subrange d cl)
      (if cl
         (if (or (eq? (caar cl) 0)
                 (> (floor (/ d (caar cl))) (floor (/ d (cdar cl))))
                 (integer? (/ d (caar cl))) (integer? (/ d (cdar cl))))
            (cons (car cl) (clean-subrange d (cdr cl)))
            (clean-subrange d (cdr cl)))
         '()))
   (if (car r)
      (cons (car r) (clean-subrange (car r) (cdr r)))
      r))
;
;   Checks to see if a value is in a range.
;
(define (in-range? d r)
   (define (in-range-aux cl)
      (and cl (>= d (caar cl))
         (or (<= d (cdar cl)) (in-range-aux (cdr cl)))))
   (and
      (or (not (car r)) (integer? (/ (car r) d)))
      (in-range-aux (cdr r))))
;
;   This function takes a given range and excludes from it all
;   values that might cause a conservative approximation of
;   const based on those values to be `illegal' (to have a
;   max greater than a min.)
;
(define (range-cons-tighten range const bmax)
   (let ((max (cdr const)) (min (car const)))
      (if (and (number? max) (number? min))
         (if (= min max)
            (if (car range)
               (cons (gcd min (car range)) (cdr range))
               (cons min (cdr range)))
            (let loop ((n 0) (r (cdr range)))
               (if (< (* n max) (* (+ n 1) min))
                  (loop (+ n 1) (sub-from-range r (/ max (+ n 1))
                                   (if (= n 0) (* bmax 2) (/ min n))))
                  (cons (car range) r))))
         range)))
;
(define (scan-constraints c max)
   (let loop ((r (range-all max)) (cl c))
      (if cl
         (loop (range-cons-tighten r (caar cl) max) (cdr cl))
         (clean-range r))))
;
;   This function does all of the timing stuff we need to do exactly
;   once for a graph.
;
(define (handle-beh-con-places)
;   (for-each (lambda (a)
;                (display-stuff "Got " (length a) " unique times" 'newline)
;                (for-each (lambda (b)
;                   (display-stuff (car b) ":" (length (cdr b)) 'newline))
;                   a))
;     (list (sgraph-beh-times *sgraph*) (sgraph-con-times *sgraph*)))
;
;   Now we go through, collect the denominators, and find the lcm of
;   the denominators and the gcd of the numerators.
;
   (let* ((den 1) (num #f) (max 0) (free-vars '())
          (add-one
             (lambda (a)
                (when (and (number? a) (> a 0))
                   (set! den (lcm den (denominator a)))
                   (when (> a max) (set! max a))
                   (set! num (if num (gcd num (numerator a))
                                (numerator a))))))
         (fix-one
            (lambda (a)
               (when (number? (car a))
                  (set-car! a (* (car a) (/ den num))))
               (when (number? (cdr a))
                  (set-cdr! a (* (cdr a) (/ den num))))))
         (check-sym
            (lambda (a)
               (when (and (symbol? a) (not (eq? a 'infinity)))
                  (when (member a free-vars)
                     (merror "Oops; a timing variable is used improperly" a))
                  (set! free-vars (cons a free-vars))))))
      (for-each (lambda (a) (add-one (caar a)) (add-one (cdar a)))
         (sgraph-beh-times *sgraph*))
      (for-each (lambda (a) (add-one (caar a)) (add-one (cdar a)))
         (sgraph-con-times *sgraph*))
      (when (not num)
         (set! num 1))
      (display-stuff " Scale: " num "/" den)
      (for-each (lambda (a) (fix-one (car a)))
         (sgraph-beh-times *sgraph*))
      (for-each (lambda (a) (fix-one (car a)))
         (sgraph-con-times *sgraph*))
      (set! max (* max (/ den num)))
      (sgraph-b-lib-r-set! *sgraph*
         (scan-constraints (sgraph-beh-times *sgraph*) max))
      (sgraph-c-lib-r-set! *sgraph*
         (scan-constraints (sgraph-con-times *sgraph*) max))
      (sgraph-tmax-set! *sgraph* (* (+ max 1) 5))
      (sgraph-global-tscale-set! *sgraph* (/ num den))
;
;   Now we do some checking.  If either value in a timing is symbolic,
;   we make sure that it only occurs once in all of the lists.
;
      (for-each (lambda (a) (check-sym (caar a)) (check-sym (cdar a)))
         (sgraph-beh-times *sgraph*))
      (sgraph-tb-assoc-set! *sgraph* (map (lambda (a) (list a 1 '() '()))
         free-vars))
      (set! free-vars '())
      (for-each (lambda (a) (check-sym (caar a)) (check-sym (cdar a)))
         (sgraph-con-times *sgraph*))
      (sgraph-tc-assoc-set! *sgraph* (map (lambda (a) (cons a (cons '() '())))
         free-vars))
;
;   Finally, we set up the sort keys.  We try to set it up so similar
;   places get sorted in the same order; this should help maximize the
;   array row and column sharing.
;
      (let ((pls (append (sgraph-beh-times *sgraph*)
                    (sgraph-con-times *sgraph*))))
         (let loop ((m (+ 1 (graph-maxorbits *graph*))) (i 0) (pls pls))
            (when pls
               (for-each (lambda (p) (splace-sort-key-set! p
                                        (+ i (node-orbit p))))
                  (cdar pls))
               (loop m (+ i m) (cdr pls)))))
      ))
;
(define (digit-star-check s)
   (let loop ((p (- (string-length s) 1)) (i 0) (d 1))
      (if (>= p 1)
         (if (<= (char->integer #\0)
                (char->integer (string-ref s p)) (char->integer #\9))
            (loop (- p 1) (+ i (* d (- (char->integer (string-ref s p))
                                       (char->integer #\0)))) (* d 10))
            (if (and (eq? (string-ref s p) #\*)
                       (< p (- (string-length s) 1)))
               (let* ((sym (string->symbol (substring s 0 p)))
                      (sv (assoc (string->symbol s)
                             (sgraph-tb-assoc *sgraph*))))
                  (set-car! (cddr sv) sym)
                  (set-car! (cdddr sv) i)
                  #t)
               #f))
         #f)))
;
;   This function sets the limits for the behavioral place variables.
;   This is also where we handle behavioral constraints of the form
;   symbol*integer.
;
(define upper? #f)
;
(define (set-beh-limits)
   (let ((min? #f) (max? #f))
      (for-each (lambda (a)
                   (if (symbol? (caar a))
                      (let ((v (cdr (assoc (caar a)
                                           (sgraph-tb-assoc *sgraph*)))))
                         (set-car! (cdr v) 0)
                         (set! min? (caar a))
                         (set-car! (cddr v) (cdar a)))
                      (when (and (symbol? (cdar a))
                               (not (eq? 'infinity (cdar a))))
                         (let ((v (cdr (assoc (cdar a)
                                          (sgraph-tb-assoc *sgraph*)))))
                            (set-car! (cdr v) (caar a))
                            (set! max? (cdar a))
                            (set-car! (cddr v) 'infinity)))))
         (sgraph-beh-times *sgraph*))
      (when (and min? max?)
         (merror "Oops; var used for both min and max!" (cons min? max?)))
      (set! upper? max?))
   (let loop ((todo (sgraph-tb-assoc *sgraph*)) (normal '()))
      (if todo
         (loop (cdr todo) (if (digit-star-check (symbol->string (caar todo)))
                             normal
                             (cons (car todo) normal)))
         (if (= (length normal) 1)
            (car normal)
            (merror "Expected only one independent beh variable." normal)))))
;
(define (update-beh-multiples)
   (for-each (lambda (a)
                (when (symbol? (caddr a))
                   (set-car! (cdr a)
                      (let ((b (cadr (assoc (caddr a)
                                        (sgraph-tb-assoc *sgraph*)))))
                         (if (eq? b 'infinity) 'infinity
                            (* (car (cdddr a)) b))))))
      (sgraph-tb-assoc *sgraph*)))
;
;   Here is where we set the verification variables.
;
;   First we consider timing.  We run through the places, looking
;   at their timing.  If all of the non-epsilon places have
;   'infty as their maximum time value, we are timing independent
;   (in some manner of speaking.)  In any case, we assign state
;   bits to each place depending on the range of their timing values.
;
;   A place with timing a..b can have a token of age 0..b,
;   so there are b+1 possibilities.  A place with timing
;   a..infty can have 0..a-1, or ready, so there are
;   a+1 possibilities.  We reserve space accordingly.
;
;   After that, we consider the transitions, and data flow.  If any
;   input arcs are found that are data flow, we set the appropriate
;   flag.
;
;   We work hard on optimizing the state vector, using the orbit information
;   collected during parsing and coalescing timing information, so that
;   our bit vectors don't dominate the state storage.
;
(define (log2 n)
   (if (> n 65536)
      (+ 16 (log2 (quotient n 65536)))
      (if (< n 2) 0 (+ 1 (log2 (quotient n 2))))))
;
(define (log2-ceil n)
   (+ 1 (log2 n)))
;
(define *verification-failed?* #f)
;
;   This one calculates a liberal approximation, given a particular scaling.
;   It is assumed that all symbolic values but 'infinity are already resolved.
;
(define *exact-approximation?* #f)
;
(define (liberal-approx range scale)
   (let ((r (cons (if (eq? (car range) 'infinity) (car range)
                      (floor (/ (car range) scale)))
                  (if (eq? (cdr range) 'infinity) (cdr range)
                      (ceiling (/ (cdr range) scale))))))
      (when *exact-approximation?*
         (when (and (not (eq? (car range) 'infinity))
                    (not (= (floor (/ (car range) scale)) (car r))))
            (set! *exact-approximation?* #f))
         (when (and (not (eq? (cdr range) 'infinity))
                    (not (= (ceiling (/ (cdr range) scale)) (cdr r))))
            (set! *exact-approximation?* #f)))
      r))
;
(define (conservative-approx range scale)
   (let ((low (if (eq? (car range) 'infinity) (car range)
                  (ceiling (/ (car range) scale))))
         (high (if (eq? (cdr range) 'infinity) (cdr range)
                  (floor (/ (cdr range) scale)))))
      (when (and (not (eq? high 'infinity)) (> low high))
         (merror "Bad conservative approximation; internal error?"
            (cons low high)))
      (when *exact-approximation?*
         (when (and (not (eq? (car range) 'infinity))
                    (not (= low (ceiling (/ (car range) scale)))))
            (set! *exact-approximation?* #f))
         (when (and (not (eq? (cdr range) 'infinity))
                    (not (= (ceiling (/ (cdr range) scale)) high)))
            (set! *exact-approximation?* #f)))
      (cons low high)))
;
;   Increasing tmax is done by multiplying it.  Tmax is used as the
;   upper limit of constraint places that are symbolic, and it is
;   increased if a run finds that the given range is insufficient.
;
;   Note that a failure in a constraint place where tmax is violated
;   is a `retry' failure, not a normal failure.
;
(define (increase-sgraph-tmax)
   (sgraph-tmax-set! *sgraph* (* 5 (sgraph-tmax *sgraph*))))
;
;   This function does any substitutions necessary to look up a range.
;
(define (resolve-range r beh?)
   (let ((lo (car r)) (hi (cdr r)))
      (cons
         (if (number? lo) lo
            (if beh? (cadr (assoc lo (sgraph-tb-assoc *sgraph*))) 0))
         (if (or (number? hi) (eq? hi 'infinity)) hi
            (if beh? (cadr (assoc hi (sgraph-tb-assoc *sgraph*)))
               (sgraph-tmax *sgraph*))))))
;
;   We set up the splace-time-marking values in all of the places with
;   timing information.  We do this so that we can calculate the number
;   of bits required, and simulate.
;
(define *breadth-first?* #t)
;
(define *increase-tmax?* #f)
;
(define *use-po-timing?* #t)
;
(define *use-po-timing-now?* #t)
;
(define *use-po-pruning?* #t)
;
(define *use-po-pruning-now?* #t)
;
(define *po-prune-hash* '())
;
(define *po-prune-sum* 0)
;
(define *canonical-sequence* '())
;
(define *next-canonical-sequence* '())
;
(define *last-esg-bit-val* 0)
;
(define (init-po-pruning)
   (set! *po-prune-sum* 0)
   (set! *po-prune-hash* (new-hash car))
   (set! *next-canonical-sequence* '()))
;
(define (po-prune-forward po)
   (when (not *breadth-first?*)
      (when (not *next-canonical-sequence*)
         (po-pruning-ok (po-transition po)))
      (po-esg-bit-set! po *last-esg-bit-val*)
      (po-canonical-sequence-set! po *canonical-sequence*)
      (set! *po-prune-sum* (+ *po-prune-sum* *last-esg-bit-val*))
      (set! *canonical-sequence* *next-canonical-sequence*))
   (set! *next-canonical-sequence* '()))
;
(define (calc-po-preds t)
   (let loop ((sofar '()) (orbs (complex-orbits t)))
      (if orbs
         (let ((pol (vector-ref *orbital-po* (car orbs))))
            (if pol
               (let ((v (po-esg-bit (car pol))))
                  (loop (if (memq v sofar) sofar (cons v sofar)) (cdr orbs)))
               (loop sofar (cdr orbs))))
         sofar)))
;
(define (cons-canon-seq t preds seq)
   (if (or (null? seq) (> *last-esg-bit-val* (car seq)))
      (cons *last-esg-bit-val* seq)
      (begin
         (when (eq? preds 'uncalc)
            (set! preds (calc-po-preds t)))
         (if (member (car seq) preds)
            (cons *last-esg-bit-val* seq)
            (cons (car seq) (cons-canon-seq t preds (cdr seq)))))))
;
(define (po-prune-backward po)
   (when (not *breadth-first?*)
      (set! *canonical-sequence* (po-canonical-sequence po))
      (set! *po-prune-sum* (- *po-prune-sum* (po-esg-bit po)))))
;
(define (po-pruning-ok t)
   (set! *last-esg-bit-val* (complex-esg-bit t))
   (set! *next-canonical-sequence*
      (cons-canon-seq t 'uncalc *canonical-sequence*))
   (let* ((hv (cons (+ *po-prune-sum* *last-esg-bit-val*)
                 *next-canonical-sequence*))
          (hm (iset-member*? *po-prune-hash* hv)))
      (if hm
         #f
         (begin
            (iset-insert *po-prune-hash* hv)
            #t))))
;
;   Here we take a sample transition and see if we have visited that
;   particular partial order before.
;
(define (backtrack-po-prune en)
   (if en
      (if (po-pruning-ok (car en))
         en
         (backtrack-po-prune (cdr en)))
      en))
;
(define (set-verification-variables tscale lib?)
   (set! *simulation-times?* #f)
   (set! *increase-tmax?* #f)
   (set! *current-state* 0)
   (set! *verification-failed?* #f)
   (set! *exact-approximation?* #f)
   (set! *num-firings* 0)
   (set! *num-unfirings* 0)
   (set! *firing-failed?* #f)
   (set! *timed-tokens* 0)
   (set! *current-region* *default-0-region*)
   (sgraph-local-tscale-set! *sgraph*
      (* tscale (sgraph-global-tscale *sgraph*)))
   (for-each (lambda (p)
                (when (not (splace-eps? p))
                   (splace-marked?-set! p #f)))
      (sgraph-places *sgraph*))
   (for-each (lambda (t)
                (strans-dirty?-set! t #t)
                (strans-fired?-set! t #f)
                (strans-enabled?-set! t #f))
      (sgraph-transitions *sgraph*))
   (for-each (lambda (esg)
                (eps-subgraph-enabled?-set! esg #f)
                (eps-subgraph-dirty?-set! esg #t)
                (eps-subgraph-state-set! esg (eps-subgraph-serial esg)))
      (sgraph-esgs *sgraph*))
   (sgraph-consider-set! *sgraph* (sgraph-transitions *sgraph*))
   (sgraph-consider-esg-set! *sgraph* (sgraph-esgs *sgraph*))
   (for-each (lambda (a)
                (let ((t ((if lib? conservative-approx liberal-approx)
                            (resolve-range (car a) #f) tscale)))
; (display-stuff "[" t "]")
                   (for-each (lambda (p)
                                (splace-time-marking-set! p t)
                                (splace-trange-seen-set! p (cons '() '())))
                      (cdr a))))
      (sgraph-con-times *sgraph*))
   (set! *exact-approximation?* #t)
   (for-each (lambda (a)
                (let ((t ((if lib? liberal-approx conservative-approx)
                            (resolve-range (car a) #t) tscale)))
; (display-stuff "[" t "]")
                   (for-each (lambda (p)
                                (splace-time-marking-set! p t))
                      (cdr a))))
      (sgraph-beh-times *sgraph*))
 (display-stuff 'newline)
   (when (or (sgraph-beh-times *sgraph*) (sgraph-con-times *sgraph*))
      (set! *simulation-times?* #t))
   (set! *use-po-timing-now?* (and *use-po-timing?* *simulation-times?*))
   (set! *use-po-pruning-now?* (and *use-po-timing-now?* *use-po-pruning?*
                                  (not *save-arcs*)
                                  (not *dataflow-important?*)))
   (let* ((cvv (make-vector (graph-maxorbits *graph*) 0))
          (cvi (make-vector (graph-maxorbits *graph*) 1))
          (oi (sgraph-orbit-info *sgraph*))
          (real-s 1))
      (let loop ((i 0) (s 2))
         (set! real-s s) ; yuck
         (when (< i (graph-maxorbits *graph*))
            (vector-set! cvi i s)
            (when (not (vector-ref (car oi) i))
               (vector-set! cvv i (vector-ref cvi i)))
            (if (= 0 (vector-ref (cdr oi) i))
               (loop (+ i 1) s)
               (loop (+ i 1)
                  (* s (pow2 (+ 1 (log2 (+ (- (vector-ref (cdr oi) i) 1)
                                              (if (vector-ref (car oi) i)
                                                 0 1))))))))))
      (set! *df-state-spread* 1)
      (when *dataflow-important?*
         (let loop ((sdfs (sgraph-dfs *sgraph*)))
            (when sdfs
               (let* ((n (car sdfs)))
                  (when (node-type? n 'J)
		     (sdf-history-set! n '())
                     (sdf-state-value-set! n real-s)
                     (sdf-curval-set! n 'unknown)
                     (set! *current-state* (+ *current-state* real-s))
                     (set! real-s (* real-s 16))
                     (set! *df-state-spread* (* *df-state-spread* 16))))
               (loop (cdr sdfs))))
         (let loop ((sdfs (sgraph-dfs *sgraph*)))
            (when sdfs
               (let* ((n (car sdfs)))
                  (when (and (node-type? n 'J) (node-label n))
		     (update-df-state n (car (node-label n)))
		     (new-dfs-value n)))
               (loop (cdr sdfs)))))
      (display-stuff " need " (log2 real-s) " state bits.")
      (for-each (lambda (p)
                   (when (not (splace-eps? p))
                      (let ((ni (node-orbit p)))
                         (splace-state-value-set! p (vector-ref cvv ni))
                         (vector-set! cvv ni (+ (vector-ref cvv ni)
                                                (vector-ref cvi ni))))))
         (sgraph-places *sgraph*))
      (for-each (lambda (t)
                   (strans-state-value-set! t 0)
                   (let loop ((tl (node-outedges t)))
                      (when tl
                         (let ((n (edge-outnode (car tl))))
                            (case (node-type n)
                               (P (strans-state-value-set! t
                                     (+ (strans-state-value t)
                                        (splace-state-value n)))))
                            (loop (cdr tl)))))
                   (let loop ((tl (node-inedges t)))
                      (when tl
                         (let ((n (edge-innode (car tl))))
                            (case (node-type n)
                               (P (strans-state-value-set! t
                                     (- (strans-state-value t)
                                        (splace-state-value n)))))
                            (loop (cdr tl))))))
         (sgraph-transitions *sgraph*))
      (set! *dataflow-transitions?* #f)
      (let loop ((tl (sgraph-transitions *sgraph*)))
         (when tl
            (let loop2 ((el (node-inedges (car tl))))
               (when el
                  (if (eq? (edge-type (car el)) 'D)
                     (set! *dataflow-transitions?* #t)
                     (loop2 (cdr el)))))
            (loop (cdr tl)))))
   (set! *timed-tokens* 0)
   (for-each (lambda (p)
                (when (not (splace-eps? p))
                   (if (node-marked p)
                      (begin
                         (mark-place p)
                         (set! *current-state*
                            (+ *current-state* (splace-state-value p))))
                      (splace-marked?-set! p #f))))
      (sgraph-places *sgraph*))
   (DBG 's
      (display-stuff "<" *current-state* ">"))
   (set! *timing-important?* (timing-important?))
   (sort-timed-tokens)
   (new-array-hash)
   (set! *big-states?*
      (or *timing-important?* *debug-flags* *save-arcs*))
   (set! *print-interval*
      (if *timing-important?* 10 50))
   (let* ((n *timed-tokens*) (v (dmake (+ n 1) 0)))
      (set! *current-region* v)
      (let fori ((i 1))
         (when (<= i n)
            (let* ((p (vector-ref *timed-token-ar* i)))
               (dset! v 0 i
                  (if (splace-time-constraint? p) 'infinity
                     (cdr (splace-time-marking p)))))
            (fori (+ i 1))))
      (apsp *current-region*)
      (my-normalize *current-region* n))
   (when *use-po-timing-now?*
      (init-po-timing #t)
      (when *use-po-pruning-now?*
         (init-po-pruning))))
;
;   This is the main simulation loop.
;
(define *simulating?* #f)
;
(define (simulate-internal)
   (logit "Simulating ")
   (set! *simulating?* #t)
   (set! *die-on-dead* #t)
   (when (and *use-po-timing-now?* (= *timed-tokens* 0))
      (init-po-timing #f)))
;
(define (step num-steps)
   (let loop ((n 0))
      (when (< n num-steps)
         (let ((enabling (prune-enabling (calculate-enabling))))
            (if (and enabling (not *firing-failed?*))
               (begin
                  (choose-and-fire-transition enabling)
		  (when (= 0 (remainder *num-firings* 1000))
		     (if (= 0 (remainder *num-firings* 10000))
			(display-stuff *num-firings*)
			(display-stuff ".")))
                  (loop (+ n 1)))
               (display-stuff "No enabled transitions" 'newline))))))
;
;   Verification is very similar.  We do depth-first only for now.
;
;   The search tree alternates states and backtrack info.  Each piece
;   of backtrack info has:
;
;      - A log of the transition firing
;   *  - The time before the firing
;   *  - A list of remaining transitions to consider
;   *  - An interval over which one of the transitions can be considered.
;
;   Stars are by the things that go around loop to loop.
;
(define *state-hash* '())
;
;   This is useful during debugging; it tells us when we've made a mistake
;   in our state calculations.  It also slows things down quite a bit.
;
(define (check-state-consistency)
   (let loop ((s 0) (pl (sgraph-places *sgraph*)))
      (if pl
         (let ((p (car pl)))
            (if (and (not (splace-eps? p)) (splace-marked? p))
               (loop (+ s (splace-state-value p)) (cdr pl))
               (loop s (cdr pl))))
         (begin
            (when *dataflow-important?*
               (let loop ((sdfs (sgraph-dfs *sgraph*)))
                  (when sdfs
                     (let* ((n (car sdfs)))
                        (when (node-type? n 'J)
                           (set! s (+ s (* (sdf-state-value n)
                                         (spread-integer (sdf-curval n)))))))
                     (loop (cdr sdfs)))))
            (when (not (= s *current-state*))
               (display-stuff "Is " s " but *current-state* is "
                  *current-state* " for a difference of "
                  (- s *current-state*) " ")
               (merror "Oops!  state inconsistent" s))))))
;
(define *start-state* '())
;
(define *die-on-dead* #f)
;
(define (find-visible-transitions-in-trace ptr init how)
   (if (vector? ptr)
      (if (> (vector-length ptr) 1)
         (let ((t (vector-ref ptr 3)))
            (find-visible-transitions-in-trace (vector-ref ptr 2)
               (cons (fire-label t) init) how))
         init)
      (if how
         (let ((t (car ptr)))
            (find-visible-transitions-in-trace (cdr ptr)
               (cons (fire-label t) init) how))
         (if (> (length ptr) 2)
            (let ((t (nth 1 (cadr ptr))))
               (find-visible-transitions-in-trace (cddr ptr)
                  (cons (fire-label t) init) how))
            init))))
;
(define (report-dead ptr how)
   (let ((rep (find-visible-transitions-in-trace ptr '() how)))
      (display-stuff 'newline "Dead trace (len=" (length rep)
         ": " rep 'newline)))
;
(define *use-backtrack-pruning?* #t)
;
;   Here we iterate through the enabling, throwing away transitions
;   that lead to states that we have definitely seen before.
;
(define (backtrack-prune-enabling en)
   (if en
      (if (iset-member? *state-hash*
             (+ *current-state* (complex-delta (car en))))
         (backtrack-prune-enabling (cdr en))
         en)
      '()))
;
;   The single argument to verify is the subroutine to use to prune
;   the enabling.  We might, for instance, substitute a subroutine
;   that calculates stubborn sets.
;
(define *small-and-slow?* #f)
;
(define (verify-internal prune-enabling)
   (set! *breadth-first?* #f)
   (set! *simulating?* #f)
   (logit "Verifying")
   (DBG 's
      (display-stuff '< *current-state* '> 'newline))
   (let* ((h (new-state-hash))
          (root (vector (add-new-state? h *current-state*)))
          (rem (prune-enabling (calculate-enabling)))
          (small-and-slow? *small-and-slow?*)
          (prune-backtrack (if (and (not *dataflow-important?*)
                                  (not *timing-important?*)
                                  (not *save-arcs*)
                                  *use-backtrack-pruning?*)
                                 backtrack-prune-enabling
                                 (begin
                                    (set! small-and-slow? #f)
                                    (if (and *timing-important?*
                                             *use-po-pruning-now?*)
                                       backtrack-po-prune
                                       (lambda (a) a)))))
          (prune-forward (lambda (a) a)))
      (when (not rem)
         (merror "No enabled transitions at start state" 0))
      (set! *state-hash* h)
      (set! *start-state* (vector-ref root 0))
      (let loop ((rem rem) (ptr root))
    (DBG 'd
       (display-stuff "B=" (length rem) " " (map fire-label rem) 'newline))
    (DBG 'c
       (check-state-consistency))
         (if rem ; still more to do
            (let* ((t (car rem))
                   (log (fire-complex t))
                   (nrem (cdr rem))
                   (ns (add-new-state? h *current-state*))
                   (dead1 (complex-dead? t))
                   (nptr (if small-and-slow?
                            (cons t ptr)
                            (vector ns log ptr t nrem)))
                   (fe (if (and (new-state?) (not dead1))
                          (prune-enabling (calculate-enabling)) '(#t)))
                   (dead? (or dead1 *firing-failed?* (null? fe))))
               (DBG 's (display-stuff "<" *current-state* ">"))
               (if (or (not (new-state?)) dead?)
                  (begin
                     (when dead?
                        (set! *verification-failed?* #t)
                        (when *big-states?*
                           (state-dead?-set! ns #t)))
                     (if (and dead? *die-on-dead*)
                        (report-dead nptr small-and-slow?)
                        (loop '() nptr)))
                  (loop (prune-backtrack fe) nptr)))
            (if (eq? ptr root)
               (display-stuff " " (hash-size h) " total states.")
               (let* ((nptr '()) (log '()) (t '()) (ci? '()))
                  (if small-and-slow?
                     (begin
                        (set! nptr (cdr ptr))
                        (set! log (car ptr))
                        (set! t (car ptr)))
                     (begin
                        (set! nptr (vector-ref ptr 2))
                        (set! log (vector-ref ptr 1))
                        (set! t (vector-ref ptr 3))))
                  (set! ci? (complex-input? t))
                  (unfire-complex log)
                  (set! *current-state* (- *current-state* (complex-delta t)))
                  (when *big-states?*
                     (when (and (not ci?) (state-dead? (vector-ref ptr 0)))
                        (state-dead?-set! (vector-ref nptr 0) #t))
                     (when (not (state-dead? (vector-ref ptr 0)))
                        (update-arcs (vector-ref nptr 0) (vector-ref ptr 0)
                                     t ci?)))
                  (loop (if (or (and *big-states?*
                                     (state-dead? (vector-ref nptr 0)))
                               (shape-destroyed? *current-region*))
                           '()
                           (prune-backtrack
                              (if small-and-slow?
                                 (prune-enabling (calculate-enabling))
                                 (vector-ref ptr 4))))
                     nptr)))))
      (logit "Done")
      (when *show-statistics?*
         (show-statistics 0))
      (when *write-state-graph-file*
         (write-state-graph-of h))))
;
;   This next is almost identical, but it does it breadth-first.  Whenever
;   a dead state is encountered, it exits.  (This is because it keeps a lot
;   more state around than depth-first; in essence, a full log for the
;   entire state graph so far explored is maintained.)
;
(define *graph-shape* '())
;
(define (old-verify-bf-internal)
   (set! *breadth-first?* #t)
   (set! *simulating?* #f)
   (logit "Verifying, breadth-first")
   (set! *graph-shape* '())
   (DBG 's
      (display-stuff '< *current-state* '> 'newline))
   (let* ((h (new-state-hash)) (root (list (add-new-state? h *current-state*)))
          (rem (prune-enabling (calculate-enabling))))
      (when (not rem)
         (merror "No enabled transitions at start state" 0))
      (set! *state-hash* h)
      (set! *start-state* (car root))
      (let loop ((rem rem) (ptr (vector (car root))) (this-lev '())
                 (next-lev '()))
         (DBG 'd (display-stuff "B=" (length rem) " "
              (map fire-label rem) " tl="
                         (length this-lev) " nl=" (length next-lev) 'newline))
         (DBG 'c (check-state-consistency))
         (if rem ; still more to do
            (let* ((t (car rem))
                   (log (fire-complex t))
                   (nrem (cdr rem))
                   (ns (add-new-state? h *current-state*))
                   (dead1 (complex-dead? t))
                   (nptr (vector ns log ptr t))
                   (fe (if (and (new-state?) (not dead1))
                          (prune-enabling (calculate-enabling)) '(#t)))
                   (dead? (or dead1 *firing-failed?* (null? fe))))
               (DBG 's (display-stuff "<" *current-state* ">"))
               (when dead?
                  (set! *verification-failed?* #t)
                  (when *big-states?*
                     (state-dead?-set! ns #t)))
               (unfire-complex log)
               (set! *current-state* (mstate-val (vector-ref ptr 0)))
               (prune-enabling (calculate-enabling))
               (if (and dead? *die-on-dead*)
                  (report-dead nptr #f)
                  (loop nrem ptr this-lev
                     (if (or (not (new-state?)) dead?) next-lev
                        (cons nptr next-lev)))))
;
;   No more to do for this transition; we go to the next in the list.
;   To get there, we need to backtrack some amount.
;
            (if this-lev
               (begin
                  (DBG 'd (display-stuff "Next group" 'newline))
                  (let loop ((from ptr) (to (car this-lev)))
                     (when (not (eq? from to))
                        (unfire-complex (vector-ref from 1))
                        (loop (vector-ref from 2) (vector-ref to 2))
                        (fire-complex (vector-ref to 3))))
                  (set! *current-state*
                     (mstate-val (vector-ref (car this-lev) 0)))
                  (let* ((fe (prune-enabling (calculate-enabling))))
                     (loop fe (car this-lev) (cdr this-lev) next-lev)))
               (if next-lev
;
;   Nothing more on this level, which means we're done with this level and
;   we've accumulated everything we need for the next level.  The only
;   problem is, we just undid the transition taking us to the next
;   level.  We have to be careful, since we might not have found a `new'
;   non-dead state for a while.  So we use the same strategy as before
;   with the loop.
;
                  (begin
                     (DBG 'd (display-stuff
                                "Next level " (length next-lev) 'newline))
                     (DBG 'x (display-stuff
                                "Next level " (length next-lev) 'newline))
                     (set! *graph-shape* (cons (length next-lev) *graph-shape*))
                     (let loop ((from ptr) (to (vector-ref (car next-lev) 2)))
                        (when (not (eq? from to))
                           (unfire-complex (vector-ref from 1))
                           (loop (vector-ref from 2) (vector-ref to 2))
                           (fire-complex (vector-ref to 3))))
                     (fire-complex (vector-ref (car next-lev) 3))
                     (set! *current-state*
                        (mstate-val (vector-ref (car next-lev) 0)))
                     (let* ((fe (prune-enabling (calculate-enabling))))
                        (loop fe (car next-lev) (cdr next-lev) '())))
                  (begin
                     (display-stuff " " (hash-size h) " total states.")
                     (logit "Done")
                     (when *show-statistics?*
                        (show-statistics 0)))))))))
;
;   The next section of code converts the nondeterministic state
;   graph created from verification into a deterministic state
;   graph.
;
;   This function does autofailure propogation.  As it handles each
;   node, it clears the noninput predecessors; this stops potential
;   infinite recursion if the circuit has an output loop.
;
(define (recurse-dead s)
   (when (state-dead? s)
      (let ((sl (state-utility s)))
         (state-utility-set! s '())
         (let loop ((sl sl))
            (when sl
               (state-dead?-set! (car sl) #t)
               (recurse-dead (car sl))
               (loop (cdr sl)))))))
;
;   This function is very similar but it works for the deterministic
;   graphs created by minimize.
;
(define (recurse-dead2 s)
   (when (state-dead? s)
      (let ((sl (state-utility s)))
         (state-utility-set! s '())
         (let loop ((sl sl))
            (when sl
               (state-succ-set! (car sl)
                  (remove-if (lambda (a) (eq? (cdr a) s))
                     (state-succ (car sl))))
               (when (null? (state-succ (car sl)))
                  (state-dead?-set! (car sl) #t)
                  (recurse-dead2 (car sl)))
               (loop (cdr sl)))))))
;
;   All dead states are only reachable through input nodes.  We use
;   depth-first search here, building up the list of nodes that are
;   reachable.
;
(define (recurse-eps s)
   (let loop ((stack (list s)) (seen '()))
      (if stack
         (let ((s (car stack)))
            (set! seen (cons s seen))
            (set! stack (cdr stack))
            (let loop ((sl (state-succ s)))
               (when sl
                  (when (and (null? (cdaar sl)) (not (memq (cdar sl) seen)))
                     (set! stack (cons (cdar sl) stack)))
                  (loop (cdr sl))))
            (loop stack seen))
         (state-utility-set! s seen))))
;
;(define (recurse-eps s)
;   (if (symbol? (state-utility s))
;      (begin
;         (if (eq? (state-utility s) 'uncalc)
;            (begin
;               (state-utility-set! s 'calculating)
;               (let loop ((sl (state-succ s)) (sv (list s)))
;                  (if sl
;                     (if (null? (cdaar sl))
;                        (loop (cdr sl) (union (recurse-eps (cdar sl)) sv))
;                        (loop (cdr sl) sv))
;                     (begin
;                        (state-utility-set! s sv)
;                        sv))))
;            (merror "Output loop" (state-seq s))))
;      (state-utility s)))
;
;   Calculates the state integer.
;
(define (min-state-val sl)
   (if (null? sl) 0
      (+ (pow2 (state-seq (car sl))) (min-state-val (cdr sl)))))
;
;   Removes all invisible transitions from the successor list.  This
;   makes our search faster.
;
(define (kill-invisible s)
   (let loop ((sl (state-succ s)) (sofar '()))
      (if sl
         (loop (cdr sl)
            (if (null? (cdaar sl))
               sofar
               (cons (cons (cdaar sl) (cdar sl)) sofar)))
         (state-succ-set! s sofar))))
;
;   This function minimizes the resulting graph by eliminating autofailures
;   and then performing the subset calculation.
;
;   The subset calculation is not terribly difficult.  We recursively
;   calculate a list of the reachable states from a given state based
;   on non-visible transitions.
;
;   These next two functions figure out the set of possible output
;   states for a given set of input states.  For input transitions,
;   *all* states must have transitions or we return the empty list
;   (since one or more may die.)
;
;   If we encounter a dead state when scanning inputs, we also return
;   '().  (Because of autofailure propogation, we don't need to
;   scan for dead states for outputs.)
;
;   First, this utility function is a combination of assoc and member,
;   for those cases when you need to see hidden entries on the assoc
;   list.
;
(define (assmem sym ls)
; (display-stuff "Lookup up " sym " in a list with " (map car ls) 'newline)
   (and ls (if (eq? sym (caar ls)) ls (assmem sym (cdr ls)))))
;
(define (min-input-next t sl)
   (let loop ((sl sl) (tl '()) (sofar '()))
      (if (null? tl)
         (if (null? sl)
            sofar
            (let ((tl (assmem t (state-succ (car sl)))))
               (if (and tl (not (state-dead? (cdar tl))))
                  (loop (cdr sl) (assmem t (cdr tl))
                     (union (state-utility (cdar tl)) sofar))
                  '())))
         (if (state-dead? (cdar tl)) '()
            (loop sl (assmem t (cdr tl))
               (union (state-utility (cdar tl)) sofar))))))
;
(define (min-output-next t sl)
   (let loop ((sl sl) (tl '()) (sofar '()))
      (if (null? tl)
         (if (null? sl)
            sofar
            (loop (cdr sl) (assmem t (state-succ (car sl))) sofar))
         (loop sl (assmem t (cdr tl))
            (union (state-utility (cdar tl)) sofar)))))
;
;   This function renumbers the states in the state graph, moving the
;   dead nodes to the end and preserving the order of the live nodes.
;
(define (renumber-state-graph h)
   (let ((v (make-vector (+ 1 (hash-size h)) '())))
      (hash-iter (lambda (a) (vector-set! v (state-seq a) a)) h)
      (let loop ((i 1) (live 1) (dead (hash-size h)))
         (if (<= i (hash-size h))
            (if (state-dead? (vector-ref v i))
               (begin
                  (state-seq-set! (vector-ref v i) dead)
                  (loop (+ i 1) live (- dead 1)))
               (begin
                  (state-seq-set! (vector-ref v i) live)
                  (loop (+ i 1) (+ live 1) dead)))
            (- live 1)))))
;
;   Minimize does a depth-first search much like verify itself.  Note that
;   we are dealing with two sets of states, the original nondeterministic
;   state graph, and the new deterministic state graph.  So if things get
;   a tiny bit confusing, don't blame me.
;
;   This is perhaps slower than it needs to be, but it is not as crucial
;   as the standard verify imperative.
;
(define (back-merge-states sl)
   (let loop ((i (car sl)) (lp (cdr sl)))
      (if lp
	 (loop (if (< (state-seq i) (state-seq (car lp))) i (car lp)) (cdr lp))
	 (for-each (lambda (s)
		      (when (not (eq? s i))
			 (state-dead?-set! s #t)
			 (state-utility-set! i
			    (append (state-utility s) (state-utility i)))
			 (for-each (lambda (dn)
			              (let loop ((su (state-utility dn)))
	                                 (when su
                                            (if (eq? (car su) s)
			                       (set-car! su i))
			                    (loop (cdr su)))))
			    (map cdr (state-succ s)))
			 (for-each (lambda (up)
			              (let loop ((sd (state-succ up)))
			                 (when sd
			                    (if (eq? (cdar sd) s)
			                       (set-cdr! (car sd) i))
			                    (loop (cdr sd)))))
			    (state-utility s))))
            sl))))
;
(define (back-merge grp)
   (let ((asslst '()) (up? #f))
      (for-each (lambda (a)
		   (let* ((succl (map cdr (state-succ a)))
		          (r? (assoc succl asslst)))
		      (if r?
                         (set-cdr! r? (cons a (cdr r?)))
			 (set! asslst (cons (list succl a) asslst)))))
         (cdr grp))
      (for-each (lambda (a)
		   (when (> (length (cdr a)) 1)
		      (set! up? #t)
		      (back-merge-states (cdr a))))
	 asslst)
      (if up?
         (set-cdr! grp (remove-if state-dead?-lambda (cdr grp))))
      up?))
;
(define (minimize nonreceptive-okay?)
   (logit "Minimizing sgraph")
   (when (not *save-arcs*)
      (merror "Can't minimize without arcs!"))
   (logit "Autofailure elimination . . .")
   (hash-iter recurse-dead *state-hash*)
   (renumber-state-graph *state-hash*)
   (logit "Eps reach calculations . . .")
   (hash-iter (lambda (s) (state-utility-set! s 'uncalc)) *state-hash*)
   (hash-iter recurse-eps *state-hash*)
   (hash-iter kill-invisible *state-hash*)
   (logit "NSG -> SG conversion")
   (set! *timed-tokens* 0)
   (set! *current-region* *default-0-region*)
   (let* ((h (new-state-hash))
          (possib (append (sgraph-inputs *sgraph*) (sgraph-outputs *sgraph*)))
          (curdetstates (state-utility *start-state*))
          (root (list (add-new-state? h (min-state-val curdetstates))))
          (rem possib))
      (let loop ((curdetstates curdetstates) (rem rem) (ptr root))
         (DBG 'd
            (display-stuff "S=" (state-val (car ptr)) " B=" rem
               " SL=" (map state-seq-lambda curdetstates)
                       " d=" (quotient (length ptr) 2) 'newline))
         (if rem ; still more to do
            (let* ((t (car rem)) (in? (memq t (sgraph-inputs *sgraph*)))
                   (nv ((if (and in? (not nonreceptive-okay?))
                           min-input-next min-output-next)
                          t curdetstates))
                   (ns (if nv (add-new-state? h (min-state-val nv)) '()))
                   (nptr (cons ns (cons (list (cdr rem) t curdetstates) ptr))))
               (DBG 's (display-stuff "<" (and ns (state-val ns)) ">"))
               (if ns
                  (if (new-state?)
                     (loop nv possib nptr)
                     (loop nv '() nptr))
                  (loop curdetstates (cdr rem) ptr)))
            (if (eq? ptr root)
               (display-stuff " " (hash-size h) " states."
                  'newline)
               (let* ((bt (cadr ptr)) (nptr (cddr ptr))
                      (t (cadr bt)) (rem (car bt)) (ov (car (cddr bt))))
                  (update-min-arcs (caddr ptr) (car ptr) t)
                  (loop ov rem nptr)))))
      (logit "Eliminating dead states . . .")
      (hash-iter (lambda (a) (state-dead?-set! a (null? (state-succ a)))) h)
      (hash-iter recurse-dead2 h)
      ;
      ;   Now we combine equivalent states.  Two states are equivalent if
      ;   their successors are equivalent.  The successors of each state are
      ;   already canonically ordered.
      ;
      (when (= 0 (renumber-state-graph h))
         (display-stuff "No remaining live states!" 'newline))
      (logit "Merging equivalent states . . .")
      (let ((asslst '()))
         (hash-iter (lambda (a)
		       (when (not (state-dead? a))
		          (let* ((ss (state-succ a))
				 (syml (map car ss))
				 (r? (assoc syml asslst)))
			    (if r?
			       (set-cdr! r? (cons a (cdr r?)))
			       (set! asslst (cons (list syml a)
					       asslst))))))
		    h)
	 (let loop ((up? #f))
            (for-each (lambda (grp)
		         (when (> (length (cdr grp)) 1)
			    (set! up? (or (back-merge grp) up?))))
	       asslst)
	    (when up?
	       (loop #f))))
      (display-stuff " " (renumber-state-graph h) " final states." 'newline)
      (when *write-state-graph-file*
         (write-state-graph-of h))
      h))
;
(define (build-large-expr-from-min wl h rok?)
   (min-check-in-wires h wl rok?))
;
(define (min-check-in-wires h wl rok?)
   (if wl
      (let* ((w (verify-wire (car wl) *the-asc*))
             (v (nth 2 w)))
         (if (eq? v 'unknown)
            (min-check-in-wires h (cdr wl) rok?)
            (cons `(wire-init-set! ,(car wl) ,v)
               (min-check-in-wires h (cdr wl) rok?))))
      (let ((s '()))
         (hash-iter
            (lambda (st)
               (when (not (state-dead? st))
                  (map (lambda (ar)
                          (when #t ; (not (state-dead? (cdr ar)))
                             (set! s (cons
                                       `(,(state-seq st)
                                        ,(if (symbol? (car ar))
                                             (car ar) (caar ar))
                                        (loop ,(state-seq (cdr ar)))) s))))
                    (state-succ st)))) h)
         `(sm loop ((s 1)) ,(fix-case `(case s ,@s))))))
;
;   Here we update our measured timing constraints according to whether
;   we used a liberal or conservative approximation.
;
(define (do-one-tc constr pl conservative?)
   (when (or (symbol? (car constr))
            (and (symbol? (cdr constr)) (not (eq? (cdr constr) 'infinity))))
      (let ((hi '()) (lo '())
            (las (assoc (car constr) (sgraph-tc-assoc *sgraph*)))
            (has (assoc (cdr constr) (sgraph-tc-assoc *sgraph*))))
         (for-each (lambda (p)
                      (when (car (splace-trange-seen p))
                         (when (or (null? lo)
                                  (< (car (splace-trange-seen p)) lo))
                            (set! lo (car (splace-trange-seen p))))
                         (when (or (null? hi)
                                  (< hi (cdr (splace-trange-seen p))))
                            (set! hi (cdr (splace-trange-seen p))))))
            pl)
      (when (and lo las)
         (set! lo (* (+ lo (car (splace-time-marking (car pl))))
                     (sgraph-local-tscale *sgraph*)))
         (if conservative?
            (when (or (null? (cddr las)) (< lo (cddr las)))
               (set-cdr! (cdr las) lo))
            (when (or (null? (cadr las)) (> lo (cadr las)))
               (set-car! (cdr las) lo)))
         (display-stuff " new: " las 'newline))
      (when (and hi has)
         (set! hi (* (+ hi (car (splace-time-marking (car pl))))
                     (sgraph-local-tscale *sgraph*)))
         (if conservative?
            (when (or (null? (cadr has)) (> hi (cadr has)))
               (set-car! (cdr has) hi))
            (when (or (null? (cddr has)) (< hi (cddr has)))
               (set-cdr! (cdr has) hi)))
         (display-stuff " new: " has 'newline)))))
;      
(define (update-tc conservative?)
   (for-each (lambda (a) (do-one-tc (car a) (cdr a) conservative?))
      (sgraph-con-times *sgraph*)))
;
;   This function does our timing verification using conservative and
;   liberal approximations.  It repeats until it runs with sufficient
;   accuracy to give a definitive answer.  For now, we ignore the
;   unknown variables.
;
;   We start with `d' equal to the maximal behavioral constraint.
;
(define (tverify-iterate prune-func)
   (let* ((bc (sgraph-b-lib-r *sgraph*))
          (cc (sgraph-c-lib-r *sgraph*))
          (both (range-intersect bc cc)))
      (let loop ((d (if (car both)
                       (/ (car both)
                          (ceiling (/ (car both) (cdar (reverse both)))))
                       (cdar (reverse both)))))
         (display-stuff "Trying a divisor of " d 'newline)
         (when (in-range? d cc)
            (display-stuff "Conservative approx." 'newline)
            (set! *verification-failed?* #f)
            (let loop ()
               (set-verification-variables d #f)
               (verify-internal prune-func)
               (when (and (not *verification-failed?*)
                          (sgraph-tc-assoc *sgraph*))
                  (update-tc #t))
               (when (and *verification-failed?* *increase-tmax?*)
                  (increase-tmax)
                  (loop))))
         (if *verification-failed?*
            (display-stuff "Timing result:  failure" 'newline)
            (begin
               (when (in-range? d cc)
                  (display-stuff "Liberal approx." 'newline)
                  (set! *verification-failed?* #f)
                  (let loop ()
                     (set-verification-variables d #t)
                     (verify-internal prune-func)
                     (when (and (not *verification-failed?*)
                                (sgraph-tc-assoc *sgraph*))
                        (update-tc #f))
                     (when (and *verification-failed?* *increase-tmax?*)
                        (increase-tmax)
                        (loop))))
               (if (and (not *verification-failed?*)
                      (null? (sgraph-tc-assoc *sgraph*)))
                  (display-stuff "Timing result:  success" 'newline)
                  (let* ((newd (new-range (/ d 2) (* d 3/4)))
                         (t1 (range-intersect both newd))
                         (t2 (range-intersect bc newd))
                         (t3 (range-intersect cc newd)))
                     (if (not (range-null? t1))
                        (loop (select-midrange t1))
                        (if (not (range-null? t2))
                           (loop (select-midrange t2))
                           (if (not (range-null? t3))
                              (loop (select-midrange t3))
                              (merror "Can't find satisfactory smaller number"
                                 new-range)))))))))))
;
;   This function tries to find any unknown behavioral values.  It uses
;   binary search.  Either an upper or lower bound must already be
;   specified.
;
;   First we run with the beh variable set so to minimize the range.
;   That must work; if it doesn't, no value will work.  Then we maximize
;   the range.  If that works, then the value is unlimited, and we
;   report that.  Otherwise, we have an upper and lower limit, and we
;   simply loop until we find the correct value.
;
;   We actually increase it a bit more aggressively than binary, since
;   if there is a failure it is generally found more rapidly than a
;   success.  We use .8.
;
(define *max-digits* 4)
;
(define (tverify-search prune-func)
   (let* ((uv (set-beh-limits)))
      (let* ((min (caddr uv)) (max (car (cdddr uv)))
             (acc 2))
         (let loop ((i 0))
            (when (< i *max-digits*)
               (set! acc (* 10 acc))
               (loop (+ i 1))))
         (if (and (not upper?) (eq? max 'infinity))
            (set! *verification-failed?* #f)
            (begin
               (set-car! (cdr uv) (if upper? min max))
               (update-beh-multiples)
               (set-verification-variables 1 #f)
               (verify-internal prune-func)))
         (if *verification-failed?*
            (display-stuff 'newline
     "Minimum range doesn't work; no successful value of " (car uv) 'newline)
            (begin
               (set-car! (cdr uv) (if upper? 'infinity 0))
               (update-beh-multiples)
               (set-verification-variables 1 #f)
               (verify-internal prune-func)
               (if (not *verification-failed?*)
                  (display-stuff 'newline
     "Maximum range works; all values for " (car uv) " work." 'newline)
                  (let loop ((d 1) (t 0))
                     (when (t= (+ min 1) max)
                        (set! min (* 10 min))
                        (when (not (eq? max 'infinity))
                           (set! max (* 10 max)))
                        (set! d (/ d 10)))
                     (set! t (if (eq? max 'infinity)
                                (if (eqv? min 0) 1 (* 2 min))
                                (if upper?
                                   (quotient (+ min (* 4 max)) 5)
                                   (quotient (+ 4 (* 4 min) max) 5))))
                     (set-car! (cdr uv) (* d t))
                     (update-beh-multiples)
                     (display-stuff 'newline "Trying "
                        (* 1.0 (* d t) (sgraph-global-tscale *sgraph*)) ":")
                     (set-verification-variables d #f)
                     (verify-internal prune-func)
                     (if (or (and *verification-failed?* upper?)
                             (not (or *verification-failed?* upper?)))
                        (set! max t)
                        (set! min t))
                     (if (and (not (eq? max 'infinity))
                              (< (* acc (- max min)) (+ max min)))
                        (let ((m (* 1.0 d (sgraph-global-tscale *sgraph*))))
                           (if upper?
                              (display-stuff 'newline (* m min) " works and "
                                 (* m max) " fails." 'newline)
                              (display-stuff 'newline (* m min) " fails and "
                                 (* m max) " works." 'newline)))
                        (loop d t)))))))))
;
;   This next section is our partial order stuff.  It's not terribly
;   complex.
;
(define *orbital-po* '())
;
(define *orbital-po-size* 0)
;
(define *po-level* 0)
;
(define *max-depth* 0)
;
;   These routines maintain a queue of small integers.  The first should
;   be `1'.  The order in which they are grabbed is pretty much
;   irrelevant so long as, whenever possible, they are smaller than
;   po-index-max.
;
(define *po-index-max* 1)
;
(define *po-avail* '())
;
(define *po-inuse* '())
;
(define (init-po-index)
   (set! *po-index-max* 0)
   (set! *po-avail* '())
   (set! *po-inuse* '()))
;
(define (po-index-release-aux n)
   (if (eq? (+ n 1) *po-index-max*)
      (begin
         (set! *po-index-max* n)
         (when (member (- n 1) *po-avail*)
            (set! *po-avail* (remove-first (- n 1) *po-avail*))
            (po-index-release-aux (- n 1))))
      (set! *po-avail* (cons n *po-avail*))))
;
(define (po-index-release n)
   (set! *po-inuse* (remove-first n *po-inuse*))
   (po-index-release-aux n))
;
(define (remove-first n l)
   (if l
      (if (= n (car l)) (cdr l)
         (cons (car l) (remove-first n (cdr l))))
      (merror "Tried to grab a nonexistent number" (cons n l))))
;
(define (po-index-grab n)
   (if (<= *po-index-max* n)
      (if (= *po-index-max* n)
         (begin
            (set! *po-inuse* (cons n *po-inuse*))
            (set! *po-index-max* (+ 1 *po-index-max*)))
         (begin
            (set! *po-avail* (cons *po-index-max* *po-avail*))
            (set! *po-index-max* (+ 1 *po-index-max*))
            (po-index-grab n)))
      (begin
         (set! *po-inuse* (cons n *po-inuse*))
         (set! *po-avail* (remove-first n *po-avail*)))))
;
;   New version:  always releases the minimum value; try to keep things
;   smaller at a slight expense in execution time.
;
(define (po-index-get)
   (if (null? *po-avail*)
      (let ((r *po-index-max*))
         (set! *po-index-max* (+ *po-index-max* 1))
         (set! *po-inuse* (cons r *po-inuse*))
         r)
      (let loop ((nl '()) (mn (car *po-avail*)) (pa (cdr *po-avail*)))
         (if pa
            (if (< (car pa) mn)
               (loop (cons mn nl) (car pa) (cdr pa))
               (loop (cons (car pa) nl) mn (cdr pa)))
            (begin
               (set! *po-avail* nl)
               (set! *po-inuse* (cons mn *po-inuse*))
               mn)))))
;
(define (po-index-max)
   *po-index-max*)
;
(define *po-index* '())
;
(define (count-joins lst sofar)
   (if lst
      (count-joins (cdr lst)
         (if (node-type? (car lst) 'J) (+ sofar 1) sofar))
      sofar))
;
(define (init-po-timing kill-cache?)
   (set! *bv-cache* (make-vector *bv-cache-max* '()))
   (set! *bv-cache-ptr* 0)
   (set! *bv-files* '())
   (set! *po-level* 0)
   (init-po-index)
   (when kill-cache?
      (for-each (lambda (t) (strans-po-stuff-cache-set! t #f))
         (sgraph-transitions *sgraph*)))
   (let ((n (* 15 (+ (graph-maxorbits *graph*) ; YUCK!  BUG BUG BUG
                    (count-joins (sgraph-dfs *sgraph*) 2)))))
      (set! *orbital-po* (make-vector n '()))
      (set! *orbital-po-size* n)
      (set! *po-index* (make-vector n '(()))))
   (set! *translate-deltas* (make-vector (graph-maxorbits *graph*) 0))
   (set! *po-pos* '())
   (set! *bv-level* -1)
   (set! *bv-matrix* '())
   (set! *num-bv-calls* 0)
   (set! *max-depth* 0)
   (do-po-work #f))
;
(define (grow-orbital-po nsize)
   (let* ((n (+ (* nsize 2) 5))
          (nvec (make-vector n '()))
          (nvec2 (make-vector n '(()))))
      (let loop ((i (- *orbital-po-size* 1)))
         (when (>= i 0)
            (vector-set! nvec i (vector-ref *orbital-po* i))
            (vector-set! nvec2 i (vector-ref *po-index* i))
            (loop (- i 1))))
      (set! *orbital-po* nvec)
      (set! *po-index* nvec2)
      (set! *orbital-po-size* n)))
;
;   This is our primary algorithm for calculating min paths for the
;   partial order.  It uses some tricks to speed up the calculations.
;   First, it uses cache arrays so that it only calculates those values
;   that it needs.  Secondly, it recognizes `infinite' maxes so these
;   values need not be computed.  Thirdly, it uses the orbital-po
;   array to quickly find predecessors in the partial order.
;   Even with all of these tricks, however, it is probably somewhat
;   slow, which is why we can turn it off.
;
(define *po-level* 0)
;
(define *po-pos* '())
;
(define (po-timing lxpo lypo)
   (if (eq? lxpo lypo)
      0
      (let ((lxl (po-level lxpo))
            (lyl (po-level lypo)))
         (if (> lyl lxl)
            (if (<= lyl *bv-level*)
               (dref *bv-matrix* (po-index lxpo) (po-index lypo))
               (vector-ref (po-max-cache lypo) (po-index lxpo)))
            (if (<= lxl *bv-level*)
               (dref *bv-matrix* (po-index lxpo) (po-index lypo))
               (vector-ref (po-min-cache lxpo) (po-index lypo)))))))
;
(define (po-timing-ij i j)
  (po-timing (car (vector-ref *po-index* i)) (car (vector-ref *po-index* j))))
;
;   This function is how it is normally called.  At a given point, we want
;   the orbit-to-orbit timing.
;
(define *translate-deltas* '())
;
(define (calc-po-timing x y)
   (t+ (- (vector-ref *translate-deltas* x) (vector-ref *translate-deltas* y))
      (po-timing (car (vector-ref *orbital-po* x))
                 (car (vector-ref *orbital-po* y)))))
;
;   Sees if two sorted lists have an element in common.
;
(define (sorted-intersect? a b)
   (and a b
      (if (< (car a) (car b)) (sorted-intersect? (cdr a) b)
         (if (= (car a) (car b)) #t
            (sorted-intersect? (cdr b) a)))))
;
(define (strans-po-data t)
   (when (null? (strans-po-stuff-cache t))
      (let ((orb '()) (min '()) (max '()) (maxi '()))
         (for-each
            (lambda (e)
               (let ((p (edge-innode e)))
                  (when (node-orbit p)
                     (set! orb (cons (node-orbit p) orb))
                     (if (splace-time-constraint? p)
                        (set! min (cons 0 min))
                        (begin
                           (set! min
                              (if (edge-ignore-timing? e) (cons 0 min)
                                 (cons (t- (car (splace-time-marking p)))
                                    min)))
                           (when (not (eq? 'infinity
                                         (cdr (splace-time-marking p))))
                              (set! max (cons (cdr (splace-time-marking p))
                                              max))
                              (set! maxi (cons (node-orbit p) maxi))))))))
            (node-inedges t))
         (strans-po-stuff-cache-set! t (list orb min max maxi))))
   (strans-po-stuff-cache t))
;
(define (complex-esg-bit t)
   (if (pair? t)
      (let loop ((tl (cdr t))
                 (v (eps-subgraph-serial (strans-eps-connected (cadr t)))))
         (if tl
            (loop (cdr tl) (+ v (strans-esg-bit (car tl))))
            v))
      (strans-esg-bit t)))
;
;   When we execute a transition, we have to create a new po structure.
;   Much of this structure is cached.
;
;   *NOTE*  (strans-orbits t) may be a superset of
;   (car (strans-po-stuff-cache t)); the former includes all orbits that
;   are affected, but the latter only includes the backwards dependencies.
;   (Structurally, the other ones are automatically taken care of.)
;
(define *pre-prune-mins?* #t)
;
(define (do-po-work t)
   (set! *po-level* (+ 1 *po-level*))
   (when (> *po-level* *max-depth*)
      (set! *max-depth* *po-level*))
   (let ((po (make-po)) (cc 0) (orbs '()) (mins '()) (max1 '()) (maxi1 '())
         (need-big-vector #f) (to-release '()) (pominc '()) (pomaxc '()))
      (if *simulating?*
         (set! *po-pos* (list po))
         (set! *po-pos* (cons po *po-pos*)))
      (po-transition-set! po t)
      (po-level-set! po *po-level*)
      (po-index-set! po (po-index-get))
      (let ((n (po-index-max)))
; (display-stuff " " n)
	 (set! pominc (make-vector n 'infinity))
	 (set! pomaxc (make-vector n 'infinity))
         (po-min-cache-set! po pominc)
         (po-max-cache-set! po pomaxc))
      (when *use-po-pruning-now?*
         (if (and (not *simulating?*) t)
            (po-prune-forward po)
            (po-esg-bit-set! po -1)))
      (if t
         (for-each
            (lambda (t)
               (let ((pos (strans-po-data t)))
                  (let loop ((os (car pos)) (om (cadr pos)))
                     (when os
                        (let ((p (safe-car (vector-ref *orbital-po*
                                               (car os)))))
                           (when (or (null? p) (>= (po-covered p) 0))
                              (set! orbs (cons p orbs))
                              (set! mins (cons (car om) mins))))
                        (loop (cdr os) (cdr om))))
                  (when (caddr pos)
                     (set! max1 (append (caddr pos) max1))
                     (set! maxi1 (append (cadr (cddr pos)) maxi1))))
               (for-each (lambda (o)
                           (let ((pol (vector-ref *orbital-po* o))
                                 (d (vector-ref *translate-deltas* o)))
                              (set! cc (+ cc 1))
                              (when (not (= 0 d))
                                 (po-trans-push o (- d)))
                              (when pol
                                 (let ((nco (- (po-covered (car pol)) 1)))
                                    (po-covered-set! (car pol) nco)
                                    (when (= nco 0)
                                    (set! to-release
                                       (cons (po-index (car pol))
                                          to-release)))))
                              (when (>= o *orbital-po-size*)
                                 (grow-orbital-po o))
                              (vector-set! *orbital-po* o
                                 (cons po
                                    (if *simulating?*
                                       (cons (car (vector-ref *orbital-po* o))
                                          '())
                                       (vector-ref *orbital-po* o))))))
                  (strans-orbits t)))
            (if (pair? t) (cdr t) (list t)))
         (let loop ((i (- (graph-maxorbits *graph*) 1)))
            (when (>= i 0)
               (let ((p (vector-ref (sgraph-token-locator *sgraph*) i)))
                  (vector-set! *orbital-po* i (cons po '()))
                  (set! cc (+ cc 1))
                  (when (and p (splace-marked? p))
                     (set! orbs (cons '() orbs))
                     (set! mins (cons 0 mins))))
               (loop (- i 1)))))
      (po-covered-set! po cc)
      (po-maxcover-set! po cc)
;
;   Now we kill off orbs and mins according to their relative distances.
;   If maxdiff(o1,o2) <= min2, then we can kill off o2 and min2.
;
      (when *pre-prune-mins?*
         (let loop1 ((o1 orbs) (m1 mins))
            (when o1
               (when (and (car m1) (cdr o1))
                  (let loop2 ((o2 (cdr o1)) (m2 (cdr m1)))
                     (when (and o2 (car m2))
                        (when (car m1)
                           (if (t< (if (eq? (car m2) 'neginfinity)
                                       0 (car m2))
                                   (po-timing (car o1) (car o2)))
                              (if (t< (if (eq? (car m1) 'neginfinity)
                                         0 (car m1))
                                      (po-timing (car o2) (car o1)))
                                 #f ; do nothing
                                 (set-car! m1 '()))
                              (set-car! m2 '())))
                        (loop2 (cdr o2) (cdr m2)))))
               (loop1 (cdr o1) (cdr m1))))
         (let loop ((o1 orbs) (m1 mins) (no '()) (nm '()))
            (if o1
               (if (car m1)
                  (loop (cdr o1) (cdr m1) (cons (car o1) no)
                        (cons (car m1) nm))
                  (loop (cdr o1) (cdr m1) no nm))
               (begin
                  (set! orbs no)
                  (set! mins nm)))))
;
;   Now we carefully add data flow if necessary.  We only add those
;   orbits that are still `fresh'.  Oh yeah, we also update the
;   join histories.
;
      (when (and *dataflow-important?* t)
	 (let ((r '()) (w '()) (to-add '()))
            (for-each (lambda (t)
		         (when (strans-df-rw t)
			    (set! r (union r (car (strans-df-rw t))))
			    (set! w (union w (cdr (strans-df-rw t))))))
               (if (pair? t) (cdr t) (list t)))
	    (when (or r w)
	       (set! r (set-difference r w))
;  (display-stuff "(" (length r) "," (length w) ")")
;
;   The reads must follow the most recent write.  The writes must follow
;   all of the most recent reads, or, if none, the most recent write.
;
	       (po-maxcover-set! po (+ (po-maxcover po) (length r) (length w)))
	       (po-covered-set! po (+ (po-covered po) (length r) (length w)))
	       (for-each (lambda (jr)
			    (let ((hist (sdf-history jr)))
			       (when (and hist (pair? (car hist)))
				  (set! hist (cdr hist)))
			       (when (and hist (not (memq (car hist) to-add)))
				  (set! to-add (cons (car hist) to-add)))))
	          r)
	       (for-each (lambda (jw)
			    (let ((hist (sdf-history jw)))
			       (when hist
			          (set! to-add
				     (if (pair? (car hist))
				        (union to-add (car hist))
					(if (memq (car hist) to-add) to-add
					   (cons (car hist) to-add)))))))
	          w)
	       (for-each (lambda (jr)
			    (let ((hist (sdf-history jr)))
			       (if (and hist (pair? (car hist)))
			          (sdf-history-set! jr
				     (cons (cons po (car hist))
				        (cdr hist)))
				  (sdf-history-set! jr
				     (cons (cons po '()) hist)))))
	          r)
	       (for-each (lambda (jw)
			    (let* ((hist (sdf-history jw))
				   (fix (and hist
					   (if (po? (car hist))
					      (list (car hist))
					      (if (cdr hist)
					         (cons (cadr hist) (car hist))
						 (car hist))))))
			       (sdf-history-set! jw
			          (if *simulating?*
				     (list po)
				     (cons po hist)))
			       (for-each (lambda (pof)
                                 (let ((nco (- (po-covered pof) 1)))
                                    (po-covered-set! pof nco)
                                    (when (= nco 0)
                                       (set! to-release
                                          (cons (po-index pof)
                                             to-release)))))
			          fix)))
	          w)
	       (for-each (lambda (po)
			    (when (>= (po-covered po) 0)
			       (set! orbs (cons po orbs))
			       (set! mins (cons 0 mins))))
	          to-add))))
      (when orbs
         (for-each (lambda (pop min)
                      (let ((i (if pop (po-index pop) 0)))
                         (vector-set! pominc i
                            (tmin (vector-ref pominc i)
                               (if (eq? min 'neginfinity) 0 min)))))
            orbs mins)
         (for-each (lambda (index)
                      (let ((poi (car (vector-ref *po-index* index))))
                         (let loop ((m (if (eq? (car mins) 'neginfinity) 0
                                       (let ((pot (po-timing (car orbs) poi)))
                                          (if (eq? pot 'infinity) 'infinity
                                             (+ pot (car mins))))))
                                    (os (cdr orbs)) (ms (cdr mins)))
                           (if os
                              (loop (if (eq? (car ms) 'neginfinity) (tmin 0 m)
                                          (let ((pot (po-timing (car os) poi)))
                                             (if (eq? pot 'infinity) m
                                                (tmin (+ pot (car ms)) m))))
                                 (cdr os) (cdr ms))
                              (vector-set! pominc index m)))))
            (cdr *po-inuse*)))
      (when max1
;        (display-stuff "max1=" max1)
         (set! maxi1 (map-inorder (lambda (i)
                             (safe-cadr (vector-ref *orbital-po* i))) maxi1))
         (for-each (lambda (index)
                      (vector-set! pomaxc index
                         (t+ (car max1)
                            (po-timing (car (vector-ref *po-index* index))
                               (car maxi1)))))
            (cdr *po-inuse*))
         (for-each (lambda (max1 maxi1)
                      (for-each (lambda (index)
                                   (vector-set! pomaxc index
                                      (tmin (vector-ref pomaxc index)
                                         (t+ max1
                                             (po-timing
                                                (car (vector-ref *po-index*
                                                        index)) maxi1)))))
                         (cdr *po-inuse*)))
            (cdr max1) (cdr maxi1))
;        (display-stuff "cache:" (po-max-cache po))
         (let loop ((ml max1) (mil maxi1))
            (when (and ml (not need-big-vector))
               (let ((max1 (car ml)) (maxi1 (car mil)))
                  (let loop ((os orbs))
                     (when os
                        (let ((t (t+ max1 (vector-ref pominc
                                             (if (car os)
                                                (po-index (car os)) 0)))))
                           (if (t< t (po-timing maxi1 (car os)))
                              (set! need-big-vector #t)
                              (loop (cdr os)))))))
               (loop (cdr ml) (cdr mil)))))
      (when (>= (po-index po) *orbital-po-size*)
         (grow-orbital-po (po-index po)))
      (vector-set! *po-index* (po-index po)
         (cons po
            (if *simulating?*
               (cons (car (vector-ref *po-index* (po-index po))) '())
               (vector-ref *po-index* (po-index po)))))
;     (display-stuff "-")
      (for-each po-index-release to-release)
      (when need-big-vector
;        (display-stuff "!BV" (po-index-max) *bv-matrix* "!")
;   (display-stuff "*")
         (new-bv po
            (map (lambda (maxi1) (if maxi1 (po-index maxi1) 0)) maxi1)))
      (when #f
         (let ((v (zero-diag (dmake (po-index-max) 'infinity))))
            (map (lambda (a)
                   (map (lambda (b)
                          (dset! v a b
                             (po-timing-ij a b)))
                        *po-inuse*))
                 *po-inuse*)
            (when (< (apsp-check v) 0)
               (display-stuff 'newline "    ****    VIOLATION    ****"
                  'newline v 'newline)
               (merror "foo" 0))))
    ))
;
;   When we find out that just a new row and column isn't accurate, we
;   allocate a brand new large matrix to hold the current state.
;
(define *bv-level* -1)
;
(define *bv-matrix* '())
;
(define *num-bv-calls* 0)
;
;   We want this to be relatively fast, but there is a lot of work that
;   has to be done.
;
;   We use this matrix if *both* of the levels are below the current
;   *bv-level*.  If either is not, then we have a fresher value in the
;   appropriate cache array.
;
;   First, we copy the new min/max arrays; these are already calculated.
;   Then, we iterate through v[t,index], checking to see if the old value
;   is good, in which case we just copy the old column, or if we need to
;   update the entire row, in which case we do the appropriate calculations.
;
;   It's pretty straightforward, except for potentially n^2 calls to
;   po-timing.
;
(define (dbmake n init)
   (if (< n 2)
      *default-0-region*
      (let ((v (make-vector n '())))
         (for-each (lambda (i)
                      (vector-set! v i (make-vector n init)))
            *po-inuse*)
         v)))
;
(define *early-removal?* #t)
;
(define (new-bv po tl)
;  (set! t3 (runtime))
   (set! *num-bv-calls* (+ 1 *num-bv-calls*))
   (DBG 'b
      (display-stuff "BV: " *timed-tokens*))
   (let* ((n (po-index-max)) (v (dbmake n 'infinity))
          (min (po-min-cache po)) (max (po-max-cache po)) (ni (po-index po)))
      ;
      ;   Save some space by not allocating when not necessary
      ;
      (for-each (lambda (index)
                   (dset! v ni index (vector-ref min index))
                   (dset! v index ni (vector-ref max index))
                   (dset! v index index 0)
                   (let ((ii (car (vector-ref *po-index* index)))
                         (vi (vector-ref v index)))
                      (for-each (lambda (j)
                                   (vector-set! vi j
                                      (po-timing ii
                                              (car (vector-ref *po-index*
                                                      j)))))
                      (cdr *po-inuse*))))
         (cdr *po-inuse*))
      (dset! v ni ni 0)
      (for-each (lambda (t)
         (let ((tpo (car (vector-ref *po-index* t)))
               (maxv (vector-ref max t)))
            (for-each (lambda (j)
               (let* ((t1 (po-timing tpo (car (vector-ref *po-index* j))))
                      (minj (vector-ref min j))
                      (t2 (t+ minj maxv)))
; (display-stuff (if (t< t2 t1) "Y" "N"))
                   (when (t< t2 t1)
                      ; first case, new value is smaller, update entire row
                      (for-each (lambda (i)
                                   (dset! v i j
                                         (tmin (dref v i j)
                                            (t+ minj (vector-ref max i)))))
                            (cdr *po-inuse*)))))
               (cdr *po-inuse*))))
         tl)
      (when (not *simulating?*)
         (save-bv-matrix po *bv-level* *bv-matrix*))
;
;   Now we try to delete some rows that might not be necessary.  If, for
;   any i, for all j that are marked timed places, a[i,j] = infinity, then
;   we can delete row i early.
;
      (when *early-removal?*
         (let ((remove '())
                  (timeds '()))
            (let loop ((jj 1))
                  (when (<= jj *timed-tokens*)
                     (let ((po (car (vector-ref *orbital-po*
                                        (node-orbit
                                           (vector-ref *timed-token-ar* jj))))))
                        (set! timeds (cons (if po (po-index po) 0) timeds)))
                     (loop (+ jj 1))))
            (for-each (lambda (i)
                            (let ((r (vector-ref v i)))
                            (let loop ((ts timeds))
                                  (if ts
                                  (when (eq? (vector-ref r (car ts)) 'infinity)
                                        (loop (cdr ts)))
                                     (set! remove (cons i remove))))))
               (cdr *po-inuse*))
               (when remove
               (DBG 'b
                  (display-stuff "remove: " remove " <" *timed-tokens* ">"
                     " {" timeds "}" 'newline))
                  (for-each (lambda (i)
                            (let ((por (car (vector-ref *po-index* i))))
                               (po-covered-set! por
                                  (- (po-covered por) 1000000))
                               (po-index-release i)
                               (vector-set! v i '())
                               (po-early-release-set! po
                                  (cons por (po-early-release po)))))
                  remove))))
      (set! *bv-level* (po-level po))
      (set! *bv-matrix* v)))
;   (log-time 'bv t3)
;
(define (po-all-orbits po)
   (complex-orbits (po-transition po)))
;
(define (complex-delta t)
   (if (pair? t)
      (cadr (cddar t))
      (strans-state-value t)))
;
(define *bv-cache* '())
;
(define *bv-cache-ptr* 0)
;
(define *bv-cache-max* 500)
;
(define *bv-files* '())
;
(define *use-disk-stack?* #f)
;
(define (next-bv-file)
   (let ((n (string-append "temp-bv-stack."
               (number->string (length *bv-files*)))))
      (set! *bv-files* (cons n *bv-files*))
      (display-stuff ">>" n)
      (open-output-file n)))
;
(define (prev-bv-file)
   (let ((n (car *bv-files*)))
      (display-stuff "<<" n)
      (open-input-file n)))
;
(define (rm-bv-file)
   (let ((n (car *bv-files*)))
      (set! *bv-files* (cdr *bv-files*))
      (system (string-append "/bin/rm -f " n))))
;
(define (shift-out-bv-cache)
   (let* ((n (/ *bv-cache-max* 2))
          (f (next-bv-file)))
      (let fori ((i 0))
         (when (< i n)
            (write (vector-ref *bv-cache* i) f)
            (vector-set! *bv-cache* i (vector-ref *bv-cache* (+ i n)))
            (vector-set! *bv-cache* (+ i n) '())
            (fori (+ i 1))))
      (close-output-port f)
      (system (string-append "rm -f " (string-append (car *bv-files*) ".Z")))
      (system (string-append "compress " (car *bv-files*)))
      (set! *bv-cache-ptr* n)))
;
(define (shift-in-bv-cache)
   (system (string-append "rm -f " (car *bv-files*)))
   (system (string-append "uncompress " (car *bv-files*)))
   (let* ((n (/ *bv-cache-max* 2))
          (f (prev-bv-file)))
      (let fori ((i 0))
         (when (< i n)
            (vector-set! *bv-cache* i (read f))
            (fori (+ i 1))))
      (set! *bv-cache-ptr* n)
      (close-input-port f)
      (rm-bv-file)))
;
(define (collect-cache pos)
   (if (and pos (po-min-cache (car pos)))
      (let* ((po (car pos)) (r (cons (po-min-cache po) (po-max-cache po))))
         (po-min-cache-set! po '())
         (po-max-cache-set! po '())
         (cons r (collect-cache (cdr pos))))
      '()))
;
(define (distribute-cache vs pos)
   (when vs
      (let ((po (car pos)) (v (car vs)))
         (po-min-cache-set! po (car v))
         (po-max-cache-set! po (cdr v)))
      (distribute-cache (cdr vs) (cdr pos))))
;
(define (save-bv-matrix po lev mat)
   (if *use-disk-stack?*
      (begin
         (when (>= *bv-cache-ptr* *bv-cache-max*)
            (shift-out-bv-cache))
         (vector-set! *bv-cache* *bv-cache-ptr* (cons (cons lev mat)
                                                   (collect-cache *po-pos*)))
         (set! *bv-cache-ptr* (+ 1 *bv-cache-ptr*)))
      (begin
         (po-min-cache-set! po lev)
         (po-max-cache-set! po mat))))
;
(define (restore-bv-matrix po)
   (if *use-disk-stack?*
      (begin
         (when (<= *bv-cache-ptr* 0)
            (shift-in-bv-cache))
         (set! *bv-cache-ptr* (- *bv-cache-ptr* 1))
        (let ((v (vector-ref *bv-cache* *bv-cache-ptr*)))
            (vector-set! *bv-cache* *bv-cache-ptr* '())
            (set! *bv-level* (caar v))
            (set! *bv-matrix* (cdar v))
            (distribute-cache (cdr v) *po-pos*)))
      (begin
         (set! *bv-level* (po-min-cache po))
         (set! *bv-matrix* (po-max-cache po)))))
;
;   When we unfire a transition, we do this.
;
(define (undo-po-work t)
   (let ((po (car *po-pos*)))
;      (display-stuff "> " (vector-length (po-min-cache po)) " "
;         (po-min-cache po) " " (po-max-cache po) 'newline)
      (when (= *po-level* *bv-level*)
         (restore-bv-matrix po))
      (set! *po-level* (- *po-level* 1))
      (vector-set! *po-index* (po-index po)
         (cdr (vector-ref *po-index* (po-index po))))
      (for-each (lambda (p)
                   (vector-set! *translate-deltas* (car p)
                      (- (vector-ref *translate-deltas* (car p)) (cdr p))))
         (po-translate-deltas po))
      (for-each (lambda (por)
                   (po-covered-set! por
                      (+ (po-covered por) 1000000))
                   (po-index-grab (po-index por)))
         (po-early-release po))
      (po-index-release (po-index po))
      (when (and *use-po-pruning-now?* (not *simulating?*))
         (po-prune-backward po)))
   (set! *po-pos* (cdr *po-pos*))
   (when *dataflow-important?*
      (let ((r '()) (w '()))
         (for-each (lambda (t)
		      (when (strans-df-rw t)
			 (set! r (union r (car (strans-df-rw t))))
			 (set! w (union w (cdr (strans-df-rw t))))))
            (if (pair? t) (cdr t) (list t)))
	 (when (or r w)
	    (set! r (set-difference r w))
	    (for-each (lambda (jr)
			 (let ((hist (sdf-history jr)))
			    (sdf-history-set! jr
			       (if (pair? (car hist))
				  (if (cdar hist)
				     (cons (cdar hist) (cdr hist))
				     (cdr hist))
				  (cdr hist)))))
	       r)
	    (for-each (lambda (jw)
			 (let* ((hist (sdf-history jw))
				(chist (cdr hist))
				(fix (and chist
					(if (po? (car chist))
					   (list (car chist))
					   (if (cdr chist)
					      (cons (cadr chist) (car chist))
					      (car chist))))))
			    (sdf-history-set! jw (cdr hist))
			    (for-each (lambda (pof)
                                 (let ((nco (+ (po-covered pof) 1)))
                                    (po-covered-set! pof nco)
                                    (when (= nco 1)
				       (po-index-grab (po-index pof)))))
			          fix)))
	          w))))
   (for-each (lambda (t)
               (for-each (lambda (o)
                           (let ((pol (cdr (vector-ref *orbital-po* o))))
                              (when (>= o *orbital-po-size*)
                                 (grow-orbital-po o))
                              (vector-set! *orbital-po* o pol)
                              (when pol
                                 (let ((nco (+ (po-covered (car pol)) 1)))
                                    (po-covered-set! (car pol) nco)
                                    (when (= nco 1)
                                       (po-index-grab
                                          (po-index (car pol))))))))
                  (strans-orbits t)))
      (if (pair? t) (cdr t) (list t))))
;
(define (po-trans-push o d)
   (let ((po (car *po-pos*)))
      (vector-set! *translate-deltas* o
         (+ (vector-ref *translate-deltas* o) d))
      (po-translate-deltas-set! po
         (cons (cons o d) (po-translate-deltas po)))))
;
(define *write-sim-graph* #f)
;
(define (stest e inputs outputs)
   (build-graph e)
   (when *write-sim-graph*
      (with-output-to-file "sim.grf"
         (lambda () (send-graph *graph*))))
   (build-sgraph *graph* inputs outputs)
   (if (sgraph-tb-assoc *sgraph*)
      (display-stuff "Can't simulate with an unknown beh value."
         (map car (sgraph-tb-assoc *sgraph*)) 'newline)
      (simulate-internal)))
;
(define *do-bf-verification?* #t)
;
(define (vtest e inputs outputs)
   (set! *die-on-dead* #t)
   (build-graph e)
   (set! *save-arcs* #f)
   (when *write-sim-graph*
      (send-graph "sim.grf" *graph*))
   (build-sgraph *graph* inputs outputs)
   (set! *do-bf-verification?* #t)
   (if (sgraph-tb-assoc *sgraph*)
      (tverify-search prune-enabling)
      (begin
         (verify-internal prune-enabling)
         (when (and *verification-failed?* *do-bf-verification?*)
            (set-verification-variables 1 #f)
            (verify-bf-internal)))))
;
(define (vtest-bf e inputs outputs)
   (set! *die-on-dead* #t)
   (build-graph e)
   (set! *save-arcs* #f)
   (when *write-sim-graph*
      (send-graph "sim.grf" *graph*))
   (build-sgraph *graph* inputs outputs)
   (set! *do-bf-verification?* #t)
   (if (sgraph-tb-assoc *sgraph*)
      (merror "Don't support bf/tverify yet" 0)
      (verify-bf-internal)))
;
(define (vtest-no-bf e inputs outputs)
   (set! *die-on-dead* #t)
   (build-graph e)
   (set! *save-arcs* #f)
   (when *write-sim-graph*
      (send-graph "sim.grf" *graph*))
   (build-sgraph *graph* inputs outputs)
   (set! *do-bf-verification?* #f)
   (if (sgraph-tb-assoc *sgraph*)
      (tverify-search prune-enabling)
      (begin
         (verify-internal prune-enabling)
         (when (and *verification-failed?* *do-bf-verification?*)
            (set-verification-variables 1 #f)
            (verify-bf-internal)))))
;
(define (statetest e inputs outputs)
   (set! *die-on-dead* #t)
   (build-graph e)
   (set! *save-arcs* #t)
   (when *write-sim-graph*
      (send-graph "sim.grf" *graph*))
   (write-state-graph "state-graph.grf")
   (build-sgraph *graph* inputs outputs)
   (verify-internal prune-enabling))
;
(define (mintest e inputs outputs nonrok)
   (set! *die-on-dead* #f)
   (build-graph e)
   (when *write-sim-graph*
      (send-graph "sim.grf" *graph*))
   (set! *save-arcs* #t)
   (write-state-graph "state-graph.grf")
   (build-sgraph *graph* inputs outputs)
   (verify-internal prune-enabling)
   (write-state-graph "min-state-graph.grf")
   (minimize nonrok))
;
(define (normalize-internal e inputs outputs nonrok)
   (set! *die-on-dead* #f)
   (build-graph e)
   (when *write-sim-graph*
      (send-graph "sim.grf" *graph*))
   (set! *save-arcs* #t)
   (build-sgraph *graph* inputs outputs)
   (verify-internal prune-enabling)
;  (write-state-graph "state-graph.grf")
   (build-large-expr-from-min (append inputs outputs)
      (minimize nonrok) nonrok))
;
(define (sstest e inputs outputs)
   (set! *die-on-dead* #f)
   (set! *expand-eps-places* #t)
   (build-graph e)
   (set! *save-arcs* #f)
   (write-state-graph "state-graph.grf")
   (when *write-sim-graph*
      (send-graph "sim.grf" *graph*))
   (build-sgraph *graph* inputs outputs)
   (verify-internal ss-prune-enabling))
;
(define (vstest e inputs outputs)
   (set! *expand-eps-places* #t)
   (set! *die-on-dead* #f)
   (build-graph e)
   (set! *save-arcs* #f)
   (write-state-graph "state-graph.grf")
   (when *write-sim-graph*
      (send-graph "sim.grf" *graph*))
   (build-sgraph *graph* inputs outputs)
   (verify-internal prune-enabling))
;
(define (ttest e inputs outputs)
   (set! *die-on-dead* #t)
   (build-graph e)
   (set! *save-arcs* #f)
   (write-state-graph "state-graph.grf")
   (when *write-sim-graph*
      (send-graph "sim.grf" *graph*))
   (build-sgraph *graph* inputs outputs)
   (tverify-iterate prune-enabling))
;
(define (tbtest e inputs outputs)
   (set! *die-on-dead* #t)
   (build-graph e)
   (set! *save-arcs* #f)
   (write-state-graph "state-graph.grf")
   (when *write-sim-graph*
      (send-graph "sim.grf" *graph*))
   (build-sgraph *graph* inputs outputs)
   (tverify-search prune-enabling))
;
(define (shape-regions)
   (let ((t '()))
      (hash-iter (lambda (s) (set! t (cons (state-shapes s) t))) *state-hash*)
      t))
;
;   Prints out a region nicely.
;
(define (region-pp v)
   (let* ((n (vector-length v))
          (len (make-vector n 1))
          (nv (dcopy v)))
      (let fori ((i 0))
         (when (< i n)
            (let forj ((j 0))
               (when (< j n)
                  (if (or (= i j)
                         (and (> i 0) (> j 0)
                            (= (dref v i j)
                               (+ (dref v i 0) (dref v 0 j)))))
                     (dset! nv i j "-")
                     (let* ((s (number->string (dref v i j)))
                            (sl (string-length s)))
                        (dset! nv i j s)
                        (when (> sl (vector-ref len j))
                           (vector-set! len j sl))))
                  (forj (+ j 1))))
            (fori (+ i 1))))
      (let fori ((i 0))
         (when (< i n)
            (display-stuff
               (if (= n 1) "<"
                  (if (= i 0) "/"
                     (if (= i (- n 1)) "\\" "|"))))
            (let forj ((j 0))
               (when (< j n)
                  (if (> j 0)
                     (display-stuff " "))
                  (let* ((s (dref nv i j)))
                     (display-stuff
                        (make-string (- (vector-ref len j)
                                        (string-length s)) #\space) s))
                  (forj (+ j 1))))
            (display-stuff
               (if (= n 1) ">"
                  (if (= i 0) "\\"
                     (if (= i (- n 1)) "/" "|"))) 'newline)
            (fori (+ i 1))))))
;
(define *show-statistics?* #t)
;
(define (pad-to n s)
   (let ((d (- n (string-length s))))
      (if (> d 0)
         (string-append (make-string d #\space) s)
         s)))
;
(define (show-statistics n)
   (define (status s v)
      (when (number? v)
         (set! v (number->string v)))
      (display-stuff s ":" (pad-to (- 17 (string-length s)) v)
         (if (= n 3) 'newline "  "))
      (set! n (if (= n 3) 0 (+ n 1))))
   (display-stuff 'newline)
   (if (and *state-hash* *timing-important?*)
      (let ((reg 0) (sum 0))
         (hash-iter (lambda (s)
                       (set! reg (+ reg (length (state-shapes s))))
                       (for-each (lambda (s)
                                    (set! sum (+ sum (vector-length s))))
                          (state-shapes s)))
            *state-hash*)
         (let ((v (if *state-hash* (hash-size *state-hash*) 0)))
            (when (> v 0)
               (status "States" v)))
         (status "Regions" reg)
         (set! sum (quotient (+ (* 20 sum) reg) (* 2 reg)))
         (status "Avg reg" (string-append (number->string (quotient sum 10))
                              "." (number->string (remainder sum 10)))))
      (let ((v (if *state-hash* (hash-size *state-hash*) 0)))
         (when (> v 0)
            (status "States" v))))
   (when (> *num-firings* 0)
      (status "Firings" *num-firings*))
   (when (and *use-po-timing?* (> *num-bv-calls* 0))
      (status "BV calls" *num-bv-calls*))
   (when (and *use-po-timing?* (> *max-depth* 0))
      (status "Max depth" *max-depth*))
   (let ((v (if *array-hash* (hash-size *array-hash*) 0)))
      (when (> v 0)
         (status "Arr cache" v)))
   (let ((v (if (and (sgraph-esg-hash *sgraph*) *use-esg-hash?*)
               (hash-size (sgraph-esg-hash *sgraph*)) 0)))
      (when (> v 0)
         (status "ESG cache" v)))
   (when (> n 0)
      (display-stuff 'newline))
   (if *verification-failed?*
      (display-stuff "Verification failed." 'newline)
      (display-stuff "Verification succeeded." 'newline)))
;
(define *graph-name* 0)
;
(define *last-graph-name* #f)
;
(define (new-graph-filename)
   (set! *graph-name* (remainder (+ 1 *graph-name*) 10))
   (string-append "orbits" (number->string *graph-name*) ".grf"))
;
(define (save-gr-recurse cs? filename by-orbits? expr args)
   (if args
      (if (eq? (car args) 'by-orbits)
         (save-gr-recurse cs? filename #t expr (cdr args))
	 (if (string? (car args))
	    (save-gr-recurse cs? (car args) by-orbits? expr (cdr args))
	    (save-gr-recurse cs? filename by-orbits? (car args) (cdr args))))
      (if filename
         (begin
	    (when (and expr cs?)
	       (merror
    "Can't save the current state of a graph to be constructed" expr))
	    (when expr
               (build-graph expr))
	    (display-stuff " graph -> " filename)
	    (set! *last-graph-name* filename)
	    (set! *use-sim-state* cs?)
	    (if by-orbits?
	       (send-subgraph-by-orbits filename *graph*)
	       (let ((f (open-output-file filename)))
                  (send-graph f *graph*)
		  (close-output-port f)))
	    (display-stuff newline))
         (save-gr-recurse cs? (new-graph-filename) by-orbits? expr args))))
;
(define (save-graph-internal args)
   (save-gr-recurse #f #f #f #f args))
;
(define (save-graph-cs-internal args)
   (save-gr-recurse #t #f #f #f args))
;
(define (save-sg-recurse norm? nonrok? filename ckt args)
   (if args
      (if (eq? (car args) 'normalize)
         (save-sg-recurse #t nonrok? filename ckt (cdr args))
	 (if (string? (car args))
	    (save-sg-recurse norm? nonrok? (car args) ckt (cdr args))
	    (if (eq? (car args) 'ignore-receptiveness)
	       (save-sg-recurse norm? #t filename ckt (cdr args))
	       (save-sg-recurse norm? nonrok? filename
	          (car args) (cdr args)))))
      (if filename
	 (begin
	    (set! *last-graph-name* filename)
	    (when (null? ckt)
	       (merror "Need a circuit to normalize." ckt))
            (let* ((ckt (verify-ckt? ckt *global-assoc*))
		   (name (macro-name ckt))
		   (margs (macro-formals ckt))
		   (ins (my-sublist margs 0 (macro-inputs ckt)))
		   (outs (my-sublist margs (macro-inputs ckt)
				  (macro-outputs ckt))))
	      (build-graph `(with-wires global ,(append ins outs)
			      ,(cons name (append ins outs))))
	      (set! *die-on-dead* #f)
	      (set! *save-arcs* #t)
	      (build-sgraph *graph* ins outs)
	      (when (not norm?)
	         (write-state-graph filename)
	         (display-stuff " -> " filename " "))
	      (verify-internal prune-enabling)
	      (when norm?
		 (write-state-graph filename)
		 (display-stuff " -> " filename " ")
	         (minimize nonrok?))
	      (display-stuff 'newline)))
         (save-sg-recurse norm? nonrok? (new-graph-filename) ckt args))))
;
(define (save-state-graph-internal args)
   (save-sg-recurse #f #f #f #f args))
;
(define (X-display-graph-internal args)
   (let ((filename (or (and args (car args)) *last-graph-name*)))
      (system (string-append "parg -x " filename " &"))))
;
(define (PS-display-graph-internal args)
   (let ((filename (or (and args (car args)) *last-graph-name*)))
      (system (string-append "parg -p " filename
			     " orbits.eps; open orbits.eps"))))
;
(define (graph-to-ps-internal args)
   (let ((filename (or (and (cdr args) (car args)) *last-graph-name*)))
      (system (string-append "parg -p " filename " " (car (reverse args))))))
;
(define (o a b)
  (lambda (c)
    (a (b c))))
;
(define (delete-row a r)
   (let* ((n (vector-length a)) (ap (dmake (- n 1) 0)))
      (let fori ((i 0) (ip 0))
         (when (= r ip)
            (set! ip (+ 1 ip)))
         (when (< i (- n 1))
            (let ((aip (vector-ref a ip)) (api (vector-ref ap i)))
               (let forj ((j 0) (jp 0))
                  (when (= r jp)
                     (set! jp (+ 1 jp)))
                  (when (< j (- n 1))
                     (vector-set! api j (vector-ref aip jp))
                     (forj (+ j 1) (+ jp 1)))))
            (fori (+ i 1) (+ ip 1))))
      ap))
;
(define (submatrix a rws)
   (let* ((n (length rws)) (ap (dmake n 0)))
      (let fori ((i 0) (rw rws))
         (when (< i n)
            (let ((aip (vector-ref a (car rw))) (api (vector-ref ap i)))
               (let forj ((j 0) (rwj rws))
                  (when (< j n)
                     (vector-set! api j (vector-ref aip (car rwj)))
                     (forj (+ j 1) (cdr rwj)))))
            (fori (+ i 1) (cdr rw))))
      ap))
;
(define (choose m n)
   (define (fchoose m n v d)
      (if (<= n 0) (/ v d)
         (fchoose (- m 1) (- n 1) (* v m) (* n d))))
   (if (or (< n 0) (> n m)) 0
      (fchoose m (if (> (+ n n) m) (- m n) n) 1 1)))
;
(define *dsc-tune* 100)
;
(define (discrete-count a)
   ; is a triangle if all maxes are equal and all interior entries are
   ; equal to 0 or max.
   (define (check-triangle a n)
      (and (apply = (cdr (vector->list (vector-ref a 0))))
         (let fori ((i 1))
             (if (< i n)
                (let forj ((j 1))
                   (if (< j n)
                      (if (or (= (dref a i j) 0) (= (dref a i j) (dref a 0 j)))
                         (forj (+ j 1))
                         #f)
                      (fori (+ i 1))))
                #t))))
   (define (swap a n i j)
      (swap-elements a i j)
      (let fork ((k 0))
         (when (< k n)
            (let ((t (dref a k i)))
               (dset! a k i (dref a k j))
               (dset! a k j t))
            (fork (+ k 1)))))
   ;
   ;   Sorts so that lower diagonal is >0.
   ;
   (define (tria-sort a n)
      (let fori ((i 2))
         (when (< i n)
            (let forj ((j 1))
               (when (< j i)
                  (when (= 0 (dref a i j))
                     (swap a n i j))
                  (forj (+ j 1))))
            (fori (+ i 1))))
      a)
   (define (tria-lower a n)
      (let fori ((i 1))
         (when (< i n)
            (let forj ((j (+ i 1)))
               (when (< j n)
                  (dset! a i j 0)
                  (forj (+ j 1))))
            (fori (+ i 1))))
      a)
   (define (quotient2pow n)
      (if (> n 1024)
         (* 1024 (quotient2pow (quotient n 1024)))
         (let loop ((n n) (v 1))
            (if (> n 2)
               (loop (quotient n 2) (* v 2))
               v))))
   (define (square-it a n)
      (let fori ((i 1))
         (when (< i n)
            (let forj ((j 1))
               (when (< j n)
                  (dset! a i j (+ (dref a i 0) (dref a 0 j)))
                  (forj (+ j 1))))
            (fori (+ i 1))))
      a)
   (define (Q-count a n)
      (- (apply * (map 1+ (vector->list (vector-ref a 0))))
         (apply + (map discrete-count
                     (dbm-difference (square-it (dcopy a) n) a)))))
   (define (split-at a n i mv)
 ; (display-stuff "S")
      (let ((a1 (dcopy a)) (a2 (dcopy a)))
         (dset! a1 0 i mv)
         (dset! a2 i 0 (- (1+ mv)))
         (apsp1 a1 0 i)
         (apsp1 a2 i 0)
         (+ (discrete-count a1) (discrete-count a2))))
   (define (square-triangle a n)
      (let fori ((i 1))
         (when (< i n)
            (when (not (= 0 (dref a i 0)))
               (let ((v (dref a i 0)))
                  (let forj ((j 0))
                     (when (< j n)
                        (dset! a i j (- (dref a i j) v))
                        (dset! a j i (+ (dref a j i) v))
                        (forj (+ j 1))))))
            (fori (+ i 1))))
      (let fori ((i 1))
         (when (< i n)
            (let forj ((j (+ i 1)))
               (when (< j n)
                  (when (< (dref a 0 i) (dref a 0 j))
                     (swap a n i j))
                  (forj (+ j 1))))
            (fori (+ i 1))))
      (cond
         ((check-triangle a n)
            (tria-sort a n)
 ; (display-stuff "T")
            (apply + (cons (choose (+ (dref a 0 1) n -1) (- n 1))
                        (map discrete-count
                           (dbm-difference a (tria-lower (dcopy a) n))))))
         ((< (dref a 0 (- n 1)) 3)
            (let ((ap1 (dcopy a)) (ap2 (dcopy a)))
               (dset! ap1 0 (- n 1) 0)
               (apsp1 ap1 0 (- n 1))
               (dset! ap2 (- n 1) 0 -1)
               (apsp1 ap2 (- n 1) 0)
               (+ (discrete-count ap1) (discrete-count ap2))))
         (#t
            (let ((v (and (<= (dref a 0 1) *dsc-tune*) (cache-array* a))))
               (or (safe-cdr v)
                  (let ((r
                     (let forj ((j 1) (mt 0) (mjv 0))
                        (if (= 1 j)
                           (forj (+ j 1) mt mjv)
                           (if (>= j n)
                              (if (> (* 4 mjv) (+ (dref a 1 0) (dref a 0 1)))
                                 (split-at a n 1 mt)
                                 (split-at a n 1 (quotient2pow (dref a 0 1))))
                              (let* ((t (- (dref a 0 j) (dref a 1 j)))
                                     (tv (min (+ t (dref a 1 0))
                                            (- (dref a 0 1) t)))
                                     (t2 (- (dref a j 1) (dref a j 0)))
                                     (tv2 (min (+ t2 (dref a 1 0))
                                             (- (dref a 0 1) t2))))
                                 (if (> tv tv2)
                                    (if (> tv mjv)
                                       (forj (+ j 1) t tv)
                                       (forj (+ j 1) mt mjv))
                                    (if (> tv2 mjv)
                                       (forj (+ j 1) t2 tv2)
                                       (forj (+ j 1) mt mjv)))))))))
  (DBG 'q (region-pp a))
                     (when v (set-cdr! v r))
                     r))))))
   (define (check-ortho n a)
      (let loop ((in '(1)) (checked '()))
         (if in
            (let ((i (car in)) (in (cdr in)))
               (let forj ((j 1))
                  (if (< j n)
                     (if (and (= (+ (dref a 0 j) (dref a i 0)) (dref a i j))
                            (= (+ (dref a 0 i) (dref a j 0)) (dref a j i)))
                        (forj (+ j 1))
                        (begin
                           (when (not (or (memq j in) (= i j)
                                         (memq j checked)))
                              (set! in (cons j in)))
                           (forj (+ j 1))))
                     (loop in (cons i checked)))))
            (if (= (length checked) (- n 1))
               (square-triangle (dcopy a) n)
 (begin ; (display-stuff "O")
               (let loop ((nin '()) (i 2))
                  (if (< i n)
                     (loop (if (memq i checked) nin (cons i nin)) (+ i 1))
                     (* (discrete-count (submatrix a (cons 0 checked)))
                        (discrete-count (submatrix a (cons 0 nin)))))))))))
   (define (check-for-equiv n a)
      (let fori ((i 1))
         (if (< i n)
            (let forj ((j 0))
               (if (< j i)
                  (if (= 0 (+ (dref a i j) (dref a j i)))
                     (discrete-count (delete-row a (max i j)))
                     (forj (+ j 1)))
                  (fori (+ i 1))))
            (check-ortho n a))))
   (let ((n (vector-length a)))
      ; (copy current head of discrete-count for this one.)
      (if (< n 5)
         (case n
            (0 1) (1 1)
            (2 (+ 1 (dref a 1 0) (dref a 0 1)))
            (3 (- (* (+ (dref a 1 0) (dref a 0 1) 1)
                     (+ (dref a 2 0) (dref a 0 2) 1))
                  (choose (- (+ (dref a 1 0) (dref a 0 2) 1) (dref a 1 2)) 2)
                  (choose (- (+ (dref a 2 0) (dref a 0 1) 1) (dref a 2 1)) 2)
               ))
            (4 (dbm3-count a)))
         (check-for-equiv n a))))
;
(define (dbm-list-diff l1 l2)
   (if l1
      (if l2
         (append (dbm-list-diff (cdr l1) l2)
            (dbm-list-diff (dbm-difference (car l1) (car l2)) (cdr l2)))
         l1)
      '()))
;
(define (discrete-state-count)
   (when *use-shape-equiv-subset?*
      (display-stuff "Warning: incorrect due to shape-equiv" 'newline))
   (when (or (not *combine-arrays?*) (not *combine-rows?*))
      (display-stuff "Warning: turning array caching on" 'newline))
   (set! *combine-arrays?* #t)
   (set! *combine-rows?* #t)
   (let loop ((s 0) (sl (shape-regions)))
;     (display-stuff 'newline "Total: " s " to go: " (length sl))
      (if sl
         (let loop2 ((msl (car sl)) (tosub '()))
            (if msl
               (let ((toadd (dbm-list-diff (list (car msl)) tosub)))
                  (set! s (+ s (apply + (map discrete-count toadd))))
                  (display-stuff " " (length toadd))
                  (loop2 (cdr msl) (append tosub toadd)))
               (loop s (cdr sl))))
         (display-stuff 'newline s " total discrete states." 'newline))))
;
;   Calculates volume of three-dimensional dbm's.
;
(define (dbm3-count a)
   (define (f a b)
      (if (< a b) (f b a)
         (+ (* (- a b) (choose (+ b 1) 2))
            (/ (* (+ (* 2 b) 1) (+ b 1) b) 6))))
   (define (g one two three four)
      (- (f (+ (- (dref a one two)) (dref a 0 two) (dref a one 0))
            (+ (- (dref a three four)) (dref a 0 four) (dref a three 0)))
         (* (+ (dref a 0 (- 6 one two)) (dref a (- 6 one two) 0) 1)
            (choose (+ (- (dref a one two)) (dref a 0 two) (dref a one 0) 1)
               2))))
   (+ (* (+ (dref a 0 1) (dref a 1 0) 1) (+ (dref a 0 2) (dref a 2 0) 1)
         (+ (dref a 0 3) (dref a 3 0) 1))
      (g 1 2 1 3) (g 1 3 2 3) (g 2 3 2 1) (g 2 1 3 1) (g 3 1 3 2) (g 3 2 1 2)))
;
;   Temporary new verify-bf-internal.
;
(define *stride* 10)
;
(define (verify-bf-internal)
   (set! *breadth-first?* #t)
   (set! *simulating?* #f)
   (logit "Verifying, breadth-first")
   (set! *graph-shape* '())
   (when (< *stride* 1)
      (merror "Stride must be at least 1" *stride*))
   (DBG 's
      (display-stuff '< *current-state* '> 'newline))
   (let* ((h (new-state-hash)) (root (list (add-new-state? h *current-state*)))
          (rem (prune-enabling (calculate-enabling)))
          (bottom 0) (top *stride*))
      (when (not rem)
         (merror "No enabled transitions at start state" 0))
      (set! *state-hash* h)
      (set! *start-state* (car root))
      (let loop ((rem rem) (ptr (vector (car root))) (this-lev '())
                 (next-lev '()))
         (DBG 'd (display-stuff "B=" (length rem) " "
              (map fire-label rem) " tl="
                         (length this-lev) " nl=" (length next-lev) 'newline))
         (DBG 'c (check-state-consistency))
         (if rem ; still more to do
            (let* ((t (car rem))
                   (log (fire-complex t))
                   (nrem (cdr rem))
                   (ns (add-new-state? h *current-state*))
                   (dead1 (complex-dead? t))
                   (nptr (if (= *stride* 1)
                            (vector ns log ptr t)
                            (vector ns log ptr t nrem)))
                   (fe (if (and (new-state?) (not dead1))
                          (prune-enabling (calculate-enabling)) '(#t)))
                   (dead? (or dead1 *firing-failed?* (null? fe))))
               (DBG 's (display-stuff "<" *current-state* ">"))
               (when dead?
                  (set! *verification-failed?* #t)
                  (when *big-states?*
                     (state-dead?-set! ns #t)))
               (if (and dead? *die-on-dead*)
                  (report-dead nptr #f)
                  (if (>= (current-depth) top)
                     (begin
                        (unfire-complex log)
                        (set! *current-state* (mstate-val (vector-ref ptr 0)))
                        (prune-enabling (calculate-enabling))
                        (loop nrem ptr this-lev
                           (if (or (not (new-state?)) dead?) next-lev
                              (cons nptr next-lev))))
                     (begin
                        (if (or (not (new-state?)) dead?)
                           (loop '() nptr this-lev next-lev)
                           (loop fe nptr this-lev next-lev))))))
;
;   No more to do for this transition; we go to the next in the list.
;   To get there, we need to backtrack some amount.
;
            (if (<= (current-depth) bottom)
               (if this-lev
                  (begin
                     (DBG 'd (display-stuff "Next group" 'newline))
                     (let loop ((from ptr) (to (car this-lev)))
                        (when (not (eq? from to))
                           (unfire-complex (vector-ref from 1))
                           (loop (vector-ref from 2) (vector-ref to 2))
                           (fire-complex (vector-ref to 3))))
                     (set! *current-state*
                        (mstate-val (vector-ref (car this-lev) 0)))
                     (let* ((fe (prune-enabling (calculate-enabling))))
                        (loop fe (car this-lev) (cdr this-lev) next-lev)))
                  (if next-lev
;
;   Nothing more on this level, which means we're done with this level and
;   we've accumulated everything we need for the next level.  The only
;   problem is, we just undid the transitions taking us to the next
;   level.  We have to be careful, since we might not have found a `new'
;   non-dead state for a while.  So we use the same strategy as before
;   with the loop.
;
                     (begin
                        (DBG 'd (display-stuff
                                   "Next level " (length next-lev) 'newline))
                        (DBG 'x (display-stuff
                                   "Next level " (length next-lev) 'newline))
                        (set! *graph-shape* (cons (length next-lev)
                                               *graph-shape*))
                        (let loop ((from ptr)
                                   (to (car next-lev)))
                           (when (not (eq? from to))
                              (if (> top bottom)
                                 (begin
                                    (set! bottom (+ bottom 1))
                                    (loop from (vector-ref to 2)))
                                 (begin
                                    (unfire-complex (vector-ref from 1))
                                    (loop (vector-ref from 2)
                                       (vector-ref to 2))))
                              (fire-complex (vector-ref to 3))))
                        (set! top (+ top *stride*))
                        (set! *current-state*
                           (mstate-val (vector-ref (car next-lev) 0)))
                        (let* ((fe (prune-enabling (calculate-enabling))))
                           (loop fe (car next-lev) (cdr next-lev) '())))
                     (begin
                        (display-stuff " " (hash-size h) " total states.")
                        (logit "Done")
                        (when *show-statistics?*
                           (show-statistics 0)))))
               (let ((log (vector-ref ptr 1)) (optr (vector-ref ptr 2)))
                  (unfire-complex log)
                  (set! *current-state* (mstate-val (vector-ref optr 0)))
                  (loop (vector-ref ptr 4) optr this-lev next-lev )))))))
